/* 
 * Copyright (c) 2001 by Matt Welsh and The Regents of the University of 
 * California. All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Author: Matt Welsh <mdw@cs.berkeley.edu>
 * 
 */

package mdw.sandStorm.internal;

import mdw.sandStorm.api.*;
import mdw.sandStorm.api.internal.*;
import mdw.sandStorm.core.*;
import mdw.sandStorm.main.*;
import java.util.*;
import mdw.util.*;

/** 
 * AggThrottle is used by thread managers to adjust their aggregation
 * level based on observations of queue length.
 *
 * @author Matt Welsh
 */
class AggThrottle {

  private static final boolean DEBUG = false;

  private StageWrapperIF stage;
  private String name;
  private ManagerIF mgr;
  private StageGraph stageGraph;

  private double lastOutputRate;
  private long lastMeasurementTime;
  private int measurementCount;

  private static final int STATE_DECREASING = 0;
  private static final int STATE_INCREASING = 1;
  private int state = STATE_DECREASING;
  private int increase_count = 0;

  private int minAggregation = 8;
  private int maxAggregation = -1;
  private int recalcWindow = 100;
  private double smoothConst = 0.5;

  private static final double REDUCE_FACTOR = 1.2;
  private static final double INCREASE_FACTOR = 1.2;
  private static final double LOW_WATER = 0.95;
  private static final double HIGH_WATER = 1.0;
  private static final double VERY_LOW_WATER = 0.25;
  private static final double VERY_HIGH_WATER = 2.0;

  private int aggregationTarget;

  AggThrottle(StageWrapperIF stage, ManagerIF mgr) {
    this.stage = stage;
    this.name = stage.getStage().getName();
    this.mgr = mgr;
    this.stageGraph = mgr.getProfiler().getGraphProfiler();
    SandstormConfig config = mgr.getConfig();

    this.minAggregation = config.getInt("global.batchController.minBatch", 
	minAggregation);
    this.maxAggregation = config.getInt("global.batchController.maxBatch", 
	maxAggregation);
    this.recalcWindow = config.getInt("global.batchController.recalcWindow",
	recalcWindow);
    this.smoothConst = config.getDouble("global.batchController.smoothConst",
	smoothConst);

    System.err.println("AggThrottle <"+name+"> created: maxAgg "+maxAggregation+", recalcWindow "+recalcWindow);
    this.aggregationTarget = this.maxAggregation;

    lastOutputRate = 0.0;
    lastMeasurementTime = System.currentTimeMillis();
    measurementCount = 0;

    mgr.getProfiler().add("AggThrottle outputRate for <"+name+">",
	new ProfilableIF() {
	public int profileSize() {
	int foo = getAggTarget(); // Recalculate
	return (int)(lastOutputRate * 1000.0);
	}
	});
    mgr.getProfiler().add("AggThrottle aggTarget for <"+name+">",
	new ProfilableIF() {
	public int profileSize() {
	int foo = getAggTarget(); // Recalculate
	return aggregationTarget;
	}
	});
  }

  public String toString() {
    return "AggThrottle <"+name+">";
  }

  synchronized int getAggTarget() {

    measurementCount++;
    if ((measurementCount % recalcWindow) != 0) {
      return aggregationTarget;
    }

    // First get measurement of output rate for this thread
    int outputCount = 0;
    StageGraphEdge edges[] = stageGraph.getEdgesFromStage(stage);
    if (edges == null) {
      // No sinks yet
      return aggregationTarget;
    }
    for (int i = 0; i < edges.length; i++) {
      StageGraphEdge edge = edges[i];
      if (edge.sink instanceof SinkProxy) {
	SinkProxy sp = (SinkProxy)edge.sink;
	outputCount += sp.enqueueCount;
	sp.enqueueCount = 0;
      }
    }
    if (DEBUG) System.err.println("AggThrottle <"+name+">: "+edges.length+" sinks, "+outputCount+" enqueue ops");

    long cur_time = System.currentTimeMillis();
    long time_elapsed = cur_time - lastMeasurementTime;
    lastMeasurementTime = cur_time;
    double curRate = ((double)outputCount * 1.0) / ((double)time_elapsed * 1.0);
    double outputRate = (smoothConst * lastOutputRate) + ((1.0 - smoothConst) * curRate);

    if (outputRate < (VERY_LOW_WATER*lastOutputRate)) {
      aggregationTarget = maxAggregation;
      state = STATE_INCREASING;
    }

    if (outputRate >= (VERY_HIGH_WATER*lastOutputRate)) {
      aggregationTarget = maxAggregation;
      state = STATE_INCREASING;
    }

    if (state == STATE_DECREASING) {
      if (outputRate < (LOW_WATER*lastOutputRate)) {
       	state = STATE_INCREASING;
      } else {
	aggregationTarget /= REDUCE_FACTOR;
	if (aggregationTarget < minAggregation) aggregationTarget = minAggregation;
      }
    }

    if (state == STATE_INCREASING) {
      if (outputRate > (HIGH_WATER*lastOutputRate)) {
	state = STATE_DECREASING;
      } else {
	aggregationTarget *= INCREASE_FACTOR;
	if (aggregationTarget > maxAggregation) aggregationTarget = maxAggregation;
      }
    }

    if (DEBUG) System.err.println("AggThrottle <"+name+">: outputRate "+MDWUtil.format(outputRate)+", last "+MDWUtil.format(lastOutputRate)+", state "+state+", aggTarget "+aggregationTarget);

    lastOutputRate = outputRate;
    return aggregationTarget;
  }

}

