/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.bloom;
import ostore.util.ByteUtils;
import ostore.util.DebugFlags;
import ostore.util.NodeId;
import ostore.util.QSException;
import ostore.util.SHA1Hash;
import ostore.util.SecureHash;
import ostore.util.TypeTable;
import ostore.util.Types;
import ostore.network.NetworkMessage;
import java.util.LinkedList;
import java.util.ListIterator;

/**
 * Update message.
 *
 * @author     Sean C. Rhea
 * @version    $Id: QueryMessage.java,v 1.3 2002/04/06 22:52:43 eaton Exp $
 */
public class QueryMessage extends NetworkMessage {

    private static final boolean DEBUG = DebugFlags.BLOOM;

    public long id;
    public SecureHash src;
    public SecureHash dest;
    public BloomQuery user_data;

    public QueryMessage (NodeId peer, long id, SecureHash src, 
	    SecureHash dest, BloomQuery user_data) {

	super (peer, false /* outbound */);
	this.id = id;
	this.src = src;
	this.dest = dest;
	this.user_data = user_data;
    }

    public QueryMessage (byte [] data, int [] offset) throws QSException {
	super (data, offset);
	id = ByteUtils.bytesToLong (data, offset);
	src = new SHA1Hash (data, offset);
	dest = new SHA1Hash (data, offset);
	int path_length = ByteUtils.bytesToInt (data, offset);
	if (path_length > 0) _path = new LinkedList ();
	for (int i = 0; i < path_length; ++i)
	    _path.addLast (new NodeId (data, offset));
        long type_code = ByteUtils.bytesToLong (data, offset);
	user_data = 
	    (BloomQuery) TypeTable.new_object (type_code, data, offset);
    }

    public void to_bytes (byte [] data, int [] offset) {
	super.to_bytes (data, offset);
	ByteUtils.longToBytes (id, data, offset);
	src.to_bytes (data, offset);
	dest.to_bytes (data, offset);
	if (_path == null)
	    ByteUtils.intToBytes (0, data, offset);
	else {
	    ByteUtils.intToBytes (_path.size (), data, offset);
	    ListIterator i = _path.listIterator ();
	    while (i.hasNext ()) {
		NodeId n = (NodeId) i.next ();
		n.to_bytes (data, offset);
	    }
	}
	ByteUtils.longToBytes (TypeTable.type_code(user_data), data, offset);
	user_data.to_bytes (data, offset);
    }

    public int path_length () { return (_path == null ? 0 : _path.size ()); }

    public void add_to_path (NodeId n) {
	if (_path == null) _path = new LinkedList ();
	_path.addLast (n);
    }

    public boolean on_path (NodeId n) {
	if (_path == null) return false;
	ListIterator i = _path.listIterator ();
	while (i.hasNext ()) {
	    NodeId m = (NodeId) i.next ();
	    if (n.equals (m)) return true;
	}
	return false;
    }

    public int type_code () {
	throw new NoSuchMethodError();
        //return Types.BLOOM_UPDATE_MESSAGE;
    }

    public Object clone () throws CloneNotSupportedException {
	QueryMessage result = (QueryMessage) super.clone ();
	result.id = id;
	result.src = src;
	result.dest = dest;
	result.user_data = user_data;

	// Need to copy this list, otherwise we get weird semantics for the
	// following code:
	//
	//   QueryMessage a = new QueryMessage (...);
	//   QueryMessage b = a.clone ();
	//   b.add_to_path (n);
	//
	// We don't want n added to a's path, only b's, so we have to copy
	// the list (although not it's elements).

	result._path = (_path == null ? null : new LinkedList (_path));

	return result;
    }

    public String toString () {
	String result = "(QueryMessage " + id + " " + src + "->" +
	    dest + (inbound ? " via " : " at ") + peer.port () + " " +
	    user_data + ")";
	return result;
    }

    private LinkedList _path;
}

