/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.bloom;

import mdw.sandStorm.api.ConfigDataIF;
import mdw.sandStorm.api.EventHandlerException;
import mdw.sandStorm.api.EventHandlerIF;
import mdw.sandStorm.api.QueueElementIF;
import mdw.sandStorm.api.SinkIF;
import mdw.sandStorm.api.SinkException;
import mdw.sandStorm.api.StagesInitializedSignal;
import ostore.util.ByteUtils;
import ostore.util.Carp;
import ostore.util.Clock;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.NodeId;
import ostore.util.Assert;
import ostore.util.PriorityQueue;
import ostore.util.QuickSerializable;
import ostore.util.SecureHash;
import ostore.util.SHA1Hash;
import ostore.util.Types;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.tapestry.api.TapestryInstance;
import ostore.tapestry.api.TapestryPublishMsg;
import ostore.tapestry.api.TapestryUnpublishMsg;
import ostore.tapestry.api.TapestryLocateReject;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.StreamTokenizer;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.Reader;

/**
 * The probabilistic location stage.
 *
 * @author     Sean C. Rhea
 * @version    $Id: SimulatedReplica.java,v 1.3 2002/02/09 01:22:45 srhea Exp $
 */
public class SimulatedReplica implements EventHandlerIF {

    private String class_tag = "SimulatedReplica";
    private static final boolean DEBUG = DebugFlags.BLOOM;

    private static Map /* NodeId -> Map SecureHash -> Integer */ _all_disks;
    private static Map /* NodeId -> Map SecureHash -> Integer */ _all_caches;
    private static Map /* SecureHash -> NodeId */ _guid_to_node_id;
    static {
	_all_disks = new HashMap ();
	_all_caches = new HashMap ();
	_guid_to_node_id = new HashMap ();
    }

    private static Map _instances;
    static {
	_instances = new HashMap ();
    }

    public static SimulatedReplica instance (NodeId node_id) {
	return (SimulatedReplica) _instances.get (node_id);
    }

    private void dispatch (QueueElementIF item) {
	final String tag = class_tag + ".dispatch";
	if (DEBUG) Debug.printtagln (tag, "dispatching " + item + ".");
	try {
	    _classifier.dispatch(item);
	} catch (SinkException e) {
	    BUG (class_tag + ".dispatch: could not dispatch " + item);
	}
    }

    private void dispatch_later (QueueElementIF item, long millis) {
	final String tag = class_tag + ".dispatch";
	if (DEBUG) Debug.printtagln (tag, "dispatching " + item + ".");
	_classifier.dispatch_later(item, millis);
    }

    private static final void BUG (String msg) {
	Carp.die (msg + "  This is a bug in ostore.bloom.SimulatedReplica.  " +
		"Please let Sean know about it ASAP.");
    }

    public long closest_replica_lat_ms (SecureHash guid) {
	soss.network.Network net = soss.network.Network.instance ();
	long closest = Long.MAX_VALUE;
	for (Iterator i = _all_disks.keySet ().iterator (); i.hasNext (); ) {
	    NodeId other_node_id = (NodeId) i.next ();
	    Map other_disk = (Map) _all_disks.get (other_node_id);
	    if (other_disk.containsKey (guid)) {
		long dist = net.route_time_ms (
			_self_node_id, other_node_id, 0);
		if (dist < closest)
		    closest = dist;
	    }
	}
	for (Iterator i = _all_caches.keySet ().iterator (); i.hasNext (); ) {
	    NodeId other_node_id = (NodeId) i.next ();
	    Map other_disk = (Map) _all_caches.get (other_node_id);
	    if (other_disk.containsKey (guid)) {
		long dist = net.route_time_ms (
			_self_node_id, other_node_id, 0);
		if (dist < closest)
		    closest = dist;
	    }
	}
	return closest;
    }

    public long closest_replica_xfer_ms (SecureHash guid) {
	soss.network.Network net = soss.network.Network.instance ();
	long closest = Long.MAX_VALUE;
	for (Iterator i = _all_disks.keySet ().iterator (); i.hasNext (); ) {
	    NodeId other_node_id = (NodeId) i.next ();
	    Map other_disk = (Map) _all_disks.get (other_node_id);
	    if (other_disk.containsKey (guid)) {
		Integer size = (Integer) other_disk.get (guid);

		// route there with a request, and back with the document
		long dist = 
		    net.route_time_ms (
			    _self_node_id, other_node_id, 0) +
		    net.route_time_ms (
			    _self_node_id, other_node_id, size.intValue ());


		if (dist < closest)
		    closest = dist;
	    }
	}
	for (Iterator i = _all_caches.keySet ().iterator (); i.hasNext (); ) {
	    NodeId other_node_id = (NodeId) i.next ();
	    Map other_disk = (Map) _all_caches.get (other_node_id);
	    if (other_disk.containsKey (guid)) {
		Integer size = (Integer) other_disk.get (guid);

		// route there with a request, and back with the document
		long dist = 
		    net.route_time_ms (
			    _self_node_id, other_node_id, 0) +
		    net.route_time_ms (
			    _self_node_id, other_node_id, size.intValue ());


		if (dist < closest)
		    closest = dist;
	    }
	}
	return closest;
    }

    public void init (ConfigDataIF config) throws Exception {

	if (DEBUG) Debug.turnOn (class_tag);

	Types.registerAllTypes ();

	_self_node_id = new NodeId (config.getString ("node_id"));
	_instances.put (_self_node_id, this);
	_simulation_log = SimulationLog.instance (_self_node_id);

	class_tag = class_tag + ".[" + _self_node_id.port () + "]";
	final String tag = class_tag + ".init";

	_classifier = Classifier.getClassifier (_self_node_id);
	_sink = config.getStage().getSink();

	_disk_files = new HashMap ();
	_all_disks.put (_self_node_id, _disk_files);
	_all_caches.put (_self_node_id, _cache_set);

	_max_cache_size = config.getInt ("max_cache_size");
	if (_max_cache_size > 0)
	    _caching = true;

	String file_list = config.getString ("file_list");
	{
	    int lineno = 1;
	    FileInputStream is = new FileInputStream (file_list);
	    Reader reader = new BufferedReader (new InputStreamReader (is));
	    StreamTokenizer tok = new StreamTokenizer (reader);
	    tok.parseNumbers ();

	    tok.nextToken ();

	    while (tok.ttype == StreamTokenizer.TT_EOL) {
		tok.nextToken ();
		//System.err.println (tok);
		lineno++;
	    }

	    while (tok.ttype != StreamTokenizer.TT_EOF) {

		if (tok.ttype != StreamTokenizer.TT_WORD) {
		    System.err.println ("Expected file name on line " + 
			    lineno + ": " + tok);
		    System.exit (1); 
		}
		String filename = tok.sval;

		tok.nextToken ();
		if (tok.ttype != StreamTokenizer.TT_NUMBER) {
		    System.err.println ("Expected size on line " + 
			    lineno + ".");
		    System.exit (1); 
		}
		int size = (int) tok.nval;

		SecureHash hash = new SHA1Hash (filename);
		if (DEBUG) Debug.printtagln (tag, "disk file \"" + 
			filename + "\" " + hash + " " + size + ".");
		_disk_files.put (hash, new Integer (size));

		tok.nextToken ();
		//System.err.println (tok);

		while (tok.ttype == StreamTokenizer.TT_EOL) {
		    tok.nextToken ();
		    //System.err.println (tok);
		    lineno++;
		}
	    }

	    is.close ();
	}

	String [] event_types = {
	    "mdw.sandStorm.api.StagesInitializedSignal",
	    "ostore.bloom.SimulatedReplica$DiskReadFinished"
	};

	for (int i = 0; i < event_types.length; ++i) {
	    Filter filter = new Filter ();
	    if (! filter.requireType (Class.forName (event_types [i])))
		BUG (tag + ": could not require type " + event_types [i]);
	    _classifier.subscribe (filter, _sink);
	}

	String [] inb_msg_types = {
	    "ostore.bloom.SimulatedRead",
	    "ostore.bloom.SimulatedReadResp"
	};

	for (int i = 0; i < inb_msg_types.length; ++i) {
	    Filter filter = new Filter ();
	    if (! filter.requireType (Class.forName (inb_msg_types [i])))
		BUG (tag + ": could not require type " + inb_msg_types [i]);
	    if (! filter.requireValue ("inbound", new Boolean (true)))
		BUG (tag + ": could not require inbound = true for " 
			+ inb_msg_types [i]);
	    _classifier.subscribe (filter, _sink);
	}
    }

    public void destroy() throws Exception {
	// TODO
    }

    public void handleEvent(QueueElementIF item) 
    throws EventHandlerException {
	final String tag = class_tag + ".handleEvent";

	if (DEBUG) Debug.printtagln (tag, 
		"got " + item.getClass () + " event.");

	if (item instanceof StagesInitializedSignal) {
	    handle_init ((StagesInitializedSignal) item);
	}
	else if (item instanceof SimulatedRead) {
	    handle_simulated_read ((SimulatedRead) item);
	}
	else if (item instanceof SimulatedReadResp) {
	    handle_simulated_read_resp ((SimulatedReadResp) item);
	}
	else if (item instanceof DiskReadFinished) {
	    handle_disk_read_finished ((DiskReadFinished) item);
	}
	else {
	    BUG (tag + ": caught unexpected event " + item + ".");
	}
    }

    public void handleEvents(QueueElementIF items[]) 
    throws EventHandlerException {
	for(int i=0; i<items.length; i++)
	    handleEvent(items[i]);
    }

    public void handle_init (StagesInitializedSignal signal) {
	final String tag = class_tag + ".handle_init";
	_guid_to_node_id.put (
		TapestryInstance.instance (_self_node_id).self_guid (), 
		_self_node_id);
	Iterator i = _disk_files.keySet ().iterator ();
	while (i.hasNext ())
	    dispatch (new TapestryPublishMsg ((SecureHash) i.next (), null));
    }

    private void add_to_cache (SecureHash guid, int size) {
	while (_cache_size + size > _max_cache_size) {
	    SecureHash file = (SecureHash) _cache_lru.extract (false);
	    Assert.assert (file != null, "not in cache lru");
	    Integer remsize = (Integer) _cache_set.remove (file);
	    Assert.assert (remsize != null, "not in cache set");
	    _cache_size -= remsize.intValue ();
	    if (! _disk_files.containsKey (file))
		dispatch (new TapestryUnpublishMsg (file, null));
	}
	// And add it in.
	_cache_set.put (guid, new Integer (size));
	_cache_lru.insert (guid, current_time ().getTime (), false);
	_cache_size += size;
	if (! _disk_files.containsKey (guid))
	    dispatch (new TapestryPublishMsg (guid, null));
    }

    public void handle_simulated_read_resp (SimulatedReadResp read) {
	final String tag = class_tag + ".handle_simulated_read_resp";

	if (_caching && (read.size <= _max_cache_size)) {

	    if (_cache_set.containsKey (read.guid)) {
		// If we already have this item in the cache, just remove
		// it's old LRU status.
		_cache_lru.remove (read.guid);
		_cache_lru.insert (read.guid, current_time ().getTime (), false);
	    }
	    else {
		// Otherwise, put in it.
		add_to_cache (read.guid, read.size);
	    }
	}
    }

    private Map _cache_set = new HashMap ();
    private PriorityQueue _cache_lru = new PriorityQueue (10);
    private int _cache_size;
    private int _max_cache_size;
    private boolean _caching;
    private static final double _disk_alpha = 10.0, _disk_beta = 0.0001;

    public class DiskReadFinished implements QueueElementIF {
	public DiskReadFinished (SecureHash g, int s) {
	    size = s; guid = g;
	}
	public int size;
	public SecureHash guid;
    }

    public void handle_simulated_read (SimulatedRead read) {
	final String tag = class_tag + ".handle_simulated_read";

	if (! (_cache_set.containsKey (read.guid) ||
		_disk_files.containsKey (read.guid))) {

	    // It was here, but now it's gone.  Send out a
	    // TapestryLocateReject to notify the location system that
	    // they've reached us in error.

	    dispatch (new TapestryLocateReject (read));
	    return;
	}
		
	soss.network.Network net = soss.network.Network.instance ();
	SimulationLog.ReadTransactionInfo xact = 
	    _simulation_log.xact (read.peer, read.guid);

	xact.location_finish_time = current_time ().getTime ();
	xact.provider = _self_node_id;
	xact.location_distance = net.route_time_ms (_self_node_id, 
		(NodeId) _guid_to_node_id.get (read.peer), 0);

	Integer size = null;
	if (_disk_files.containsKey (read.guid))
	    size = (Integer) _disk_files.get (read.guid);
	else
	    size = (Integer) _cache_set.get (read.guid);
	
	if ((!_caching) || _cache_set.containsKey (read.guid)) {
	    if (_caching) {
		// Update LRU
		_cache_lru.remove (read.guid);
		_cache_lru.insert (read.guid, current_time ().getTime (), 
			false);
	    }

	    if (DEBUG) Debug.printtagln (tag, 
		    "data in cache, returning it now.");

	    dispatch (new SimulatedReadResp (read.peer, read.guid, 
			size.intValue ()));
	}
	else {
	    synchronized (_disk_reads_in_progress) {
		Set in_progress = (Set) _disk_reads_in_progress.get (read.guid);
		if (in_progress == null) {

		    if (DEBUG) Debug.printtagln (tag, 
			    "data on disk, starting disk read.");

		    in_progress = new HashSet ();
		    in_progress.add (read.peer);
		    _disk_reads_in_progress.put (read.guid, in_progress);
		    long delay = Math.round (
			    size.intValue () * _disk_beta + _disk_alpha);
		    dispatch_later (new DiskReadFinished (read.guid, 
				size.intValue ()), delay);
		}
		else {
		    if (DEBUG) Debug.printtagln (tag, 
			    "data on disk, disk read in progress.");

		    in_progress.add (read.peer);
		}
	    }
	}
    }

    public void handle_disk_read_finished (DiskReadFinished fin) {
	final String tag = class_tag + ".handle_disk_read_finished";

	if (DEBUG) Debug.printtagln (tag, "finished disk read for " + 
		fin.guid);

	synchronized (_disk_reads_in_progress) {
	    if (fin.size <= _max_cache_size) {
		add_to_cache (fin.guid, fin.size);
	    }

	    Set in_progress = (Set) _disk_reads_in_progress.remove (fin.guid);
	    for (Iterator i = in_progress.iterator (); i.hasNext (); ) {
		SecureHash peer = (SecureHash) i.next ();
		if (DEBUG) Debug.printtagln (tag, "sending data to " + peer);
		dispatch (new SimulatedReadResp (peer, fin.guid, fin.size));
	    }
	}
    }

    private Date current_time () {
	return Clock.current_date (_self_node_id);
    }

    private static final void assert (boolean expr) {
	if (DEBUG) Assert.assert (expr, "BUG");
    }

    private SinkIF _sink;
    private NodeId _self_node_id;
    private Classifier _classifier;
    private Map _disk_reads_in_progress = new HashMap ();

    /**
     * Maps file names on the disk to their sizes.
     */
    private Map /* SecureHash -> Integer */ _disk_files;
    private SimulationLog _simulation_log;
    
}


