/**
 * Filter.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.dispatch;

import ostore.util.Debug;
import ostore.util.Carp;
import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.LinkedList;
import mdw.sandStorm.api.QueueElementIF;

/**
 * A specification for an event class.
 * Each <code>Filter</code> consists of a set of restrictions, the
 * intersection of which describe the set of <code>QueueElementIF</code>
 * events objects which satisfy the <code> Filter </code>.
 *
 * @author	Dennis Geels
 * @version $Id: Filter.java,v 1.20 2002/02/09 01:41:50 geels Exp $
 */

public class Filter
{
    ////////////////////       CONSTANTS	////////////////////

    private static final String tag = "Filter";    

    private static final boolean DEBUG =
	ostore.util.DebugFlags.DISPATCH_FILTER;
    
    ////////////////////    STATIC VARIABLES 	////////////////////

    /**
     * A <code>HashMap</code> containing the set of valid types for fields
     * whose value we would like to restrict.
     * If the type is a Java primitive wrapper, it maps to the
     * <code>Class</code> object for the associated primitive.
     * <p>
     * Update this HashMap (through the static initializer) if you want to
     * add an allowable type restriction.
     * <b>NOTE</b> all added types should override the default
     * <code>hashCode</code> method with their own deterministic version.
     */
    protected static HashMap valid_types;

    ////////////////////       VARIABLES	////////////////////

    /**
     * The fraction of qualifying events which the <code>Classifier</code>
     * should pass through the <code>Filter</code>. 
     * Must be between 0 and 1, inclusive.
     * Use this variable if we want to sample the event stream.
     */
    public double acceptance_rate;

    /**
     * A <code>HashMap</code> containing all the restrictions for this
     * <code>Filter</code>.
     * The key is a <code>Field</code> object, mapped to the value
     * which the <code>Field</code> should have.
     */
    public HashMap restrictions;

    /**
     * A <code>List</code> containing all the restrictions in the order
     * they were added.
     */
    public LinkedList ordered_restrictions;
    
    /**
     * The most recently required type.  This <code>Class</code> must be the
     * most restrictive one mentioned, so we can forget all previous
     * <code>requireType</code> calls and store just the last one.
     */
    public Class required_type;

    ////////////////////  STATIC CONSTRUCTORS	////////////////////

    static {
	valid_types = null;
	try {
	    valid_types = new HashMap();
	    // Map wrapper objects to their primitive types.  Otherwise, the
	    // value is ignored;  use <code>TRUE</code>.
	    
	    // Allow all the basic java.lang Objects:
	    valid_types.put( Class.forName("java.lang.String"), Boolean.TRUE );
	    valid_types.put( Class.forName("java.lang.Character"), Character.TYPE );
	    valid_types.put( Class.forName("java.lang.Boolean"), Boolean.TYPE );
	    valid_types.put( Class.forName("java.lang.Byte"), Byte.TYPE );
	    valid_types.put( Class.forName("java.lang.Short"), Short.TYPE );
	    valid_types.put( Class.forName("java.lang.Integer"), Integer.TYPE );
	    valid_types.put( Class.forName("java.lang.Long"), Long.TYPE );
	    valid_types.put( Class.forName("java.lang.Float"), Float.TYPE );
	    valid_types.put( Class.forName("java.lang.Double"), Double.TYPE );

	    // Allow all the primitive types:
	    valid_types.put( Character.TYPE, Boolean.TRUE );
	    valid_types.put( Boolean.TYPE, Boolean.TRUE );
	    valid_types.put( Byte.TYPE, Boolean.TRUE );
	    valid_types.put( Short.TYPE, Boolean.TRUE );
	    valid_types.put( Integer.TYPE, Boolean.TRUE );
	    valid_types.put( Long.TYPE, Boolean.TRUE );
	    valid_types.put( Float.TYPE, Boolean.TRUE );
	    valid_types.put( Double.TYPE, Boolean.TRUE );

	    // Allow QuickSerializable types
  	    valid_types.put( Class.forName("ostore.util.QSInt"), Boolean.TRUE );
  	    valid_types.put( Class.forName("ostore.util.QSLong"), Boolean.TRUE );
  	    valid_types.put( Class.forName("ostore.util.QSString"), Boolean.TRUE );

	    // Allow OceanStore-specific types:
//  	    valid_types.put( Class.forName("ostore.client.OSAppId"), Boolean.TRUE );
  	    valid_types.put( Class.forName("ostore.util.NodeId"), Boolean.TRUE );
  	    valid_types.put( Class.forName("ostore.util.SecureHash"), Boolean.TRUE );

	} catch ( java.lang.ClassNotFoundException e ) {
	    Carp.die( e );
	}

	if( DEBUG ) Debug.turnOn( tag );
    }

    ////////////////////      CONSTRUCTORS	////////////////////
    
    public Filter() {

	acceptance_rate = 1.0;
	restrictions = new HashMap();
	ordered_restrictions = new LinkedList();

	// initialize required_type to the <code>QueueElementIF</code> interface.
	try {
	    required_type =
	      Class.forName("mdw.sandStorm.api.QueueElementIF" );
	} catch ( java.lang.ClassNotFoundException e ) {
	    Carp.die( e );
	}
    }

    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Add a type restriction.
     * This method just looks up the named <code>Class</code> and calls
     * <code>requireType( Class )</code>.
     *
     * @param type_name	The name of the <code>Class</code> of which filtered
     * 			<code>QueueElementIF</code>s must be a member.
     * @return <code>true</code> iff the requirement was registered without
     * 		errors.
     */
    public boolean requireType( String type_name )
    {
	if( DEBUG ) Debug.printtagln( tag+".requireType",
				      "called: " + type_name );
	boolean valid = true;
	// Find the <code>Class</code> object.
	Class type = null;
	try {
	    type = Class.forName( type_name );
	} catch ( ClassNotFoundException e ) {
	    if( DEBUG ) Debug.printtagln( tag+".requireType",
					  "Invalid class name \"" +
					  type_name + "\"." );
	    valid = false;
	}
	if( valid ) {
	    valid = requireType( type );
	}
	return valid;
    }

    /**
     * Add a type restriction.
     * <code>QueueElementIF</code> objects must now be members of the
     * specified class, or they will not pass the filter.
     * The class must be a subclass of all classes required so far, which
     * makes sense.
     *
     * @param type	The <code>Class</code> of which filtered
     * 			<code>QueueElementIF</code>s must be a member. 
     * @return <code>true</code> iff the requirement was registered without
     *		errors.
     */
    public boolean requireType( Class type )
    {
	if( DEBUG ) Debug.printtagln( tag+".requireType",
				      "called: " + type );
	boolean valid = true;

	// If the type already exists in the requirements, return.
	// FIXME!! remove check - always false.
	// Don't use Hashmap here?
	if( restrictions.containsKey( type ) ) {
	    valid = true;	// trivially.
	} else if( required_type.isAssignableFrom( type ) ) {
	    valid = true;	// A subclass of the last required type.
	    // We need only remember this type now.
	    required_type = type;
	} else {
	    if( DEBUG ) Debug.printtagln( tag+".requireType",
					  "Invalid type restriction; "
					  + type.toString() +
					  " is not a subclass of " +
					  required_type.toString() );
	    valid = false;
	}

	return valid;
    }

    /**
     * Add a restriction of a field's value.
     * This method just looks up the named <code>Field</code> and calls
     * <code>requireValue( Field, Object )</code>.
     *
     * @param field_name	The name of the <code>Field</code> whose value
     * 				we are restricting.
     * @param value	The desired value.
     * @return <code>true</code> iff the requirement was registered without
     *		errors.
     */
    
    public boolean requireValue( String field_name, Object value )
    {
	if( DEBUG ) Debug.printtagln( tag+".requireValue", "called: " +
				      field_name + ", " + value );
	boolean valid = true;;
	// Find the <code>Field</code> object.
	Field field = null;
	try {
	    field = required_type.getField( field_name );
	} catch ( NoSuchFieldException e ) {
	    if( DEBUG ) Debug.printtagln( tag+".requireValue",
					  "No public field \"" + field_name +
					  "\" in " +required_type);
	    valid = false;
	}
	if( valid ) {
	    valid = requireValue( field, value );
	}
	return valid;
    }
    
    /**
     * Add a restriction of a field's value.
     * The appropriate <code>Field</code> of the events must
     * match the specified <code>value</code>, or it will not pass the filter.
     *
     * @param field	The <code>Field</code> whose value we are restricting.
     * 			It must be a member of the most recently specified
     * 			<code>Class</code>.    
     * @param value	The desired value.  It must be of the same type as the
     * 			specified <code>Field</code>, or the appropriate
     * 			wrapper if the <code>Field</code> is a primitive.  
     * 			Currently only simple types are allowed, such as
     * 			Integer, GUID, String, etc. 
     * @return <code>true</code> iff the requirement was registered without
     *		errors.
     */
    
    public boolean requireValue( Field field, Object value )
    {
	if( DEBUG ) Debug.printtagln( tag+".requireValue", "called: " +
				      field + ", " + value );
	boolean valid = true;
	Class field_type = field.getType();
	Class value_type = value.getClass();

	// Assume the restriction is valid, then perform a few checks:
	
	if( ! field.getDeclaringClass().isAssignableFrom( required_type ) ) {
	    // The field is not valid for the most recently required type.
	    if( DEBUG ) Debug.printtagln( tag+".requireValue",
					  field.toString() +
					  " is not a member of class "
					  + required_type.toString() );
	    valid = false;
	} else if( ! valid_types.containsKey( field_type ) ) {
	    // The field has an invalid type.
	    if( DEBUG ) Debug.printtagln( tag+".requireValue",
					  "Cannot restrict fields of type "
					  + field_type );
	    valid = false;
	} else if( ! field_type.equals(value_type) &&
		   ! field_type.equals(getPrimitiveType(value_type)) ) {
	    // The type of the value is neither the same as the field's type
	    // nor a wrapper for the field's (primitive) type.
	    if( DEBUG ) Debug.printtagln( tag+".requireValue",
					  "Field type does not match " +
					  "required value type: " +
					  field_type + " != " + value_type );
	    valid = false;
	} else if( restrictions.containsKey(field) &&
		   ! restrictions.get(field).equals(value) ) {
	    // We have conflicting restrictions for this field.
	    if( DEBUG ) Debug.printtagln( tag+".requireValue",
					  "Conflicting restriction for field "
					  + field_type + ": " + value + " != "
					  + restrictions.get(field) );
	    valid = false;
	} else {
	    // add this field/object pair to the set of restrictions.
	    restrictions.put( field, value );
	    ordered_restrictions.addLast( field );
	}

	return valid;
    }

    /**
     * Find the <code>Class</code> object for the primitive type corresponding
     * the the specified wrapper <code>Object</code>.
     *
     * @param wrapper	The <code>Class</code> object corresponding to a
     * 			primitive type's wrapper.
     * @return The <code>Class</code> object representing the proper primitive
     * type, or <code>null</code> if <code>wrapper</code> is not a primitive
     * wrapper.
     */

    private Class getPrimitiveType( Class wrapper )
    {
	Class primitive = (Class)valid_types.get( wrapper );
	// We've overloaded <code>valid_types</code> to map wrapper objects to
	// their primitive counterparts.  Use that.
	if( ( primitive != null ) &&
	    ( primitive.equals(Boolean.TRUE) ) ) {
	    primitive = null;
	}
	return primitive;
    }
     
    /**
     * Name the fraction of qualifying events which the
     * <code>Classifier</code> should pass through the <code>Filter</code>.
     *
     * @return <code>true</code> iff the rate was set without errors.
     */
    
    public boolean setAcceptanceRate( double rate )
    {
	if( DEBUG ) Debug.printtagln( tag+".setAcceptanceRate",
				      "called: " + rate );
	boolean valid = true;
	
	if( (0 <= rate) && (rate <= 1) ) {
	    acceptance_rate = rate;
	} else {
	    if( DEBUG ) Debug.printtagln( tag+".setAcceptanceRate",
					  "rate must be between 0 and 1." );
	    valid = false;
	}
	return valid;
    }

    public String toString () {
	return "ostore.dispatch.Filter";
    }
}
