/**
 * Resender.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.dispatch;

import ostore.util.Assert;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.NodeId;
import ostore.util.Clock;
import mdw.sandStorm.api.QueueElementIF;
import mdw.sandStorm.api.SinkFullException;
import java.util.Iterator;
import java.util.TreeSet;
import java.util.HashSet;

/**
 * A utility class that automatically re-dispatches events.
 * <p>
 * If an attempt to <code>dispatch</code> an event throws a
 * <code>SinkFullException</code>, the <code>Resender</code> schedules the
 * event for a later attempt and returns a unique, opaque token.  That
 * token can be used later to confirm or cancel delivery.
 * <p>
 * <b>NOTE:</b> A <code>Resender</code> does not guarantee any delivery
 * order for events which are initially delayed.  If in-order delivery is
 * important, do not use the automatic re-dispatching features, or avoid
 * this class completely.
 * <p>
 * Any stage which uses a <code>Resender</code> must
 * {@link Signal#register} for {@link Signal#ALRM} signals targetted at the
 * name given to the <code>Resender</code> constructor.  The
 * <code>Resender</code> will set its own alarms; the stage must catch the
 * resulting <code>AlarmSignal</code>s and call {@link #process_queues}.
 *
 * @author	Dennis Geels
 * @version $Id: Resender.java,v 1.2 2002/02/15 20:41:19 geels Exp $
 */

public class Resender
{
    ////////////////////       CONSTANTS	////////////////////

    private static final String class_tag = "Resender";
    private static final boolean DEBUG = DebugFlags.DISPATCH_RESENDER;
    
    ////////////////////    STATIC VARIABLES 	////////////////////
    ////////////////////       VARIABLES	////////////////////

    /**
     * The set of all events to retry at a later date, sorted by that date.
     */
    private TreeSet pending_events;

    /**
     * The set of all events which have expired.  They were retried their
     * maximum number of times (at least once) and never dispatched
     * successfully.
     */
    private HashSet waste_bin;

    /**
     * The <code>NodeId</code> for this machine.
     */
    private NodeId node_id;

    /**
     * The name with which to target alarm signals.
     */
    private String stage_name;
    
    /**
     * The underlying <code>Classifier</code> object through which to
     * dispatch events.
     */
    private Classifier classifier;

    /**
     * The <code>Signal</code> object for this machine.
     */
    private Signal signal;

    /**
     * The earliest deadline of any pending event, or
     * <code>Long.MAX_VALUE</code> if none pending.
     */
    private long earliest_deadline;

    /**
     * The token for the <code>AlarmSignal</code> we expect to arrive at
     * <code>earliest_deadline</code>. 
     */
    private Object alarm_token;

    /**
     * For efficiency, we cache one state object here.  We create a new one
     * only when the old one is returned to the caller, which binds it to a
     * single event.  That case happens only when a dispatch fails.
     */
    private EventState spare_state;
    
    ////////////////////  STATIC CONSTRUCTORS	////////////////////
    ////////////////////      CONSTRUCTORS	////////////////////

    /**
     * Creates a new <code>Resender</code> object that uses the
     * default <code>Classifier</code> for the specified machine.
     *
     * @param node_id	The ID for this machine - needed to locate the
     * correct <code>Classifier</code> and <code>Signal</code>.
     * @param stage_name	The name of the calling stage - needed to
     * address <code>AlarmSignal</code>s properly.
     */
    public Resender( NodeId node_id, String stage_name )
    {
	this.pending_events = new TreeSet();
	this.waste_bin = new HashSet();
	this.node_id = node_id;
	this.stage_name = stage_name;
	this.classifier = Classifier.getClassifier( node_id );
	this.signal = Signal.getSignal( node_id );
	this.earliest_deadline = Long.MAX_VALUE;
	this.alarm_token = null;
	this.spare_state = new EventState();
    }

    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Processes the queue of pending events, sending any whose deadlines
     * have been reaced.
     * <p>
     * This method should be called each time the stage receives an
     * <code>AlarmSignal</code>. 
     */
    public synchronized void process_queues()
    {
	final String tag = class_tag+".process_queues";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	// Scan the head of the queue until we reach an event which isn't
	// ready yet.
	long now = Clock.current_date( node_id ).getTime();
	if( DEBUG ) Debug.printtagln( tag, "now: " + now + " pending: " +
					  pending_events.size() );

	if( now >= earliest_deadline ) {
	    while( ! pending_events.isEmpty() ) {
		EventState next = (EventState)pending_events.first();

		if( DEBUG ) Debug.printtagln( tag, "next: " + next.deadline );
		if( next.deadline > now ) {
		    // No more events ready yet - all done.
		    earliest_deadline = next.deadline;
		    reset_alarm();
		    break;
		} else {
		    pending_events.remove( next );
		    // Try to dispatch the event
		    Object token =
			dispatch_internal( next.event, next.num_retries,
					   next.delay, next );
		    // <code>dispatch_internal</code> will put the event back
		    // on the queue if necessary.
		}
	    }
	    if( pending_events.isEmpty() ) {
		earliest_deadline = Long.MAX_VALUE;
	    }
	} else {
	    if( DEBUG ) Debug.printtagln( tag, "not ready" );
	    reset_alarm();
	}
	
	if( DEBUG ) Debug.printtagln( tag, "done" );
    }

    /**
     * Sets an alarm to wake us before the next event must be sent,
     * clearing any old alarm to save resources. 
     */
    private void reset_alarm()
    {
	final String tag = class_tag+".reset_alarm";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	if( alarm_token != null ) {
	    signal.cancel( alarm_token );
	}
	if( DEBUG ) Debug.printtagln( tag, "earliest_deadline: " +
				      earliest_deadline );
	if( earliest_deadline < Long.MAX_VALUE ) {
	    long now = Clock.current_date( node_id ).getTime();
	    alarm_token = signal.alarm( stage_name, (earliest_deadline-now) );
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );
    }
    
    /**
     * Dispatches an event through the <code>Classifier</code>.
     * <p>
     * If the dispatch fails, the event is stored in a trash pile.  Events
     * on the trash pile can be recovered by the {@link #clear} method.
     *
     * @param event	Any event to dispatch
     *
     * @return a token, used to reference the event later, or
     * <code>null</code> if the object was successfully dispatched.
     *
     * @see #dispatch(mdw.sandStorm.api.QueueElementIF, int, long)
     * @see Classifier#dispatch
     */
    public synchronized Object dispatch( QueueElementIF event )
    {
	// Do not retry the dispatch.
	return dispatch( event, 0, 0 );
    }
    
    /**
     * Dispatch the specified event, making at most
     * <code>num_retries</code> additional attempts, waiting
     * <code>millis</code> milliseconds in between attempts.
     * <p>
     * After enough attempts have failed, the event is stored in a trash
     * pile.  Events on the trash pile can be recovered by the
     * {@link #clear} method.
     *
     * @param event	Any event to dispatch
     * @param num_retries	The number (<code>>=0) of times to try
     * dispatching the event if the initial attempt fails.
     * @param millis	The number of milliseconds to wait after each
     * attempt.
     *
     * @return a token, used to reference the event later, or
     * <code>null</code> if the object was successfully dispatched on the
     * first try.
     *
     * @see #dispatch(mdw.sandStorm.api.QueueElementIF)
     * @see Classifier#dispatch
     */
    public synchronized Object dispatch( QueueElementIF event,
					 int num_retries, long millis )
    {
	final String tag = class_tag+".dispatch";
	if( DEBUG ) Debug.printtagln( tag, "called" );
	// Set a few of the state fields only once per event.
	spare_state.event = event;
	spare_state.delay = millis;
	// Now call the real method.
	Object token = dispatch_internal( event, num_retries, millis,
					  spare_state );
	if( token != null ) {	// implies <code>token == spare_state</code>
	    if( DEBUG ) Assert.assert( (token == spare_state),
				       "unexpected token" );
	    // Create a new state object for next time.
	    spare_state = new EventState();
	} 	// else reuse spare_state.
	
	return token;
    }

    /**
     * Performs the operations described in {@link
     * #dispatch(mdw.sandStorm.api.QueueElementIF, int, long)}, 
     * using the specified <code>EventState</code> object if the dispatch
     * fails. 
     *
     * @param event		Any event to dispatch
     * @param num_retries	The number (<code>>=0) of times to try
     * dispatching the event if the initial attempt fails.
     * @param millis		The number of milliseconds to wait after
     * 				each attempt.
     * @param state		The <code>EventState</code> object to use
     * 				to store the event if dispatch fails.
     *
     * @return <code>state</code> if <code>Classifier.dispatch</code>
     * throws an exception, or <code>null</code> if it worked.
     *
     * @see #dispatch(mdw.sandStorm.api.QueueElementIF, int, long)     
     */
    private Object dispatch_internal( QueueElementIF event, int num_retries,
				      long millis, EventState state )
    {
	final String tag = class_tag+".dispatch_internal";
	if( DEBUG ) Debug.printtagln( tag, "called: " + event + "(" +
				      num_retries + ")(" + millis + ")" );
	Object token;
	try {
	    classifier.dispatch( event );
	    if( DEBUG ) Debug.printtagln( tag, "dispatch succeeded" );	    
	    token = null;	// <code>null</code> means successful
	} catch( SinkFullException sfe ) {
	    if( DEBUG ) Debug.printtagln( tag, "dispatch failed" );
	    // Dispatch failed.  Store the failed event in an
	    // <code>EventState</code> object and pass it back to
	    // the caller as the event's unique token.
	    if( num_retries > 0 ) {
		long now = Clock.current_date( node_id ).getTime();
		long next_time = now + millis;
		if( DEBUG ) Debug.printtagln( tag, "rescheduled for " +
					      next_time );
		if( next_time < earliest_deadline ) {
		    earliest_deadline = next_time;
		    reset_alarm();
		}
		state.deadline = next_time;
		state.num_retries = num_retries-1;
		pending_events.add( state );
	    } else {
		// Don't worry about the other state fields.
		if( DEBUG ) Debug.printtagln( tag, "throwing away" );
		waste_bin.add( state );
	    }
	    token = state;	// non-<code>null</code> means failure.
	}
	
	if( DEBUG ) Debug.printtagln( tag, "done" );
	return token;
    }

    /**
     * Cancels the specified event.
     * All state relating to the pending event is removed from the
     * <code>Resender</code>.  If the event has already dispatched
     * successfully, this method does nothing. 
     *
     * @param token	A token returned by {@link #dispatch}.
     *
     * @return <code>true</code> iff the event has not yet been dispatched
     * successfully. 
     *
     * @see #dispatch(mdw.sandStorm.api.QueueElementIF, int, long)
     */
    public synchronized boolean cancel( Object token )
    {
	final String tag = class_tag+".cancel";
	if( DEBUG ) Debug.printtagln( tag, "called: " + token );

	boolean still_here = pending_events.remove( token );
	if( ! still_here ) {
	    still_here = waste_bin.remove( token );
	}
	if( still_here ) {
	    if( DEBUG ) Debug.printtagln( tag, "removed" );
	    EventState state = (EventState)token;
	    if( earliest_deadline == state.deadline ) {
		// Reschedule the next alarm.
		if( ! pending_events.isEmpty() ) {
		    state = (EventState)pending_events.first();
		    earliest_deadline = state.deadline;
		} else {
		    earliest_deadline = Long.MAX_VALUE;
		}
		reset_alarm();
	    }
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );
	return still_here;
    }

    /**
     * Checks whether the specified event was been successfully dispatched.
     *
     * @param token	A token returned by {@link #dispatch}.
     *
     * @return <code>true</code> iff the event has timed out.
     *
     * @see #dispatch(mdw.sandStorm.api.QueueElementIF, int, long)
     */
    public synchronized boolean timed_out( Object token )
    {
	final String tag = class_tag+".timed_out";
	if( DEBUG ) Debug.printtagln( tag, "called: " + token );

	boolean dead = waste_bin.contains( token );
	
	if( DEBUG ) Debug.printtagln( tag, "done: " + dead );
	return dead;
    }

    /**
     * Checks whether the specified event is scheduled for further attempts.
     *
     * @param token	A token returned by {@link #dispatch}.
     *
     * @return <code>true</code> iff the event is on the queue for another
     * attempted dispatch.
     *
     * @see #dispatch(mdw.sandStorm.api.QueueElementIF, int, long)
     */
    public synchronized boolean pending( Object token )
    {
	final String tag = class_tag+".pending";
	if( DEBUG ) Debug.printtagln( tag, "called: " + token );

	boolean waiting = pending_events.contains( token );
	
	if( DEBUG ) Debug.printtagln( tag, "done: " + waiting );
	return waiting;
    }

    /**
     * Checks whether the specified event has been successfully dispatched.
     * <p>
     * Equivalent to <code>( !pending(token) && !timed_out(token) )</code>.
     *
     * @param token	A token returned by {@link #dispatch}.
     *
     * @return <code>true</code> iff the <code>Resender</code> no longer
     * has any memory of the <code>token</code>, which <emph>implies</emph>
     * that the event was previously dispatched.
     *
     * @see #dispatch(mdw.sandStorm.api.QueueElementIF, int, long)
     */
    public synchronized boolean was_sent( Object token )
    {
	final String tag = class_tag+".was_sent";
	if( DEBUG ) Debug.printtagln( tag, "called: " + token );

	boolean still_here = ( pending_events.contains( token ) ||
			       waste_bin.contains( token ) );
	if( DEBUG ) {
	    Debug.printtagln( tag, (still_here?"not sent":"sent") );
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );	
	return ( ! still_here );
    }

    /**
     * Cancels all events which have timed out.  If <code>all ==
     * true</code>, also cancels all pending events.
     *
     * @param all	<code>true</code> iff pending events should also be
     * 			cancelled. 
     *
     * @return an array containing all cancelled events.
     *
     * @see #dispatch(mdw.sandStorm.api.QueueElementIF, int, long)
     * @see #cancel
     */
    public synchronized QueueElementIF[] clear( boolean all )
    {
	final String tag = class_tag+".was_sent";
	if( DEBUG ) Debug.printtagln( tag, "clear" );

	// Make an array to hold the dead events
	int len = waste_bin.size();
	if( all ) {	// Include the pending events.
	    len += pending_events.size();
	}
	QueueElementIF[] cancelled = new QueueElementIF[len];

	// Dump events from the waste bin into the array.
	int i = 0;
	Iterator states = waste_bin.iterator();
	while( states.hasNext() ) {
	    cancelled[i] = ((EventState)states.next()).event;
	    i++;
	}
	waste_bin.clear();
	if( all ) {
	    // Repeat for pending events.
	    states = pending_events.iterator();
	    while( states.hasNext() ) {
		cancelled[i] = ((EventState)states.next()).event;
		i++;
	    }	    
	    pending_events.clear();

	    // No events pending:
	    earliest_deadline = Long.MAX_VALUE;
	    reset_alarm();
	}

	if( DEBUG ) Debug.printtagln( tag, "done" );
	return cancelled;
    }
       
    
    ////////////////////     INNER CLASSES	////////////////////

    /**
     * An internal object that stores state for an event which did not
     * dispatch properly.
     */
    private class EventState implements Comparable
    {
	////////////////////     CONSTANTS	////////////////////
	////////////////////  STATIC VARIABLES 	////////////////////
	////////////////////     VARIABLES	////////////////////

	/**
	 * The original event.
	 */
	public QueueElementIF event;

	/**
	 * The number of times to try to dispatch <code>event</code> if the
	 * next attempt fails.
	 */
	public int num_retries;
	
	/**
	 * The time at which to try next.
	 */
	public long deadline;

	/**
	 * The number of milliseconds to wait before trying again.
	 */
	public long delay;
	
	////////////////////    CONSTRUCTORS	////////////////////
	////////////////////   STATIC METHODS 	////////////////////
	//////////////////// 	  METHODS	////////////////////

	/**
	 * Impose an order on <code>EventState</code>s based on their
	 * <code.deadline</code>s.
	 */
	public int compareTo( Object o )
	{
	    EventState other = (EventState)o;
	    if( deadline < other.deadline ) {
		return -1;
	    } else if( deadline > other.deadline ) {
		return 1;
	    } else {
		return 0;
	    }
	}
	/**
	 * Returns a human-readable representation of this
	 * <code>EventState</code>
	 */
	public String toString()
	{
	    return ("<Token for: " + event + ">");
	}
	
	////////////////////   INNER CLASSES	////////////////////
    }
}
