/**
 * TrieRoot.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.dispatch;

import java.util.Vector;

/**
 * The root of a <code>Classifier</code> trie.
 * Each trie corresponds to a single event type.  The <code>TrieRoot</code>
 * contains links to the <code>Class</code> objects for the neighbors of
 * this trie's type in the class hierarchy. 
 *
 * @author	Dennis Geels
 * @version	$Id: TrieRoot.java,v 1.5 2002/02/09 01:22:45 srhea Exp $
 */

public class TrieRoot extends TrieNode
{
    ////////////////////       CONSTANTS	////////////////////
    ////////////////////    STATIC VARIABLES 	////////////////////
    ////////////////////       VARIABLES	////////////////////

    /**
     * The parent of this node's type in the event class hierarchy.
     */
    protected Class superclass;

    /**
     * The children of this node's type in the event class hierarchy.
     */
    protected Vector subclasses;
    
    ////////////////////  STATIC CONSTRUCTORS	////////////////////
    ////////////////////      CONSTRUCTORS	////////////////////

    /**
     * Creates a new <code>TrieRoot</code> with empty
     * <code>superclass</code> and <code>subclasses</code> fields.
     */
    protected TrieRoot() {
	super();
	superclass = null;
	subclasses = new Vector();
    }

    /**
     * Creates a new <code>TrieRoot</code> with the specified fields.  This
     * constructor is used within <code>clone</code>.
     */
    private TrieRoot( Vector sinks, Vector priorities, Vector rates,
		      Vector fields, Vector value_maps,
		      Class superclass, Vector subclasses )
    {
	this.sinks = sinks;
	this.priorities = priorities;
	this.rates = rates;
	this.fields = fields;
	this.value_maps = value_maps;
	this.superclass = superclass;
	this.subclasses = subclasses;
    }

    
	
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Performs a deeper copy than <code>Object.clone</code> would.
     * The entire trie is cloned, and may be modified independently of the
     * original.  The <code>Field</code>s and <code>SinkIF</code>s
     * referenced by the <code>TrieNode</code>s are not cloned.
     *
     * @return A new <code>TrieRoot</code> with <code>cloned</code>
     * internal data structures.
     */
    public Object clone() {
  	TrieNode stub = (TrieNode)super.clone();
  	Vector subclasses_clone = (Vector)subclasses.clone();
  	TrieRoot cloned = new TrieRoot( stub.sinks, stub.priorities,
					stub.rates, stub.fields,
					stub.value_maps, superclass,
					subclasses_clone );
	return cloned;
    }
    ////////////////////     INNER CLASSES	////////////////////

    public String toString () {
	return "ostore.dispatch.TrieRoot";
    }

}
