/**
 * HMACDHHandler.
 *
 * @author    Gifford Cheung
 * @version    $Id: HMACDHHandler.java,v 1.21 2002/04/05 19:52:04 gkcheung Exp $
 */  

/**
 * Future TODO: implement message/event logging....?
 */

package ostore.mac;
import java.util.*;
import java.io.*;
import ostore.util.*;
import ostore.dispatch.*;
import ostore.security.*;
import mdw.sandStorm.api.*;
import mdw.sandStorm.core.ssTimer;
import mdw.sandStorm.internal.SinkProxy;

// Security stuff
import java.security.*;
import java.security.spec.*;
import javax.crypto.*;
import javax.crypto.spec.*;
import javax.crypto.interfaces.*;
import cryptix.jce.provider.*;
import cryptix.jce.provider.random.*;
import cryptix.jce.provider.key.*;

/**
 * MacHandler that uses HMAC for the message authentication and  
 * Diffie-Hellman Key Generation.  <br></br><br></br>
 * <b><i>Station-To-Station Protocol Implentation:</i></b> <br></br><br></br>
 * <b>Basic Protocol:</b><br></br>
 * There are four states: (1)UNINITIALIZED, (2)WAITFORRESPONSE, (3)WAITFORFINAL,
 * (4)ESTABLISHED.<br></br>
 * There are three messages (across the network): (1)INITIALIZE, (2)RESPOND,
 * (3)FINALIZE. <br></br>
 * There are three events (internal): (1)INIT (request to send out message 1), 
 * (2) STS_TIMEOUT, (3)DH_TIMEOUT <br></br>
 * State table<br></br>
 * <table border = 1><tr><td><b>State</b></td>
 * <td><b>Event</b></td><td><b>Next Stage</b>
 * </td><td><b>Action</b></td></tr>
 * <tr><td>UNINIT</td><td>Tapestry sends (1)INIT event</td>
 *  <td>WAITFORRESPONSE</td><td>Send message (1)</td></tr>
 * <tr><td>UNINIT</td><td>Receive message (1)</td>
 *  <td>WAITFORFINAL</td><td>Compute secret, Send message (2)</td></tr>
 * <tr><td>WAITFORRESPONSE</td><td>Receive message (2)</td>
 *  <td>ESTABLISHED</td><td>Confirm signature and Compute secret, 
 *  Send message (3)</td></tr>
 * <tr><td>WAITFORRESPONSE</td><td>Receive message (1)</td>
 *  <td>WAITFORRESPONSE or <br></br> WAITFORFINAL</td><td>If my guid is 
 *    greater that their guid, I become WAITFORFINAL, else ignore message.
 *  </td></tr>
 * <tr><td>WAITFORFINAL</td><td>Receive message (3)</td>
 *  <td>ESTABLISHED</td><td>Confirm signature</td></tr></table>
 * <br></br><br></br>
 * 
 * <b>establishedParties & pendingParties:</b><br></br>
 * Needed: To be able to go through a false or faulty handshake, 
 * one that is cancelled before completion, and not lose the existing
 * connection. <br></br>
 * Solution: Keep two Hashtables (indexed by guid) one for established
 * connections (<tt>establishedParties</tt>) and another for in-progress
 * connections (<tt>pendingParties</tt>). If a <tt>pendingParties</tt> 
 * connection successfully finishes a handshake, that entry replaces the
 * one in <tt>establishedParties</tt>. <br></br>
 * Functional access to this MacHandler only accesses 
 * <tt>establishedParties</tt><br></br><br></br>
 * <b>Timeouts:</b><br></br>
 * The HMACDHHandler handles timeouts (both session and Diffie-Hellman key
 * expirations) by sending itself delayed messages. The time (in milliseconds)
 * of these are determined by config variables: <tt>timeoutlength</tt> and 
 * <tt>dhlifespan</tt><br></br><br></br>
 * TODO describe sts timeout and dh timeout<br></br>
 * <b>Synchronicity:</b><br></br>
 * Internally: <tt>establishedParties</tt>I
 * This stage is not designed to be replicated. Replication may introduce
 * race conditions that have not been thoroughly investigated.
 */
public class HMACDHHandler extends MacHandler {

    public static final String tag = "HMACDHHandler";
    public static final boolean DEBUG       = DebugFlags.HMAC_DH_HANDLER;
    public static final boolean ANNOUNCE       
                                = DebugFlags.HMAC_DH_HANDLER_ANNOUNCEMENTS;
    private static final int _keysize       = 1024; //4096;
    private static HMACDHHandler oneAndOnly = null; 
    
    private Hashtable establishedParties = new Hashtable();
    private Hashtable pendingParties = new Hashtable();
    private KeyPairGenerator _kpgen;
    private KeyAgreement _kagree;
    private SecureRandom _r;
    private Signature _signEngine;
    private String _pkeyfile;
    private String _skeyfile;
    private QSPublicKey _sigPub;
    private QSPrivateKey _sigPri;
    private Mac _hmac;
    private SecureHash _myGuid;
    private SinkIF _this_sink;
    private int _stsTime;
    private int _retryLimit;
    private int _dhlifespan;
    private boolean _selfTest;
    private Classifier _classifier;

    public void handleEvent(QueueElementIF item)
        throws EventHandlerException {
        if (ANNOUNCE) Debug.printtagln(tag,".");
        if (DEBUG) {
            Debug.printtagln(tag, "[~.~GOT EVENT--hmacdhhandler--" + _myGuid +
                           " ~.~]");
        }
    
        if (item instanceof TellSelf) {
            TellSelf tell = (TellSelf) item;
            if (tell.msg == TellSelf.INIT) {
			  if (DEBUG) {
				Debug.printtagln(tag, "HMDHH " + _myGuid + 
								 ": TellSelf.init recv'd");
				Debug.printtagln(tag, "HMDHH " + _myGuid + 
								 ": printing pending");
				for (Enumeration e = pendingParties.elements(); 
					 e.hasMoreElements(); ) {
				  ((PartyInfo)e.nextElement()).print();
				  Debug.printtagln(tag, "---------------------");
				}
				Debug.printtagln(tag, "***************************");
				Debug.printtagln(tag, "HMDHH " + _myGuid + 
								 ": printing established");
				for (Enumeration e = establishedParties.keys(); 
					 e.hasMoreElements(); ) {
				  Debug.printtagln(tag, _myGuid + 
								   e.nextElement().toString ());
				  Debug.printtagln(tag, "---------------------");
				}
				for (Enumeration e = establishedParties.elements(); 
					 e.hasMoreElements(); ) {
				  if (DEBUG)
					((PartyInfo)e.nextElement()).print();
				  Debug.printtagln(tag, "---------------------");
				}
			  }
                // if establishedParties(guid) gone then nothing establishED
                // if pendingParties(guid) gone then nothing establishING
                // if nothing is establishED or establishING for this guid,
                // then get something started.

                // if we are already establishED: reannounce and return.
                if (establishedParties.containsKey(tell.guid)) {
                    QueueElementIF reannounce 
                        = new KeyEstablished( tell.guid);
                    try {
                        _classifier.dispatch( reannounce);
                        if (DEBUG) Debug.printtagln(tag, "HMAC" + _myGuid + 
                                        ": REANNOUNCING.");
                    } catch (SinkException e) {
                        // TODO what if this is a SinkFullException ... ?
                        Carp.die( "ostore.mac.HMACDHHandler: sink exception.");
                    }
                    return;
                }
                // if we are establishING: return and do nothing
                if ( pendingParties.containsKey(tell.guid) ) 
                    return; 

                // create a new PartyInfo and place into pendingParties 
                PartyInfo freshEntry = new PartyInfo();
                pendingParties.put(tell.guid, freshEntry);

                // generate new DH pair 
                // this ISN'T synchronized, even if it alters a pendingParties
                // element because  (1) this is a slow operation
                //                  (2) this particular PartyInfo will NOT
                //                      be next altered UNTIL a new DHMessage
                //                      comes in, referring to it and that won't
                //                      come in until I send out the 1st 
                //                      message, I haven't till fifteen lines 
                //                      from now.
                makeNewDHKeys( freshEntry);
                
                // make INITIALIZE message
                DHMessage out = new DHMessage(tell.guid,
                                            freshEntry.dhpub,
                                            null,
                                            new QSHMACMD5(),
                                            null, null,
                                            DHMessage.INITIALIZE);
                out.inbound = false;

                // change state to WAITFORRESPONSE
                freshEntry.state = freshEntry.WAITFORRESPONSE;

                // send message INITIALIZE and
                // send out TIMER for session timeout
                //TellSelf st = new TellSelf( tell.guid, TellSelf.STS_TIMEOUT);
                TellSelf st = new TellSelf( tell.guid, TellSelf.STS_TIMEOUT);
                if (DEBUG) Debug.printtagln(tag, "HMDHH" + _myGuid + 
                                ": Sending init msgs...");
                try {
                    _classifier.dispatch(out);
                    freshEntry.timerToken = 
                        _classifier.dispatch_later(st, _stsTime);
                } catch (SinkException e) {
                    // TODO what if this is a SinkFullException ... ?
                    Carp.die( "ostore.mac.HMACDHHandler: sink exception.");
                }
                    
            } else if (tell.msg == TellSelf.STS_TIMEOUT) {
                try {
                    if (ANNOUNCE) 
                        Debug.printtagln(tag, "HMDHH" + _myGuid +
                                        ":TellSelf.STS_TIMEOUT rcv'd"
                                           + "against: " + tell.guid);
                    QueueElementIF announce;

                    // *******************************************
                    // IT IS STILL POSSIBLE FOR A TIMEOUT TO QUEUE  
                    // UP EVEN THOUGH WE ARE ESTABLISHED
                    // ******************************************

                    // a session has timed out
                    // (1) retrieve the session
                    PartyInfo info = (PartyInfo) pendingParties.get( tell.guid);
                    // (2) do nothing if we retrieved nothing
                    //     reason for this typically is
                    //     a timeout was placed in the queue before
                    //     a successful STS run. 
                    //     If this isn't the reason, we still would be
                    //     able to do anything
                    if (info == null) {
                            // do nothing
                            // this statement skips steps 3/4  
                            if (ANNOUNCE) Debug.printtagln(tag,"HMACDHH" + 
                                            _myGuid + ": " +
                                           "TIMEOUT received, but no info. " +
                                           "may have already established.");
                    } else {
                        
                        // (3) increment the retry counter
                        info.sessionRetries += 1;
                        // (4) if retry limit not reached start all over again. 
                        if (info.sessionRetries >= _retryLimit) {
                            announce = new HandshakeFailed( tell.guid);
                        } else {
                            announce = new DHMessage( tell.guid,
                                                      info.dhpub,
                                                      null,
                                                      new QSHMACMD5(),
                                                      null, null,
                                                      DHMessage.INITIALIZE);
                            ((DHMessage) announce).inbound = false;
    
                            info.state = info.WAITFORRESPONSE;
                            TellSelf st = 
                                new TellSelf( tell.guid, TellSelf.STS_TIMEOUT);
                                info.timerToken = 
                                _classifier.dispatch_later(st, 
                                _stsTime);
                                //JUMP SinkException
                        }

                        _classifier.dispatch(announce);
                        // JUMP SinkException
                    }
                } catch (SinkException e) {
                        // TODO what if this is a SinkFullException ... ?
                    Carp.die( "ostore.mac.HMACDHHandler: sink exception.");
                }
                    
            } else if (tell.msg == TellSelf.DH_TIMEOUT) {
                if (ANNOUNCE) Debug.printtagln(tag, "DH TIMEOUT:" + _myGuid);
                // (1)snag lock 
                    /* NOTE on establishedParties
                     * establishingParties is accessed both by handleEvent and 
                     * functional access to this stage, thus the need for
                     * "synchronized" access to it.
                     */
                synchronized (establishedParties) {
                    // (2)wipe all keys and timers
                    for (Enumeration e = pendingParties.elements(); 
                         e.hasMoreElements() ;) {
                        PartyInfo i = (PartyInfo) e.nextElement();
                        if (i.timerToken != null) {
                        _classifier.cancel(i.timerToken);
                        }
                    }
                    establishedParties.clear();
                    // (3) usedKeys cleared out by virtue of emptying 
                    // establishedParties. 
                }
                _classifier.dispatch_later(tell,_dhlifespan);
               }
        } else if (item instanceof DHMessage) {
            //=============================================
            // Prep dhmsg for Stage evaluation
            //=============================================
            DHMessage dhmsg = (DHMessage) item;
            if (DEBUG) {
                Debug.printtagln(tag, "HMDHH" + _myGuid);
                dhmsg.print();
            }

            // Condition(4.b <Prep>) and to send out message: extract guid
            SecureHash otherguid = dhmsg.peer;

            // Condition (1) record msgType
            int msgType = dhmsg.messageType;

            // Condition (2) grab and verify the MACType
            // drop the message if MACType not identified
            if (dhmsg.MACType == null) return;
            // drop the message if MACType incorrect (not MACTYPE_HMACMD5)
            if (dhmsg.MACType.type_code() != Types.MACTYPE_HMACMD5) return;

            // Condition (3) get associated PartyInfo
            PartyInfo info;
            if (!pendingParties.containsKey( otherguid)) {
                if (msgType == DHMessage.INITIALIZE) {
                    // if the PartyInfo entry doesn't exist and we have an 
                    // appropriate message, create it.
                    PartyInfo newPending = new PartyInfo();
                    pendingParties.put(otherguid, newPending);
                    info = newPending;
                } else return; // no partyinfo entry means nothing can happen
            } else {
                info = (PartyInfo) pendingParties.get( otherguid);
            }
            if (DEBUG) {
                Debug.printtagln(tag, "HMDHH" + _myGuid);
                info.print();
            }

            //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- 
            // Conditions satisfied at this point:
            // (1) msgType retreived 
            // (2) MACType retreived and == Types.MACTYPE_HMACMD5  
            // (3) info is a non-null PartyInfo entry linked w/ the sender guid
            // Conditions needed to be satisfied:
            // (4) the rest of dhmsg need to be checked (at least non-null) 
            //        according to the state:
            //        (a)QSkey, (b)sigkey, (c)verify
            //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-    

            // State evaluation: get the state 
            int state = info.state;
            // =================================================
            if (state == PartyInfo.UNINIT) {
            // ----------------------------------STAGE UNINIT
             try {
                if (msgType == DHMessage.INITIALIZE) {
                    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
                    // Conditions:
                    // Condition (4.abc)
                    // an INITIALIZE message needs a non-null entry
                    // for (a)QSkey and nothing in (b)sigkey or (c)verify
                    // if we catch something wrong here, we drop the message
                    // and reset the pendingParties 
                    // TODO: this still allows for Denial of Service attacks...
                    if ( dhmsg.QSkey == null || 
                         dhmsg.sigkey != null ||    
                         dhmsg.mac != null) {
                        // ignore (TODO write error to log file)
                        pendingParties.remove( otherguid);
                        return;
                    }
                    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

                    // (1)
                    // TODO a check for a bad key submission
                    //        one that takes advantage of the 
                    //        math involved such as a QS public key
                    //        of value 0 and for replay attacks (keysused)
                    if (establishedParties.containsKey( otherguid)) {
                        PartyInfo check = 
                            (PartyInfo) establishedParties.get( otherguid);
                        if (check.usedKeys.contains(dhmsg.QSkey)) {
                            // then we've already seen this DH key
                            // TODO write event to log file and ignore msg
                            return;
                        }
                    }
                    // (1.a) store their DHKey. Also checks out we 'save it'.
                    info.dhpubOther = (DHPublicKey) dhmsg.QSkey.key();
                    info.usedKeys.add(dhmsg.QSkey);

                    // (2) do we need to compute a new DH?
                    if (establishedParties.containsKey( otherguid)) {
                        PartyInfo e = 
                          (PartyInfo) establishedParties.get(otherguid);
                        info.dhpub = e.dhpub;
                        info.dhpri = e.dhpri;
                    }
                    else {
                        makeNewDHKeys( info);
                    }
                    if (DEBUG) Debug.printtagln(tag, "HMACDHH" + _myGuid + 
                                    ": computed DH");

                    // (3) compute secret
                    _kagree.init(info.dhpri); //JUMP InvalidKeyException
                    _kagree.doPhase(dhmsg.QSkey.key(), true);
                    byte[] seed = _kagree.generateSecret();
                    info.secret = new RawSecretKey("DH seed", seed);
                    if (DEBUG) Debug.printtagln(tag, "HMACDHH" + _myGuid + 
                                    ": computed secret");
                    
                    // (4) make and send DHMESSAGE RESPOND
                    DHMessage out = new DHMessage(otherguid,
                                            info.dhpub,
                                            (QSRsaPublicKey)_sigPub,
                                            new QSHMACMD5(),
                                            null, null, // makeDHVerify later
                                            DHMessage.RESPOND);
                    out.inbound = false;

                    // compute verification info and place into DHVerify
                    //  MAC(Signature(Hash(both DH keys and Mac type)))
                    makeDHVerify( new QSDHPublicKey(info.dhpub), 
                                  dhmsg.QSkey, 
                                  info.secret, 
                                  out);
                    // JUMP SignatureException

                    // send out 
                    if (DEBUG) Debug.printtagln(tag, "HMACDHH: " + _myGuid +
                                                  "sending out RESPOND  msg");
                    _classifier.dispatch(out);
                    // JUMP SinkException

                    // (5) change state to WAITFORFINAL
                    info.state = PartyInfo.WAITFORFINAL;
                    
                    // (6) TIMER send out session timer
                    TellSelf st = new TellSelf(otherguid, TellSelf.STS_TIMEOUT);
                    info.timerToken =
                        _classifier.dispatch_later(st, _stsTime);

                }    
             } catch (InvalidKeyException e) {
                // key issue reset the stage and ignore message 
                // TODO write error to log file
                Carp.out( "ostore.mac.HMACDHHandler: InvalidKeyException.");
                info.state = PartyInfo.UNINIT;
                info.secret = null;
             } catch (SignatureException e) {
                // unable to sign the message
                // TODO write error to log file
                Carp.out( "ostore.mac.HMACDHHandler: signature exception.");
                info.state = PartyInfo.UNINIT;
                info.secret = null;
             } catch (SinkException e) {
                // Unable to dispatch the message
                // TODO write error to log file
                // TODO what if this is a SinkFullException ... ?
                Carp.out( "ostore.mac.HMACDHHandler: sink exception.");
                info.state = PartyInfo.UNINIT;
                info.secret = null;
             } 
            // =================================================
            } else if (state == PartyInfo.WAITFORRESPONSE) {
            // ----------------------------------STAGE WAITFORRESPONSE
             try {
                if (msgType == DHMessage.RESPOND) {
                    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
                    // Conditions:
                    // Condition (4.abc)
                    if ( dhmsg.QSkey == null ||
                         dhmsg.sigkey == null ||
                         !otherguid.verify(dhmsg.sigkey) ||
                         dhmsg.mac == null) {
                        // ignore (TODO write error to log file)
                        if (DEBUG) Debug.printtagln(tag, _myGuid + 
                                            "HMDHH: dhmsg failed " +
                                            "Condition 4.abc check");
                        return;
                    }
                    
                    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

                    // (1) Store the public signature key and QSkey
                    info.sigpub = dhmsg.sigkey;
                    info.dhpubOther = (DHPublicKey)dhmsg.QSkey.key();

                    // (2)
                    // TODO check for a bad key submission
                    //        one that takes advantage of the 
                    //        math involved such as a QS public key
                    //        of value 0 and for replay attacks (keysused)
                    // THEN check against usedKeys
                    if (establishedParties.containsKey( otherguid)) {
                        PartyInfo check = 
                            (PartyInfo) establishedParties.get( otherguid);
                        if (check.usedKeys.contains(dhmsg.QSkey)) {
                            // then we've already seen this DH key
                            // TODO write event to log file and ignore msg
                            return;
                        }
                    }
                    // (2.a) if their DHKey checks out we 'save it'.
                    info.usedKeys.add( dhmsg.QSkey);

                    // (3) compute secret
                    _kagree.init(info.dhpri); 
                        // JUMP InvalidKeyException
                    _kagree.doPhase(dhmsg.QSkey.key(), true); 
                        // JUMP InvalidKeyException
                    byte[] seed = _kagree.generateSecret();
                    info.secret = new RawSecretKey("DH seed", seed);
                    
                    // (4) check dhmsg.verify
                    // JUMP InvalidKeyException
                    // JUMP SignatureException
                    // JUMP TypeTable.NoSuchTypeCode
                    if (!checkDHVerify( dhmsg.user_data, dhmsg.mac,dhmsg,info)) {
                        // ignore (TODO write error to log file)
                        return;
                    }

                    // (5) make and send FINALIZE
                    DHMessage out = new DHMessage(otherguid,
                                                  null,
                                                  (QSRsaPublicKey)_sigPub,
                                                  new QSHMACMD5(),
                                                  null, null, // DHVerify later
                                                  DHMessage.FINALIZE);
                    out.inbound = false;

                    // compute verification info and place into DHVerify
                    //  MAC(Signature(Hash(both DH keys and Mac type)))
                    makeDHVerify( new QSDHPublicKey(info.dhpub), 
                                  dhmsg.QSkey, 
                                  info.secret, 
                                  out);
                    // JUMP SignatureException

                    // send out 
                    _classifier.dispatch(out);
                    // JUMP SinkException
                    
                    // (6) change state to ESTABLISHED
                    info.state = PartyInfo.ESTABLISHED;

                    if (ANNOUNCE) 
                        Debug.printtagln(tag, "HMACDHHandler: established by " +
                                            _myGuid + "(me) with " +
                                            otherguid + "(them)");
                               
                            

                    // (7) move info to establishedParties and
                    //       remove info from pendingParties 
                    //       cancel timer
                    synchronized (establishedParties) {
                        _classifier.cancel(info.timerToken);
                        info.timerToken = null;
                        if (establishedParties.contains( otherguid)) {
                            PartyInfo old =
                                (PartyInfo) establishedParties.get( otherguid);
                            info.usedKeys.addAll( old.usedKeys);
                        }
                        pendingParties.remove( otherguid);
                        establishedParties.put( otherguid, info);
                        QueueElementIF announce 
                            = new KeyEstablished( otherguid);
                        _classifier.dispatch( announce);
                            //JUMP SinkException
                    }

                    if (DEBUG) { 
                        Debug.printtagln(tag, "HMACDHHANDLER info for"+_myGuid);
                        info.print();
                    }
                } else if (msgType == DHMessage.INITIALIZE) {
                    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
                    // Conditions:
                    // Condition (4.abc)
                    if ( dhmsg.QSkey == null || 
                         dhmsg.sigkey != null || 
                         dhmsg.mac != null) {
                        // ignore (TODO write error to log file)
                        return;
                    }
                    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

                    // Conditions are met, so this is a readable message,
                    // Now, should we ignore this message? Because in this
                    // case, both sides sent (1) and are stage WAITFORRESPONSE
                    // If my guid is greater than other guid change my state.
                    int compare = _myGuid.compareTo(otherguid);
                    if (compare > 0) {
                        if (ANNOUNCE) {
                            Debug.printtagln(tag,"HMACDHH: DUAL-INIT conflict" +
                                               " I, " + _myGuid + 
                                               " WILL BECOME UNINIT FOR " +
                                               otherguid + " AND WILL SEND" +
                                               " MYSELF THE MESSAGE AGAIN");
                        }
                        info.state = PartyInfo.UNINIT;
                        // send the message back to myself
                        _classifier.dispatch(dhmsg);
                            //JUMP SinkFullException
                    }
                    if (compare == 0) 
                           Carp.die("Attempting to handshake with self!");
                    if (ANNOUNCE && compare <= 0) {
                        Debug.printtagln(tag,"HMACDHH: DUAL-INIT conflict" +
                                           " I, " + _myGuid + 
                                           " WILL IGNORE THIS MESSAGE FOR " +
                                           otherguid);
                    }
                }    
             } catch (InvalidKeyException e) { 
                // TODO write error to log file
                Carp.out( "ostore.mac.HMACDHHandler: invalid key exception.");
                info.state = PartyInfo.WAITFORRESPONSE;
                info.secret = null;
             } catch (SignatureException e) {
                // TODO write error to log file
                Carp.out( "ostore.mac.HMACDHHandler: signature exception.");
                info.state = PartyInfo.WAITFORRESPONSE;
                info.secret = null;
             } catch (TypeTable.NoSuchTypeCode e) {
                // TODO write error to log file
                Carp.out( "ostore.mac.HMACDHHandler:nosuchtypecode exception.");
                info.state = PartyInfo.WAITFORRESPONSE;
                info.secret = null;
             } catch (SinkException e) {
                // Unable to dispatch the message
                // TODO write error to log file
                // TODO what if this is a SinkFullException ... ?
                Carp.out( "ostore.mac.HMACDHHandler: sink exception.");
                info.state = PartyInfo.WAITFORRESPONSE;
                info.secret = null;
             } catch ( ArrayIndexOutOfBoundsException e) {
                // TODO write error to log ifle
                Carp.out( "ostore.mac.HMACDHHandler: array error copying old.");
             }
            // =================================================
            } else if (state == PartyInfo.WAITFORFINAL) {
            // ----------------------------------STAGE WAITFORFINAL
             try {
                if (msgType == DHMessage.FINALIZE) {
                    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
                    // Conditions:
                    // Condition (4.abc)
                    if ( dhmsg.QSkey != null ||
                         dhmsg.sigkey == null ||
                         !otherguid.verify( dhmsg.sigkey) ||
                         dhmsg.mac == null ) {
                        if (DEBUG) Debug.printtagln(tag, "HMACDHHandler: " +
                                                     "message ignored. :"  +
                                                     _myGuid);
                        // ignore. (TODO write error to log file)
                        return;    
                    }        
                    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

                    // (1) Store the public signature key
                    info.sigpub = dhmsg.sigkey;

                    // (2) check dhmsg.verify
                    if (!checkDHVerify( dhmsg.user_data, dhmsg.mac, dhmsg, info)){
                            //JUMP InvalidKeyException
                            //JUMP SignatureException
                            //JUMP TypeTable.NoSuchTypeCode
                        // ignore (TODO write error to log file)
                        if (DEBUG) Debug.printtagln(tag, "HMACDHHandler: " +
                                                     _myGuid +
                                                     "dhverify failed.");
			if (DEBUG) {
                        Debug.printtagln(tag, _myGuid.toString());
                        info.print();
                        return;
                    }
                    
                    // (3) change state to ESTABLISHED
                    info.state = PartyInfo.ESTABLISHED;

                    if (ANNOUNCE) 
                        Debug.printtagln(tag,"HMACDHHandler: established by " +
                                            _myGuid + "(me) with " +
                                            otherguid + "(them)");

                    // (4) move info to establishedParties and
                    //       remove info from pendingParties 
                    synchronized(establishedParties) {
                        _classifier.cancel(info.timerToken);
                        info.timerToken = null;
                        if (establishedParties.contains( otherguid)) {
                            PartyInfo old =
                                (PartyInfo) establishedParties.get( otherguid);
                            info.usedKeys.addAll( old.usedKeys);
                        }
                        pendingParties.remove( otherguid);
                        establishedParties.put( otherguid, info);
                        QueueElementIF announce 
                            = new KeyEstablished( otherguid);
                        _classifier.dispatch( announce);
                            //JUMP SinkException
                    
                        if (DEBUG)  
                            Debug.printtagln(tag, "HMACDHHANDLER info" + 
                                            _myGuid);
                            info.print();
                        }
                    }
                }    
             } catch (InvalidKeyException e) {
                // TODO write error to log file
                Carp.out( "ostore.mac.HMACDHHandler: invalid key exception.");
                info.state = PartyInfo.WAITFORFINAL;
             } catch (SignatureException e) {
                // TODO write error to log file
                Carp.out( "ostore.mac.HMACDHHandler: signature exception.");
             } catch ( TypeTable.NoSuchTypeCode e) {
                // TODO write error to log file
                Carp.out( "ostore.mac.HMACDHHandler:nosuchtypecode exception.");
             } catch ( ArrayIndexOutOfBoundsException e) {
                     // TODO write error to log ifle
                Carp.out( "ostore.mac.HMACDHHandler: array error copying old.");
             } catch ( SinkException e) {
                // TODO write error to log file 
                // TODO what if this is a SinkFullException ... ?
                Carp.out( "ostore.mac.HMACDHHandler: sink exception" + e);
             }
            }
            // =================================================
        }
    }

    public void handleEvents(QueueElementIF item[])
        throws EventHandlerException {

        for (int i = 0; i < item.length; i += 1)
            handleEvent(item[i]);
    }

    private boolean checkMAC( byte[] src, byte[] MAC, SecretKey macSecret) {
        try {
            _hmac.init(macSecret);
            return ByteUtils.equals(MAC, _hmac.doFinal(src));
        } catch (InvalidKeyException e) {
            Carp.die("HMACDHHandler: could not init hmac engine.");
            return false;
        }
    }

    private static byte [] data_to_check (QuickSerializable user_data) {
	byte [] result = new byte [4+ByteUtils.qs_length (user_data)];
	int [] offset = {0};
	ByteUtils.intToBytes (user_data.type_code (), result, offset);
	user_data.to_bytes (result, offset);
	return result;
    }

    private boolean checkDHVerify( QuickSerializable user_data,
				   QSByteArray mac,
                                   DHMessage dhmsg, 
                                   PartyInfo info) 
     throws InvalidKeyException, SignatureException, TypeTable.NoSuchTypeCode {
        boolean allzgood;
        SignedQS s = null;
        STSVerify recreatedSTS;
        byte[] recreated, extracted;

        // check MAC
        allzgood = checkMAC(data_to_check (user_data), 
		mac.bytes (), info.secret);

        if (allzgood) {
            // extract signature
            s = (SignedQS) user_data;
        
            // verify Signature
            // next line: workaround cryptix bug
            try {reInitSigEngine();} catch (Exception e) {Carp.die(e);}
            allzgood = s.verify(info.sigpub.key(), _signEngine);
        }
        if (DEBUG) Debug.printtagln(tag, "checkDH: sig. allzgood = " + 
                                        allzgood + " " + _myGuid);
            
        // check the HASH
        if (allzgood) {
            // recreate STS (as if they would have done, i.e.
            // "my Public + other Public + MACType"
            // becomes: 
            // "other Public + my Public + MACType"
            recreatedSTS = new STSVerify(new QSDHPublicKey(info.dhpubOther), 
                                         new QSDHPublicKey(info.dhpub), 
                                         Types.MACTYPE_HMACMD5);

            // verify the HASH is equivalent
            allzgood = ((SHA1Hash)s.user_data()).verify(recreatedSTS);
        }

        if (DEBUG) Debug.printtagln(tag, "checkDH: sig & hash allzgood = " + 
                                        allzgood + " " + _myGuid);
        return allzgood;
            
    }

    private void makeDHVerify( QSDHPublicKey myPu, 
                             QSDHPublicKey otherPu, 
                             SecretKey macSecret,
                             DHMessage destination ) 
     throws SignatureException, InvalidKeyException {
                            
        // (1) create STSVerify info my Public + other Public + MACType
        STSVerify sts = new STSVerify( myPu, otherPu, Types.MACTYPE_HMACMD5);
        // (2) Hash the info
        SecureHash digested = new SHA1Hash( sts);
        // (3) Sign the Hash
        // next line: workaround cryptix bug
        try {reInitSigEngine();} catch (Exception e) {Carp.die(e);}
        SignedQS sig = new SignedQS( digested, 
                                     new QSSha1RsaSignature(),
                                     _sigPri.key(),
                                     _r,
                                     _signEngine);
        // (4) MAC the Signature (the SHA1Hash)
        _hmac.reset();
        _hmac.init(macSecret);
        QSByteArray mac = new QSByteArray( _hmac.doFinal(data_to_check( sig)));
        // (5) insert into destination
        destination.user_data = sig;
        destination.mac = mac;
    }

    private void makeNewDHKeys( PartyInfo info) {
        // TODO in the future, we could create a keygenerator
        // with functional access that we could simple grab
        // a ready made key. (might save some time, this can take up to 8 secs)

        KeyPair pair = _kpgen.generateKeyPair();
        info.dhpub = (DHPublicKey) pair.getPublic();
        info.dhpri = (DHPrivateKey) pair.getPrivate();
    }

    public void init(ConfigDataIF config) throws Exception {

        Debug.printtagln(tag, "HMACDHHandler initializing.------------");
        Types.registerAllTypes();

		_classifier = Classifier.getClassifier (
                new NodeId (config.getString ("node_id")));

        oneAndOnly = this;

		String stageName = config.getStage().getName();
		try 
		  {
			if(config.getBoolean("Profile"))
			  {
				config.getManager().getProfiler().add(
								 "ostore.mac.HMACDHHandler", this);
				SinkIF sink =
				  config.getManager().getStage(stageName).getSink();
				_this_sink = new SinkProxy(sink,
										   config.getManager(),
										   config.getStage().getWrapper());
			  }
			else
			  _this_sink =
				config.getManager().getStage(stageName).getSink();
		  } 
		catch (NoSuchStageException e) 
		  {
			Carp.die (stageName + ".init: could not get own sink.");
		  }


        // sts timeout length (milliseconds)
        _stsTime = config.getInt("timeoutlength");
        if (ANNOUNCE) System.out.println("timeout length= " + _stsTime);

        // dh key lifespan (milliseconds)
        _dhlifespan = config.getInt("dhlifespan");
    
        // get sts retry limit
        _retryLimit = config.getInt("sessionretries");

        // get my signature key filenames
        _pkeyfile = config.getString("pkey");
        _skeyfile = config.getString("skey");

        // are we running the selfTest? (warning this will system.exit
        // on completion.)
        _selfTest = config.getBoolean("selftest");
        
        // subscribe ourselves to certain types of messages
        // DHMessages (inbound)
        Filter filter = new Filter();
        if (! filter.requireType (Class.forName
            ("ostore.mac.DHMessage")))
        Carp.die ("HMACDHHandler: could not require type.");
        if (! filter.requireValue ("inbound", new Boolean (true)))
            Carp.die ("HMACDHHandler: could not require inbound.");
        _classifier.subscribe( filter, _this_sink);
    
        filter = new Filter();
        if (! filter.requireType( Class.forName( 
            "ostore.mac.HMACDHHandler$TellSelf")))
        Carp.die ("HMACDHHandler: could not require type.");
        _classifier.subscribe( filter, _this_sink);

        try {
            initProviders();
            initKeyGenEngines();
            initSigEngine();
            initMacEngine();
            //if (_selfTest) 
                selfTest(); //TODO for some reason, skipping the selfTest breaks things...?? .... 2/1/2002 -- is this true anymore?...
            if (DEBUG || ANNOUNCE) Debug.turnOn(tag);
        } catch (Exception e) {
            Carp.die(e);
        }

        // begin DH key life countdown
        TellSelf dhexpire = new TellSelf( null, TellSelf.DH_TIMEOUT);
        _classifier.dispatch_later(dhexpire, _dhlifespan);
        
        Debug.printtagln(tag, "HMACDHHandler Initiated.---------------");

    }

    /**
     * Return the HMACDHHandler used by this machine. <br></br> 
     * Envisioned usage: Pass the string <tt>HMAC
     */
    public static HMACDHHandler getInstance() { 
	Assert.assert (oneAndOnly != null, 
		"HMACDHHandler getInstance (): no MAC stage running!");
	return oneAndOnly; 
    }

    /**
     *  Establish a connection between the guid and this instance of sandstorm
     * <br></br> 
     * Once a connection is established, the MacHandler will broadcast
     * a {@link MacHandler.KeyEstablished} announcement.
     */
    public void establish(SecureHash guid) {
        try {
            // send a message back to myself through to the eventhandler
            // to initiate a DH handshake (STS) with this guid.
            TellSelf msg = new TellSelf (guid, TellSelf.INIT);
            _classifier.dispatch(msg);
                //JUMP SinkFullException
        } catch (SinkFullException e) {
                // TODO what do do if the sink is full?
                if (DEBUG) System.out.println("sink full exception");
        }
    }

    /**
     * Note: a failure may mean that a message was sent right before
     * the sender decided to change keys that that message arrived 
     * after a new key was established (though this is unlikely) it
     * is a possibility that needs to be accounted for.
     * @return <code>true</code> iff this message is authentic, i.e.
     * verified by the internal MAC
     */
    public boolean authenticate (QuickSerializable user_data, 
	    QSByteArray incoming_mac, SecureHash guid)
        throws GuidNotEstablished {

        SecretKey secret;
        int state;
        byte[] computed_mac;
        
        // get secret or throw GuidNotEstablished
        // since the get operation is synchronized, that's all we
        // need for synchronization here.
        PartyInfo info = (PartyInfo) establishedParties.get(guid);
        if (info == null) 
            throw new GuidNotEstablished(guid.toString());
        state = info.state;

        if (state != PartyInfo.ESTABLISHED) 
            throw new GuidNotEstablished(guid.toString());
        secret = info.secret;

        // init engine and do mac
        try {
            _hmac.init(secret);
            computed_mac = _hmac.doFinal(data_to_check (user_data));
        } catch (InvalidKeyException e) {
            //TODO does this mean the key simply expired?
            Carp.out("STRANGE THINGS: HMACDHHandler: InvalidKeyException" +
                     "while authenticating");
            return false;
        }

        // compare and return result
        return ByteUtils.equals(computed_mac, incoming_mac.bytes ());
    }

    /**
     * Computes the MAC code for the this message
     * @return <code>true</code> iff the MAC was computed successfully.
     */
    public QSByteArray doMAC(QuickSerializable user_data, SecureHash guid)
     throws GuidNotEstablished {

        SecretKey secret;
        int state;
        byte[] mac = null;
        
        // get secret or throw GuidNotEstablished
        PartyInfo info = (PartyInfo) establishedParties.get(guid);
        if (info == null) 
            throw new GuidNotEstablished(guid.toString());
        secret = info.secret;

        // init engine and do mac
        try {
        _hmac.init(secret);
        mac = _hmac.doFinal( data_to_check( user_data));
        } catch (InvalidKeyException e) {
            //TODO does this mean the key simply expired?
            Carp.out("STRANGE THINGS: HMACDHHandler: InvalidKeyException" +
                     "while doingMAC");
        }

	return new QSByteArray (mac);
    }

    public String idString() { return "HMAC-DH Mac Handler"; }

    public SecureHash myGuid() { return _myGuid; }

    void initProviders()
     throws NoSuchProviderException,
            NoSuchAlgorithmException {    
        Debug.printtagln(tag, "Adding Provider: CryptixJCE");
        Provider cryptixJCE = new cryptix.jce.provider.CryptixCrypto();
        Security.addProvider(cryptixJCE);
        //cryptixJCE.list(System.out);

        Debug.printtagln(tag, "Adding Provider: CryptixRandom");
        Provider cryptixRandom = new cryptix.jce.provider.CryptixRandom();
        Security.addProvider(cryptixRandom);
        //cryptixRandom.list(System.out);

        // Initialize

          // Initialize SecureRandom (Random number generator)
          boolean available = DevRandom.isAvailable();
          Debug.printtagln(tag, "DevRandom.isAvailable = " + available);
          // Note that we are using DevRandom from CryptixRandom
          // which pulls their numbers from /usr/random
          // which, by the way, is butt fast.
          _r = SecureRandom.getInstance("DevRandom","CryptixRandom");
    }

    void initKeyGenEngines() 
     throws NoSuchAlgorithmException,
            NoSuchProviderException {
        _kpgen = KeyPairGenerator.getInstance("DH","CryptixCrypto");
        _kpgen.initialize( _keysize, _r);
        _kagree = KeyAgreement.getInstance("DH", "CryptixCrypto");
    }

    void initSigEngine() 
     throws NoSuchProviderException,
            NoSuchAlgorithmException,
            InvalidKeyException,
            QSException {
        _signEngine = Signature.getInstance("SHA-1/RSA/PKCS#1", "CryptixCrypto");
        //TODO should we be grabbing this from some global config for
        // this OceanStore instance? we ought to have a EncryptedQSIO

        // Grab keys from _pkeyfile and _skeyfile
        try {
            _sigPub = (QSPublicKey) QSIO.read(_pkeyfile);
            _sigPri = (QSPrivateKey) QSIO.read(_skeyfile);
        } catch (IOException e) {
            Carp.die ("HMACDHHandler: IO Exception reading " +
                      _pkeyfile + " or " + _skeyfile );
        }
        _signEngine.initSign(_sigPri.key(), _r);

        // get my guid
        _myGuid = new SHA1Hash (_sigPub);

        if (ANNOUNCE) Debug.printtagln(tag,"HMACDH: My Guid is " + _myGuid);
    }

    void reInitSigEngine()
     throws NoSuchProviderException,
            NoSuchAlgorithmException,
            InvalidKeyException {
        _signEngine = Signature.getInstance("SHA-1/RSA/PKCS#1", "CryptixCrypto");
    }

    void initMacEngine() 
     throws NoSuchAlgorithmException,
            NoSuchProviderException {
        // Initalize the Mac Engine
        _hmac = Mac.getInstance("HMAC-MD5", "CryptixCrypto");
    }

    void selfTest() {
        Debug.printtagln(tag, "STARTING SELFTEST ----------------------------");

        // testing timeoutlength ... and dhlifespan maybe 
        Debug.printtagln(tag, "testing timing--------------------------------");
        // timeoutlength
       // TODO TODO TODO 


        
        PartyInfo AlfInfo = new PartyInfo();
        PartyInfo BigFootInfo = new PartyInfo();

        // create DHKeys for Alf 
        Debug.printtagln(tag, "(1) Alf: create a set of dh keys of size " + 
                         _keysize + " ...");
        makeNewDHKeys(AlfInfo);
        Debug.printtagln(tag, "done.");

        // create DHKeys for Bigfoot 
        Debug.printtagln(tag, "(2) Bigfoot: create a set of dh keys of size " + 
                           _keysize + " ...");
        makeNewDHKeys(BigFootInfo);
        Debug.printtagln(tag, "done.");

        // get signature keys from quick.pri and quick.pub
        // we did this already and got: _sigPub and _sigPri
        // testing w/ two DIFFERENT sig keys is done in StageTester.java
        
        // generate secret
        Debug.printtagln(tag, "(3) Alf: generate shared secret ...");
        try {
            _kagree.init(AlfInfo.dhpri); //JUMP InvalidKeyException
            _kagree.doPhase(BigFootInfo.dhpub, true);
            byte[] seed = _kagree.generateSecret();
            AlfInfo.secret = new RawSecretKey("DH seed", seed);
        } catch (InvalidKeyException e) {
            Carp.die(e);
        }
        Debug.printtagln(tag, "done.");

        // create test DHMessage 
        Debug.printtagln(tag, "(4) Alf: create test DHMessage (RESPOND) ...");
        DHMessage dhmTest = new DHMessage(new SHA1Hash(),     //toWhom
                                          AlfInfo.dhpub,     //DH Public key
                                          _sigPub,            //Sign Pub key
                                          new QSHMACMD5(),    //mac type
                                          null, null,         //verification
                                          DHMessage.RESPOND); // type 
        
        // **** makeDHVerify 
        try {
            makeDHVerify( new QSDHPublicKey(AlfInfo.dhpub), 
                          new QSDHPublicKey(BigFootInfo.dhpub), 
                          AlfInfo.secret, 
                          dhmTest); 
            // JUMP SignatureException 
            // JUMP InvalidKeyException
        } catch (SignatureException e) { 
            Carp.die(e);
        } catch (InvalidKeyException e) {
            Carp.die(e);
        }
        Debug.printtagln(tag, "done.");
        // encode DHMessage
        Debug.printtagln(tag, "(5) Alf: encoding DHMessage...");
        int[] offset = {0};
        dhmTest.to_bytes( null, offset);
        byte[] encoding = new byte[offset[0]];
        offset[0] = 0;
        dhmTest.to_bytes( encoding, offset);
        Debug.printtagln(tag, "done.");
        // decode DHMessage
        Debug.printtagln(tag, "(6) BigFoot: decode DHMessage...");
        offset[0] = 0;
        try {
            dhmTest = new DHMessage( encoding, offset); 
            // JUMP QSException
        } catch ( QSException e) {
            Carp.die(e);
        }
        Debug.printtagln(tag, "done.");
        // BigFoot figures out secret key.
        Debug.printtagln(tag, "(7) BigFoot: figure out secret...");
        try {
            _kagree.init(BigFootInfo.dhpri); // JUMP InvalidKeyException
            _kagree.doPhase( dhmTest.QSkey.key(), true );
            byte[] seed = _kagree.generateSecret();
            BigFootInfo.secret = new RawSecretKey("DH seed", seed);
        } catch (InvalidKeyException e) {
            Carp.die(e);
        }
        BigFootInfo.dhpubOther = (DHPublicKey)dhmTest.QSkey.key();
        BigFootInfo.sigpub = dhmTest.sigkey;
        Debug.printtagln(tag, "done.");
        // **** checkDHVerify
        Debug.printtagln(tag, "(8) BigFoot: checkDHVerify...");
        try {
            boolean check = checkDHVerify( dhmTest.user_data, dhmTest.mac, 
                                                 dhmTest, 
                                              BigFootInfo); 
            // JUMP InvalidKeyException
            // JUMP SignatureException
            // JUMP TypeTable.NoSuchTypeCode
            Debug.printtagln(tag, "[returns:" + check + "]...");
            if (!check) {
                Debug.printtagln(tag, "checkDHVerify FAILED");
                Carp.die();
            }
        } catch (InvalidKeyException e) {
            Carp.die(e);
        } catch (SignatureException e) {
            Carp.die(e);
        } catch (TypeTable.NoSuchTypeCode e) {
            Carp.die(e);
        }   
        Debug.printtagln(tag, "done.");
        Debug.printtagln(tag, "SELFTEST DONE----------------------------------");
		if (_selfTest) System.exit(0);
    }

    public void destroy() throws Exception {
        //TODO
        // free timer?
    }

    //  methods for ProfilableIF
    /* Specified by mdw.sandStorm.api.ProfilableIF */
    public int profileSize()
    {
	return _this_sink.size();
    }
    //  end methods for ProfilableIF

    private class TellSelf implements QueueElementIF {
        public static final int INIT        = 0;
        public static final int STS_TIMEOUT = 1; 
        public static final int DH_TIMEOUT  = 2;

        public SecureHash guid;
        public int msg;
        public TellSelf( SecureHash guid, int msg) {
            this.guid = guid;
            this.msg = msg;
        }
    }
}
