/**
 * Network.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.network;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StreamTokenizer;
import java.net.InetAddress;
import java.util.HashMap;
import java.util.Map;
import mdw.sandStorm.api.ConfigDataIF;
import mdw.sandStorm.api.EventHandlerException;
import mdw.sandStorm.api.EventHandlerIF;
import mdw.sandStorm.api.NoSuchStageException;
import mdw.sandStorm.api.NoSuchStageException;
import mdw.sandStorm.api.ProfilableIF;
import mdw.sandStorm.api.QueueElementIF;
import mdw.sandStorm.api.SinkClosedEvent;
import mdw.sandStorm.api.SinkClosedException;
import mdw.sandStorm.api.SinkException;
import mdw.sandStorm.api.SinkFlushedEvent;
import mdw.sandStorm.api.SinkIF;
import mdw.sandStorm.core.BufferElement;
import mdw.sandStorm.internal.SinkProxy;
import mdw.sandStorm.lib.aSocket.ATcpClientSocket;
import mdw.sandStorm.lib.aSocket.ATcpConnectFailedEvent;
import mdw.sandStorm.lib.aSocket.ATcpConnection;
import mdw.sandStorm.lib.aSocket.ATcpInPacket;
import mdw.sandStorm.lib.aSocket.ATcpListenFailedEvent;
import mdw.sandStorm.lib.aSocket.ATcpListenSuccessEvent;
import mdw.sandStorm.lib.aSocket.ATcpServerSocket;
import mdw.sandStorm.lib.aSocket.aSocketInputStream;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.dispatch.Signal;
import ostore.util.Assert;
import ostore.util.ByteUtils;
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.NodeId;
import ostore.util.Pair;
import ostore.util.PriorityQueue;
import ostore.util.Queue;
import ostore.util.QuickSerializable;
import ostore.util.TypeTable;
import soss.network.TransitStub;

/**
 * A stage to tie the OceanStore dispatch model to Sandstorms asyncronous
 * networking functionality.  To send a message on the network, inherit it
 * from NetworkMessage, set inbound = false, and dispatch it.  This stage
 * will catch the event, open a connection (if one is not open already),
 * and send it.  To receive messages, just listen for the correct type
 * conditioned on inbound == true.
 *
 * @author     Sean C. Rhea
 * @version    $Id: Network.java,v 1.38 2002/04/03 02:50:13 srhea Exp $
 */
public class Network implements EventHandlerIF, ProfilableIF {

    private String class_tag = "Network";
    public static final boolean DEBUG = 
	ostore.util.DebugFlags.NETWORK_NETWORK;

    private void dispatch (QueueElementIF item) {
	final String tag = class_tag + ".dispatch";
	if (DEBUG) Debug.printtagln (tag, "dispatching " +
		item.getClass () + " event: " + item + ".");
	try {
	    _classifier.dispatch(item);
	} catch (SinkException e) {
	    BUG (class_tag + ".dispatch: could not dispatch " + item);
	}
    }

    private static final void BUG (String msg) {
	Carp.die (msg + "  This is a bug in ostore.network.Network.  " +
		"Please let Sean know about it ASAP.");
    }

    private void handle_alarm () {
	final String tag = class_tag + ".handle_alarm";

	if (! _close_me.empty ()) {
	    long now = System.currentTimeMillis ();
	    long top = _close_me.topPriority (false);

	    while (now > top) {
		ATcpConnection conn_to_close = 
		    (ATcpConnection) _close_me.extract (false);
		NodeId peer = new MyNodeId (conn_to_close.getPort (),
			conn_to_close.getAddress ());

		try {
		    conn_to_close.close (_this_sink);
		}
		catch (SinkClosedException e) {
		    if (DEBUG) Debug.printtagln (tag,
			    "Tried to close already closed "
			    + "client connection.");
		}

		if (_close_me.empty ())
		    break;
		else
		    top = _close_me.topPriority (false);
	    }
	}
	Signal.getSignal (_this_node_id).alarm ("NetworkStage", 10000);
    }

    private void handle_atcp_connection (ATcpConnection connection) {
	final String tag = class_tag + ".handle_atcp_connection";

	NodeId peer = new MyNodeId (connection.getPort (),
		connection.getAddress ());

	if (DEBUG) Debug.printtagln (tag, 
		"Network got ATcpConnection event from " + peer + ".");

	if (_pending.get (peer) != null) {

	    if (DEBUG) Debug.printtagln (tag, 
		    "This connection was established locally so"
		    + " the port number is correct.");

	    // We initiated this connection, so the port number is
	    // correct, but we need to send an IdentityMessage to
	    // let the other side know our server port before we
	    // send any of the pending messages.  We will complete
	    // this when we get the SinkFlushedEvent later.

	    IdentityMessage ident = 
		new IdentityMessage (peer, false, _this_node_id);
	    byte [] buffer = msg_to_bytes (ident);

	    try {
		connection.enqueue (new BufferElement (buffer));
	    }
	    catch (SinkException e) {
		if (DEBUG) Debug.printtagln (tag,
			"Got SinkException sending "
			+ "IdentityMessage. (This is bad.)"); 
		// TODO: what to do here?
	    }

	    try {
		connection.flush (_this_sink);
	    }
	    catch (SinkClosedException e) {
		BUG (tag + ": Got SinkClosedException on flush.  "
			+ "(This is bad.)"); 
	    }
	}
	else {

	    if (DEBUG) Debug.printtagln (tag, 
		    "This connection was established remotely so"
		    + " we don't know the server port yet.");

	    // We did not initiate this connection, so the port
	    // number is (possibly) different from the port number
	    // the sender listens on.  We'll hang onto this
	    // connection in anticipation of a message mapping it
	    // to the listening port.  We also store the current
	    // time, so we can timeout if that message never comes.

	    Pair p = new Pair ();
	    p.first = connection;
	    p.second = new Long (System.currentTimeMillis ());

	    _mystery_connections.put (peer, p);
	}

	// Start allowing incoming packets from this connection.

	connection.startReader (_this_sink);
    }

    public void handle_sink_flushed_event (SinkFlushedEvent flush) {
	final String tag = class_tag + ".handle_sink_flushed_event";

	ATcpConnection conn = null;

	try {
	    conn = (ATcpConnection) (flush.sink);
	}
	catch (ClassCastException e) {
	    BUG (tag + ": Got SinkFlushedEvent for a sink which "
		    + "was not an ATcpConnection.");
	    /*if (DEBUG) Debug.printtagln (tag,
	      "Got SinkFlushedEvent for a sink which "
	      + "was not an ATcpConnection.");
	      return;*/
	}

	NodeId peer = new MyNodeId (conn.getPort (), conn.getAddress ());

	if (DEBUG) Debug.printtagln (tag, 
		"Locally established connection to " + peer 
		+ " flushed.  Sending pending messages.");

	// First off, we need to cache this connection so further
	// outgoing messages can be enqueued on it.  In the process, we
	// check for pending outgoing messages.

	Queue pending = (Queue) _pending.get (peer);
	if (pending == null) {
	    BUG (tag + ": No pending messages!  Confused.");
	    /*if (DEBUG) Debug.printtagln (tag, 
	      "No pending messages!  Confused.");
	      return;*/
	}

	// We initiated this connection, so the port number is
	// correct, and we can just go ahead and add it to our
	// list of connections.
	Assert.assert (! _connections.containsKey (peer),
		tag + ": connection already exists.");
	if (DEBUG) Debug.printtagln (tag,
		"Adding connection for " + peer + ".");
	_connections.put (peer, conn);

	// Moreover, we need to remove the list of pending
	// messages and send them out.  We remove the list here
	// and send them out below.
	_pending.remove (peer);

	send_pending_messages (pending, conn);
    }

    private void handle_atcp_inpacket (ATcpInPacket packet) {
	final String tag = class_tag + ".handle_atcp_inpacket";

	ATcpConnection conn = packet.getConnection ();
	NodeId peer = new MyNodeId (conn.getPort (), conn.getAddress ());
	aSocketInputStream is = 
	    (aSocketInputStream) _connection_streams.get (peer);
	if (is == null) {
	    is = new aSocketInputStream ();
	    _connection_streams.put (peer, is);
	}

	is.addPacket (packet);

	while (true) {

	    // Decode the packet and send off an IncomingMessage as a
	    // result.

	    int [] offset = {0};
	    int size = 0;
	    byte [] data = null;

	    try {
		if (is.available () < 4) {
		    // Not enough to even read the size of the next
		    // message.
		    return;
		}

		is.mark (0);
		byte [] size_bytes = new byte [4];
		if (is.read (size_bytes) != 4) {
		    BUG (tag + ": should have caught this earlier.");
		}

		size = ByteUtils.bytesToInt (size_bytes, offset);

		if (is.available () < size) {
		    // Whole message is not yet here.
		    is.reset ();
		    return;
		}

		data = new byte [size];
		if (is.read (data) < size) {
		    BUG (tag + ": should have caught this earlier.");
		}
		offset [0] = 0;
	    }
	    catch (IOException e) {
		BUG (tag + ": caught " + e);
	    }

	    if (_client_port_to_server_port.containsKey (peer)) {
		NodeId server_port = 
		    (NodeId) _client_port_to_server_port.get (peer);
		if (DEBUG) Debug.printtagln (tag, "Translating " + 
			peer + " to " + server_port + ".");
		peer = server_port;
	    }

	    if (DEBUG) Debug.printtagln (tag,
		    "Received a message from " + peer + ".");

	    if (size < ByteUtils.SIZE_INT)  {
		if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (tag + ": "
		+
			"Received message from " + peer + 
			" too small to have a type code.  " +
			"Dropping it.");
		return;
	    }

	    long type_code = ByteUtils.bytesToLong (data, offset);
	    NetworkMessage msg = null;   
	    try {
		msg = (NetworkMessage) 
		    TypeTable.new_object (type_code, data, offset);
	    }
	    catch (Exception e) {
		if (DebugFlags.LOG_BAD_NETWORK_EVENTS) {
		    Carp.out (tag + ": " +
			    "Received message of type " + 
			    Long.toHexString (type_code) + " from " + peer + 
			    " that caused exception " + e + 
			    " on decode.  Dropping it.  " + 
			    "(Stack trace follows.)"); 
		    e.printStackTrace ();
		}
		return;
	    }

	    msg.peer = peer;
	    msg.inbound = true;

	    if (DEBUG) Debug.printtagln (tag, "Message is " + msg + ".");

	    if (msg instanceof IdentityMessage) {
		process_identity_message ((IdentityMessage) msg);
	    }
	    else if (_ts != null) {
		Integer Src = (Integer) _node_id_to_vertex.get (peer);
		if (Src == null) {
		    Carp.die ("Got message from " + peer + 
			    ", who is not in the topology mapping.");
		}
		Integer Dst = (Integer) _node_id_to_vertex.get (_this_node_id);
		if (Dst == null) {
		    Carp.die ("We (" + _this_node_id + ") are somehow " +
			    "not in the topology mapping.");
		}
		int src = Src.intValue ();
		int dst = Dst.intValue ();
		TransitStub.PathInfo pi = _ts.shortest_path (
			src, dst, size, true /* record */, type_code);
		long delay = Math.round (pi.route_time * 1000.0);
		if (delay == 0)
		    dispatch (msg);
		else {
		    if (DEBUG) Debug.printtagln (tag, 
			    "Delaying msg from " + src + 
			    " to " + dst + " by " + delay + " ms.");
		    _classifier.dispatch_later (msg, delay);
		}
	    }
	    else {
                /*
                dispatch((QueueElementIF)
                         new RecordMsgEvent(msg.type_code(), size));
                */
		dispatch (msg);
	    }

	    if (DEBUG) Debug.printtagln (tag,
		    "Message correctly decoded and dispatched.");
	}
    }

    private void handle_outbound_network_message (NetworkMessage msg) {
	final String tag = class_tag + ".handle_outbound_network_message";

	if (DEBUG) Debug.printtagln (tag, "Network got " + msg 
		+ " to send to " + msg.peer + ".");

	// If this is the first time we have sent a message to a node,
	// we need to open up a connection to it first.

	ATcpConnection conn = null;

	if ((! _connections.containsKey (msg.peer)) || 
		_closing.containsKey (msg.peer)) {
	    // If we either do not have a connection to this peer as a
	    // remote server, or if we do have such a connection, but
	    // it is being closed, then see if we have a connection
	    // from it as a remote client.

	    if (_server_port_to_client_port.containsKey (msg.peer)) {
		NodeId client_port = 
		    (NodeId) _server_port_to_client_port.get (msg.peer);
		conn = (ATcpConnection) _connections.get (client_port);
	    }
	}
	else {
	    // Otherwise, we can use the connection to this peer as a
	    // remote server.

	    conn = (ATcpConnection) _connections.get (msg.peer);
	}

	if (conn == null) {
	    // If we still have not found a connection, we must
	    // establish one.  If there is already a pending connection
	    // open, add this message to it's queue.  If there is no
	    // pending connection, start one.

	    Queue pending = (Queue) _pending.get (msg.peer);
	    if (pending == null) {

		// Create a queue for pending messages
		pending = new Queue ();
		_pending.put (msg.peer, pending);

		if (DEBUG) Debug.printtagln (tag, 
			"Openning new socket to " + msg.peer 
			+ ".");
		// Open the socket
		new ATcpClientSocket (
			msg.peer.address (), 
			msg.peer.port (), 
			_this_sink);
	    }
	    else if (DEBUG) {
		Debug.printtagln (tag, 
			"Socket exists but is not yet open.");
	    }

	    Debug.printtagln (tag, "Queuing message for now.");
	    pending.enqueue (msg);
	}
	else {

	    Debug.printtagln (tag, "Sending message out on network.");
	    // If there is already a connection open to this
	    // destination, enqueue the message to it.

	    byte [] buffer = null;
            //  PRE: RecordMsgEvent is a way to track bandwidth.
            //RecordMsgEvent record = null;
	    try {
		buffer = msg_to_bytes (msg);
                //record = new RecordMsgEvent(msg.type_code(), buffer.length);
	    }
	    catch (Exception e) {
		Carp.out(tag + ": to_bytes error on type " +
			msg.getClass () + ".");
		e.printStackTrace(System.err);
		Carp.die(e);
		return;
	    }
	    BufferElement buffer_element = new BufferElement (buffer);

	    try {
		conn.enqueue (buffer_element);
                //dispatch(record);
	    }
	    catch (SinkException e) {
		if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (tag + ": "
		+
			"Got SinkException on enqueue of "
			+ "outgoing message.  Dropping it.");
	    }
	}
    }

    private void handle_sink_closed_event (SinkClosedEvent sce) {
	final String tag = class_tag + ".handle_sink_closed_event";

	if (sce.sink instanceof ATcpConnection) {
	    ATcpConnection conn = (ATcpConnection) sce.sink;
	    NodeId peer = new MyNodeId (conn.getPort (),
		    conn.getAddress ());

	    ATcpConnection conn2 = 
		(ATcpConnection) _connections.get (peer);
	    if (conn2 == null) {
		BUG (tag + ":got SinkClosedEvent for non-existant "
			+ "connection: " + peer);
	    }
	    else if (conn2 == conn) {

		if (DEBUG) Debug.printtagln (tag, 
			"Removing connection " + peer + " because "
			+ "of SinkClosedEvent.");

		_connections.remove (peer);
		_connection_streams.remove (peer);
		_closing.remove (peer);
		if (_server_port_to_client_port.containsKey (peer)) {
		    NodeId image = (NodeId)
			_server_port_to_client_port.remove (peer);
		    _client_port_to_server_port.remove (image);
		}
		else 
		    _client_port_to_server_port.remove (peer);

		Assert.assert (! _pending.containsKey (peer),
			tag + ": should not be a pending conn.");
	    }
	    else {
		// We closed this connection ourselves, probably
		// because it was duplicated.  Do nothing.
		BUG (tag + ": duplicate connection to " + peer + ".");
	    }
	}
	else {
	    BUG (tag + ": got SinkClosedEvent for: " + sce.sink);
	}
    }

    private void handle_atcp_connect_failed_event (
	    ATcpConnectFailedEvent failure) {
	final String tag = class_tag + ".handle_atcp_connect_failed_event";

	ATcpClientSocket socket = failure.getSocket ();

	NodeId peer = new MyNodeId (socket.getPort (), socket.getAddress ());
	boolean sensible = false;

	if (_pending.containsKey (peer)) {
	    _pending.remove (peer);
	    if (DEBUG) Debug.printtagln (tag, 
		    "Removing pending connection " + peer + " because "
		    + "of ATcpConnectFailedEvent.");
	    sensible = true;
	}

	if (_connections.containsKey (peer)) {
	    if (DEBUG) Debug.printtagln (tag, 
		    "Removing existing connection " + peer + " because "
		    + "of ATcpConnectFailedEvent.");
	    _connections.remove (peer);
	    _connection_streams.remove (peer);
	    Assert.assert (! _closing.containsKey (peer),
		    tag + ": should not be in _closing.");
	    sensible = true;
	}

	if (_server_port_to_client_port.containsKey (peer)) {
	    _server_port_to_client_port.remove (peer);
	    if (DEBUG) Debug.printtagln (tag, 
		    "Removing server->client map " + peer + " because "
		    + "of ATcpConnectFailedEvent.");
	    sensible = true;
	}

	if (_client_port_to_server_port.containsKey (peer)) {
	    _client_port_to_server_port.remove (peer);
	    if (DEBUG) Debug.printtagln (tag, 
		    "Removing client->server " + peer + " because "
		    + "of ATcpConnectFailedEvent.");
	    sensible = true;
	}

	/* TODO: this doesn't seem to cause any problems... */
	{
	  String str = null;
	  {
	    str = failure.toString();
	    String errorStr = new String("error");
	    int index = str.lastIndexOf(errorStr);
	    if(index >= 0)
	    {
		String failStr = new String("fail ");
		byte[] bytes = str.getBytes();
		byte[] failBytes = failStr.getBytes();
		for(int i = 0; i < failBytes.length; i++)
		    bytes[index + i] = failBytes[i];
		str = new String(bytes);
	    }
	  }
	  
	  System.err.println(tag + ":could NOT form "
			     + "connection to " + peer + 
			     " because " + str + "."  );
	  System.err.flush();
	}
	if (! sensible ) {
	    BUG (tag + ":Got ATcpConnectFailedEvent for unknown "
		    + "connection to " + peer + ".");
	}
    }

    public void handleEvent (QueueElementIF item) 
    throws EventHandlerException {

	final String tag = class_tag + ".handleEvent";

	if (DEBUG)
	{
	    String errorStr = new String("error");
	    String str = item.toString();
	    int index = str.lastIndexOf(errorStr);
	    if(index >= 0)
	    {
		String failStr = new String("fail ");
		byte[] bytes = str.getBytes();
		byte[] failBytes = failStr.getBytes();
		for(int i = 0; i < failBytes.length; i++)
		    bytes[index + i] = failBytes[i];
		str = new String(bytes);
	    }
	    Debug.printtagln (tag, 
		    "handleEvent called with " + str + " event.");
	}

	if (item instanceof ostore.dispatch.Signal.AlarmSignal) {
	    handle_alarm ();
	}
	else if (item instanceof ATcpConnection) {
	    handle_atcp_connection ((ATcpConnection) item);
	}
	else if (item instanceof SinkFlushedEvent) {
	    handle_sink_flushed_event ((SinkFlushedEvent) item);
	}
	else if ((item instanceof IdentityMessage) && 
		(((IdentityMessage) item).inbound == true)) {
	    process_identity_message ((IdentityMessage) item);
	}
	else if (item instanceof ATcpInPacket) {
	    handle_atcp_inpacket ((ATcpInPacket) item);
	}
	else if ((item instanceof NetworkMessage) && 
		(((NetworkMessage) item).inbound == false)) {
	    handle_outbound_network_message ((NetworkMessage) item);
	}
	else if (item instanceof SinkClosedEvent) {
	    handle_sink_closed_event ((SinkClosedEvent) item);
	}
	else if (item instanceof ATcpConnectFailedEvent) {
	    handle_atcp_connect_failed_event ((ATcpConnectFailedEvent) item);
	}
	else if (item instanceof ATcpListenSuccessEvent) {
	    ATcpListenSuccessEvent e = (ATcpListenSuccessEvent) item;
	    ATcpServerSocket s = e.getSocket ();
	    int port = s.getPort ();
	    System.out.println (class_tag + " now listening on port " + port + ".");
	}
	else if (item instanceof ATcpListenFailedEvent) {
	    ATcpListenFailedEvent e = (ATcpListenFailedEvent) item;
	    ATcpServerSocket s = e.getSocket ();
	    int port = s.getPort ();
	    System.err.println ("ERROR: Network could not listen on " + 
		    "port " + port + ".  There is probably another " + 
		    "program using the port.  Dying...");
	    System.exit (1);
	}
	else {
	    Carp.die (tag + ": got unknown event of type " + 
		    item.getClass ().getName () + ": " + item + ".");
	}
    }

    public void handleEvents(QueueElementIF items[]) 
    throws EventHandlerException {

	final String tag = class_tag + ".handleEvents";

	if (DEBUG) Debug.printtagln (tag, 
		"handleEvents called with " + items.length + " events.");

	for(int i=0; i<items.length; i++)
	    handleEvent(items[i]);
    }

    public void init (ConfigDataIF config) throws Exception {
	
	String stageName = config.getStage().getName();
	class_tag = stageName;
	final String tag = class_tag + ".init";

        if (DEBUG) Debug.turnOn (class_tag);

	_connections = new HashMap ();
	_connection_streams = new HashMap ();
	_mystery_connections = new HashMap ();
	_server_port_to_client_port = new HashMap ();
	_client_port_to_server_port = new HashMap ();
	_pending = new HashMap ();
	_closing = new HashMap ();
	_close_me = new PriorityQueue (10);

	String topology_graph = config.getString ("topology_graph");
	String topology_mapping = config.getString ("topology_mapping");

	if ((topology_graph == null) ^ (topology_mapping == null)) {
	    Carp.die (tag + ": must specify topology_graph " +
		    " and topology_mapping together.");
	}

	if (topology_graph != null) {

	    // Open the graph file.
	    try {
		_ts = new TransitStub (topology_graph);
	    }
	    catch (IOException e) {
		Carp.die (tag + ": caught exception when creating " +
			"graph: " + e + ".");
	    }

	    // Read the IP:Port to Vertex mapping
	    _node_id_to_vertex = new HashMap ();
	    StreamTokenizer tok = null;
	    try {
		FileInputStream is = new FileInputStream (topology_mapping);
		Reader reader = 
		    new BufferedReader (new InputStreamReader (is));
		tok = new StreamTokenizer (reader);
	    }
	    catch (Exception e) {
		Carp.die (tag + ": caught exception " + e + " trying " +
			"to read topology mapping " + topology_mapping);
	    }

	    tok.resetSyntax ();
	    tok.whitespaceChars ((int) ' ', (int) ' ');
	    tok.wordChars ((int) 'a', (int) 'z');
	    tok.wordChars ((int) 'A', (int) 'Z');
	    tok.wordChars ((int) '0', (int) '9');
	    tok.wordChars ((int) ':', (int) ':');
	    tok.wordChars ((int) '.', (int) '.');

	    int lineno = 1;
	    tok.nextToken ();
	    while (tok.ttype == StreamTokenizer.TT_EOL) {
		tok.nextToken ();
		lineno++;
	    }

	    while (tok.ttype != StreamTokenizer.TT_EOF) {

		// Read the IP:Port
		if (tok.ttype != StreamTokenizer.TT_WORD) {
		    Carp.die (tag + ": expected IP:Port on line " + 
			    lineno + ": " + tok);
		}
		NodeId nid = null;
		try {
		    nid = new NodeId (tok.sval);
		}
		catch (Exception e) {
		    Carp.die (tag + ": Bad IP:Port string on " +
			    "line " + lineno + ": " + e);
		}

		// Read the Vertex
		tok.nextToken ();
		if (tok.ttype != StreamTokenizer.TT_WORD) {
		    Carp.die (tag + ": expected IP:Port on line " + 
			    lineno + ": " + tok);
		}
		Integer vertex = null;
		try {
		    vertex = new Integer (Integer.parseInt (tok.sval));
		}
		catch (NumberFormatException e) {
		    Carp.die (tag + ": number format error reading " +
			    "vertex on line " + lineno + ": " + e);
		}

		// Put them in the mapping.
		if (_node_id_to_vertex.containsKey (nid)) {
		    Carp.die (tag + ": duplicate node id " + nid +
			    " in topology mapping " + topology_mapping);
		}
		_node_id_to_vertex.put (nid, vertex);

		tok.nextToken ();
		while (tok.ttype == StreamTokenizer.TT_EOL) {
		    tok.nextToken ();
		    lineno++;
		}
	    }
	}
    
	_this_node_id = new NodeId (config.getString ("node_id"));	
  	int port = _this_node_id.port();
	_classifier = Classifier.getClassifier (_this_node_id);

	if (DEBUG) Debug.printtagln (tag, "Starting on port " + port + ".");

	try 
	{
	    if(config.getBoolean("Profile"))
	    {
		config.getManager().getProfiler().add("ostore.network.Network", this);
		SinkIF sink =
		    config.getManager().getStage(stageName).getSink();
		_this_sink = new SinkProxy(sink,
			      config.getManager(),
			      config.getStage().getWrapper());
	    }
	    else
		_this_sink =
		    config.getManager().getStage(stageName).getSink();
	} 
	catch (NoSuchStageException e) 
        {
	    BUG (tag + ": could not get own sink.");
	}
		
	Filter filter = new Filter ();
	if (! filter.requireType (
		    Class.forName ("ostore.network.IdentityMessage")))
	    BUG (tag + ": could not require type.");
	if (! filter.requireValue ("inbound", new Boolean (true)))
	    BUG (tag + ": could not require outbound.");
	_classifier.subscribe (filter, _this_sink);

	filter = new Filter ();
	if (! filter.requireType (
		    Class.forName ("ostore.network.NetworkMessage")))
	    BUG (tag + ": could not require type.");
	if (! filter.requireValue ("inbound", new Boolean (false)))
	    BUG (tag + ": could not require outbound.");
	_classifier.subscribe (filter, _this_sink);

	Signal.getSignal (_this_node_id).register (
		Signal.ALRM, "NetworkStage", _this_sink);
	Signal.getSignal (_this_node_id).alarm ("NetworkStage", 10000);
	
	// Open the server socket
	if (DEBUG) Debug.printtagln (tag, 
		"Openning server socket on port " + port + ".");
	new ATcpServerSocket (port, _this_sink);  

	// TODO: start timeout thread for _mystery_connections
    }

    public void destroy() throws Exception {
	// TODO
    }

    //  methods for ProfilableIF
    public int profileSize()
    {
	return _this_sink.size();
    }
    //  end methods for ProfilableIF

    protected NodeId process_identity_message (IdentityMessage ident) {

	final String tag = class_tag + ".process_identity_message";
	NodeId result = null;

	if (DEBUG) Debug.printtagln (tag,
		"Got an IdentityMessage from " + ident.peer 
		+ " purporting to be " + ident.node_id + ".");

	// Look up the mystery connection that this message corresponds
	// to.
	Pair p = (Pair) _mystery_connections.remove (ident.peer);
	if (p == null) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (tag + ": " +
		    "Got unexpected IdentityMessage from " 
		    + ident.peer + ".  Dropping it.");
	    return result;
	}

	_connections.put (ident.peer, p.first);

	result = ident.node_id;

	// TODO: authenticate it

	ATcpConnection conn_to_remote_server = 
	    (ATcpConnection) _connections.get (ident.node_id);

	if (conn_to_remote_server != null) {
	    // We have openned this connection from both sides
	    // simultaneously, so we need to pick a client and
	    // server from TCP's point of view.  Let the node with
	    // the lowest NodeId be the server.

	    if (_this_node_id.less_than (ident.node_id)) {
		// We are the server.  Close our outgoing connection to the
		// remote server and use the incoming one instead.

		if (DEBUG) Debug.printtagln (tag,
			"We are the server.  Closing old conn.");

		_closing.put (ident.node_id, conn_to_remote_server);
		_close_me.insert (conn_to_remote_server,
			System.currentTimeMillis () + 10000, false);

		if (DEBUG) Debug.printtagln (tag,
			"Adding connection for " + ident.peer + ".");
		if (DEBUG) Debug.printtagln (tag,
			"Adding translation " + ident.peer + " to " 
			+ ident.node_id + ".");
		_client_port_to_server_port.put (ident.peer, ident.node_id);
		_server_port_to_client_port.put (ident.node_id, ident.peer);
	    }
	    else {
		// We are the client.  Close our incoming connection and
		// use the outgoing one instead.

		if (DEBUG) Debug.printtagln (tag,
			"We are the client.  Closing incoming conn.");

		// Even though we are closing this connection, we may have
		// already received some other messages from it, so store
		// the mapping until it is really closed.

		_closing.put (ident.peer, p.first);
		_client_port_to_server_port.put (ident.peer, ident.node_id);
	    }
	}
	else {
	    // This connection was not yet open, so we are cool.

	    if (DEBUG) Debug.printtagln (tag,
			"Adding connection for " + ident.peer + ".");
	    if (DEBUG) Debug.printtagln (tag,
		    "Adding translation " + ident.peer + " to " 
		    + ident.node_id + ".");
	    _client_port_to_server_port.put (ident.peer, ident.node_id);
	    _server_port_to_client_port.put (ident.node_id, ident.peer);
	}

	return result;
    }

    private void send_pending_messages (Queue pending, ATcpConnection conn) {
	final String tag = class_tag + ".send_pending_messages";

	Assert.assert (pending.size () > 0, tag + ": pending queue bug.");

	QueueElementIF [] elements = new QueueElementIF [pending.size ()];
	int i = 0;
	while (! pending.empty ()) {
	    NetworkMessage msg = (NetworkMessage) pending.dequeue ();

	    if (DEBUG) Debug.printtagln (tag, 
		    "Sending pending message " + msg + ".");
	    
	    byte [] buffer = null;
	    try {
		buffer = msg_to_bytes (msg);
	    }
	    catch (Exception e) {
		if (DEBUG) Debug.printtagln (tag, "Got exception " + 
			e + " on encoding.  Dropping message.");
		return;
	    }
	    elements [i++] = new BufferElement (buffer);
	}

	try {
	    if (elements.length == 1)
		conn.enqueue (elements [0]);
	    else
		conn.enqueue_many (elements);
	}
	catch (SinkException e) {
	    if (DEBUG) Debug.printtagln (tag, "Got SinkException on " +
		    "enqueue of pending messages.  Dropping them.");
	}
    }

    private byte [] msg_to_bytes (QuickSerializable msg) {
	int [] offset = {0};
	long type_code = TypeTable.type_code (msg);
	ByteUtils.longToBytes (type_code, null, offset);
	msg.to_bytes (null, offset);
	int size = offset [0];
	byte [] buffer = new byte [size+4];
	offset [0] = 0;
	ByteUtils.intToBytes (size, buffer, offset);
	ByteUtils.longToBytes (type_code, buffer, offset);
	msg.to_bytes (buffer, offset);
	return buffer;
    }

    private NodeId _this_node_id;

    private Classifier _classifier;

    private HashMap _connections;
    private HashMap _connection_streams;
    private HashMap _mystery_connections;
    private HashMap _client_port_to_server_port;
    private HashMap _server_port_to_client_port;
    private HashMap _pending;
    private HashMap _closing;
    private SinkIF _this_sink;
    private PriorityQueue _close_me;
    private Map _node_id_to_vertex;
    private TransitStub _ts;
}

