/**
 * HMMStage.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.oil;

import java.lang.reflect.Field;
import ostore.util.QSIO;
import ostore.util.InvalidConfigFileException;
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.NodeId;
import ostore.util.DebugFlags;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import mdw.sandStorm.api.ConfigDataIF;
import mdw.sandStorm.api.EventHandlerIF;
import mdw.sandStorm.api.EventHandlerException;
import mdw.sandStorm.api.QueueElementIF;
import mdw.sandStorm.api.SinkIF;

/**
 * A superclass for stages which employ <code>HMM</code>s.
 * This class initializes a few useful <code>Class</code> objects,
 * and initializes an <code>HMM</code> and event stream filter.
 * Subclasses should call <code>ModelStage.init</code> at the
 * <b>beginning</b> of their own <code>init</code> method.
 *
 * @see HMM
 *
 * @author	Dennis Geels
 * @version	$Id: HMMStage.java,v 1.7 2001/12/13 20:50:29 geels Exp $
 */

public abstract class HMMStage implements EventHandlerIF
{
    ////////////////////       CONSTANTS	////////////////////

    // Turn on debugging iff subclasses turn theirs on.
    private static final boolean DEBUG =
	(DebugFlags.OIL_HMMBUILDER || DebugFlags.OIL_HMMPREDICTOR);

    // A few strings used in the config files:
    private static final String IMAGE_STR 	= "read_snapshot";
    private static final String EVENT_CLASS_STR	= "observed_class";
    private static final String EVENT_FIELD_STR	= "observed_field";
    private static final String M_STR		= "num_states";
    private static final String N_STR		= "num_outputs";
    private static final String THRESHOLD_STR	= "epsilon";
    
    ////////////////////    STATIC VARIABLES 	////////////////////

    // Use the subclass tag.
    protected static String class_tag;

    /**
     * The <code>Class</code> object for the <code>QueueElementIF</code>
     * Interface.
     */
    protected static Class QueueElementIF_class;

    /**
     * The <code>Class</code> object for the <code>QuickSerializable</code>
     * Interface.
     */
    protected static Class QuickSerializable_class;
    
    ////////////////////       VARIABLES	////////////////////

    /**
     * The name by which to refer to this stage.  It should be
     * the name of the corresponding <code>StageIF</code>.
     */
    protected String name;

    /**
     * The learning model to use.
     */
    protected HMM model;
    
    /**
     * The type of events to observe.
     */
    protected Class observed_type;

    /**
     * The location of the observable field within the events watched.
     * This field comprises the observed output for the <code>HMM</code>.
     */
    protected Field accessor;

    protected Classifier classifier;

    ////////////////////  STATIC CONSTRUCTORS	////////////////////

    static {
	// Initialize the <code>Class</code> objects we need to analyze our
	// configuration parameters.
	try{
	    QueueElementIF_class =
		Class.forName( "mdw.sandStorm.api.QueueElementIF" );
	    QuickSerializable_class =
		Class.forName( "ostore.util.QuickSerializable" );
	} catch( ClassNotFoundException cnfe ) {
	    Carp.die( cnfe );
	}
    }
    
    ////////////////////      CONSTRUCTORS	////////////////////
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /* Specified by mdw.sandStorm.api.EventHandlerIF */
    public void handleEvents(QueueElementIF elemarr[]) 
	throws EventHandlerException
    {
	final String tag = class_tag+".handleEvents";
	if( DEBUG ) Debug.printtagln( tag, name + " called: "
				      + elemarr.length + " elements" );
	for( int i=0; i<elemarr.length; i++ ) {
	    handleEvent( elemarr[i] );
	}
	if( DEBUG ) Debug.printtagln( tag, name + " done" );
    }

    /* Specified by mdw.sandStorm.api.EventHandlerIF */
    /**
     * Performs stage initialization common to <code>HMMBuilder</code>s and
     * <code>HMMPredictor</code>.
     *
     * @exception InvalidConfigFileException if the
     * <code>ConfigDataIF</code> contains an invalid parameter or does not
     * contain a required field.
     * @exception InvalidTypeException if a parameter specifies a
     * <code>Class</code> or <code>Field</code> which does not implement
     * the proper java interface.
     */
    public void init(ConfigDataIF config) throws Exception
    {
	final String tag = class_tag+".init";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	name = config.getStage().getName();
	boolean valid_type = true;
	
	////////////////////////////////
	// Initialize model

	String image_name = config.getString( IMAGE_STR );
	if( image_name == null ) {
	    if( DEBUG ) Debug.printtagln( tag, "no snapshot found; " +
					  "initializing new model" );

	    int num_states = config.getInt( M_STR );
	    if( num_states <= 0 ) 
		throw new InvalidConfigFileException( name, M_STR );
	    
	    int num_outputs;
	    if( config.contains( N_STR ) ) {
		num_outputs = config.getInt( N_STR );
		if( num_outputs <= 0 ) 
		    throw new InvalidConfigFileException( name, N_STR );
	    } else {
		// Just guess for now; the model should be fairly robust to errors
		// in this parameter.
		num_outputs = num_states*100;
	    }
	    
	    double epsilon = config.getDouble( THRESHOLD_STR );
	    if( (epsilon <0) || (1<epsilon) )
		throw new InvalidConfigFileException( name, THRESHOLD_STR );
	    
	    model = new HMM( HMM.INIT_CODE_CLUSTERING, num_states,
			     num_outputs, epsilon );
	    
	} else {
	    if( DEBUG ) {
		Debug.printtagln( tag, "reading model from " +
				  "snapshot: \"" + image_name + "\"" );
	    }
	    try {
		model = (HMM)QSIO.read( image_name );
	    } catch( Exception e ) {
		if( DEBUG ) {
		    Carp.carp( "invalid model image" );
		    Carp.carp( e );
		}
		model = null;
	    }
	    if( model == null ) {
		throw new InvalidConfigFileException( name, IMAGE_STR );
	    }
	}

	if( DEBUG ) Debug.printtagln( tag, "Model initialized" );
	
	////////////////////////////////
	// Initialize accessor

	String observed_name = config.getString( EVENT_CLASS_STR );
	if( observed_name == null ) {
	    throw new InvalidConfigFileException( name, EVENT_CLASS_STR );
	}
	try { 
	    observed_type = Class.forName( observed_name );
	    valid_type =
		QueueElementIF_class.isAssignableFrom( observed_type );
	} catch( ClassNotFoundException cnfe ) {
	    valid_type = false;
	}
	if( !valid_type ) {
	    throw new InvalidTypeException( name, EVENT_CLASS_STR,
					    QueueElementIF_class,
					    observed_name );
	}
	
	String field_name = config.getString( EVENT_FIELD_STR );
	if( field_name == null ) {
	    throw new InvalidConfigFileException( name, EVENT_FIELD_STR );
	}
	try { 
	    accessor = observed_type.getField( field_name );
	    valid_type =
		QuickSerializable_class.isAssignableFrom(accessor.getType());
	} catch( NoSuchFieldException nsfe ) {
	    throw new InvalidConfigFileException( name, EVENT_FIELD_STR );
	}
	if( !valid_type ) {
	    throw new InvalidTypeException( name, EVENT_FIELD_STR,
					    QuickSerializable_class,
					    accessor.toString()+":"+
					    accessor.getClass().getName() );
	}
	
	////////////////////////////////
	// Initialize Filters
	
	classifier = Classifier.getClassifier(
                new NodeId (config.getString ("node_id")));
	SinkIF my_sink = config.getStage().getSink();

	// Subscribe to all events we are supposed to observe
	Filter filter = new Filter();
	if( !filter.requireType( observed_type ) ||
	    !classifier.subscribe( filter, my_sink ) )
	    throw new InvalidConfigFileException( "Classifier does not " + 
						  "recognize field: " +
						  observed_type.getName() );
	
	if( DEBUG ) Debug.printtagln( tag, "watching field " + field_name +
				      " of class " + observed_name );

	if( DEBUG ) Debug.printtagln( tag, "done: " + name );	
    }
    
    /* Specified by mdw.sandStorm.api.EventHandlerIF */
    public void destroy()
    {
	final String tag = class_tag+".destroy";
	if( DEBUG ) Debug.printtagln( tag, "destroyed: " + name  );
    }

    ////////////////////     INNER CLASSES	////////////////////

    /**
     * An exception thrown when a config file requests an operation
     * (e.g. emit_predictions) that a particular <code>Builder</code> does not
     * provide.
     *
     * @author	Dennis Geels
     * @version	$Id: HMMStage.java,v 1.7 2001/12/13 20:50:29 geels Exp $
     */
    public static class UnsupportedConfigurationException
	extends InvalidConfigFileException
    {
	/**
	 * Constructs a new UnsupportedConfigurationException
	 *
	 * @param stage		The name of the stage throwing the exception
	 * @param tag		The config file tag for the requested
	 * 			operation
	 * @param value		The requested operation
	 */
	public UnsupportedConfigurationException( String stage, String tag,
						  String value )
	{
	    super( "Stage \"" + stage + "\" does not support operation \""
		   + value + "\" for parameter \"" + tag + "\"" );
	}
    }

    /**
     * An exception thrown when a config file supplies the name of a class or
     * field that does not satisfy type restrictions.
     *
     * @author	Dennis Geels
     * @version	$Id: HMMStage.java,v 1.7 2001/12/13 20:50:29 geels Exp $
     */
    public static class InvalidTypeException
	extends InvalidConfigFileException
    {
	/**
	 * Constructs a new InvalidTypeException
	 *
	 * @param stage		The name of the stage throwing the exception
	 * @param tag		The config file tag
	 * @param required	The <code>Class</code> required that paramater. 
	 * @param supplied	The <code>String</code> from the config file
	 */
	public InvalidTypeException( String stage, String tag,
				     Class required, String supplied )
	{
	    super( "Tag \"" + tag + "\" for Stage \"" + stage +
		   "\" requires a Class of type \"" + required.getName()
		   + "\"; received \"" + supplied + "\"" );
	}
    }
}
