/**
 * SegmentedModel.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.oil;

import ostore.util.QuickSerializable;
import mdw.sandStorm.api.QueueElementIF;

/**
 * An interface for introspective models which participate as part of
 * large, distributed models.
 * In addition to normal <code>Model</code> functionality,
 * <code>SegmentedModel</code>s understand how to swap tagged portions of
 * the model state (<code>Segment</code>s) from a foreign distributor.
 *
 * @see Model
 * @see Segment
 * @see HMM
 *
 * @author	Dennis Geels
 * @version	$Id: SegmentedModel.java,v 1.6 2001/11/28 23:03:15 geels Exp $
 */

public interface SegmentedModel extends Model
{
    ////////////////////       CONSTANTS	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Incorporates the portion of the model contained in the specified
     * <code>Segment</code>.
     *
     * @param s	 The <code>Segment</code> to add.
     * @exception IllegalArgumentException if the <code>Segment</code> object
     * is not of the correct subclass.
     *
     * @see #choose_segments
     */
    void add_segment( Segment s );
    
    /**
     * Selects the most-relevant <code>num</code> segments to retain,
     * discarding all others.
     *
     * @param num	The number of <code>Segment</code>s to keep.
     */
    void choose_segments( int num );	

    ////////////////////     INNER CLASSES	////////////////////

    /**
     * A single, tagged portion of the global <code>Model</code> state.
     * The contents and structure of s <code>Segment</code> may vary
     * widely, depending on the particular <code>Model</code>.
     *
     * @see HMM.Segment
     *
     * @author	Dennis Geels
     * @version	$Id: SegmentedModel.java,v 1.6 2001/11/28 23:03:15 geels Exp $
     */
    interface Segment extends QueueElementIF, QuickSerializable
    {
	/**
	 * Returns the unique ID for this <code>Segment</code>.  This ID
	 * serves as the only key identifying a <code>Segment</code> -
	 * <code>Segment.equals</code> should never be called.
	 * <code>(Segment.id()).equals</code> and
	 * <code>(Segment.id()).hashCode</code>, however, may be called
	 * quite often.  Avoid using <code>QuickSerializable</code> IDs
	 * for which those methods are expensive.
	 *
	 * @return the unique ID for this <code>Segment</code>.
	 */
	QuickSerializable id();
    }
}
