/**
 * PatchworkStage.java
 *
 * @author	Jeremy Stribling
 * @version	$Id: PatchworkStage.java,v 1.3 2002/01/15 23:19:34 strib Exp $
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.oil.patchwork;

import java.util.HashMap;
import ostore.util.InvalidConfigFileException;
import ostore.util.Assert;
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.tapestry.impl.RoutingTable;
import ostore.tapestry.impl.RouteEntry;
import ostore.util.NodeId;
import mdw.sandStorm.api.ConfigDataIF;
import mdw.sandStorm.api.UnknownEventException;
import mdw.sandStorm.api.EventHandlerIF;
import mdw.sandStorm.api.EventHandlerException;
import mdw.sandStorm.api.QueueElementIF;
import mdw.sandStorm.api.SinkIF;
import mdw.sandStorm.api.SinkFullException;   
import mdw.sandStorm.api.StagesInitializedSignal;   
import mdw.sandStorm.core.ssTimer;
import ostore.oil.Model;

/**
 * <p>
 * The <code>PatchworkStage</code> code actively collects network
 * statistics for the purpose of optimizing Tapestry routing.  It keeps
 * track of these statistics in a <code>PatchworkModel</code> object that
 * implements the <code>ostore.oil.Model</code> interface.  Other models
 * that clone this model can be constructed by sending a 
 * <code>PatchworkDeltaRequestMsg</code> to the stage.  The
 * <code>PatchworkDeltaResponseMsg</code> sent back will contain a
 * <code>PatchworkModel.Delta</code> that can be added to the local
 * clone of the model through the model's <code>add_delta</code>
 * function.  The stage will also automatically compute a delta
 * periodically and send out a response message, based on the
 * <code>recalc_period</code> tag in the config file.</p><p>
 *
 * The configuration file for this stage may contain the following
 * tags:</p><p>
 *
 * <code>latency_period</code>: the period of the latency measurements
 * of the stage in in milliseconds, i.e. how often the stage actively collects
 * latency statistics.  If this tag is not present, it defaults to a 
 * value of 5000 milliseconds.  Specify a 0 to never collect
 * latency measurements.</p><p>
 *
 * <code>bandwidth_period</code>: the period of the bandwidth measurements 
 * of the stage in in milliseconds, i.e. how often the stage actively collects
 * bandiwdth statistics.  If this tag is not present, it defaults to a 
 * value of 10000 milliseconds.  Specify a 0 to never collect
 * bandwidth measurements.</p><p>
 *
 * <code>loss_period</code>: the period of the loss measurements
 * of the stage in in milliseconds, i.e. how often the stage actively collects
 * loss statistics.  If this tag is not present, it defaults to a 
 * value of 1000 milliseconds.  Specify a 0 to never collect
 * loss measurements.</p><p>
 *
 * <code>recalc_period</code>: the period of recalculation of the 
 * <code>PatchworkModel</code> of the stage in in milliseconds.  If this 
 * tag is not present, it defaults to a value of 20000 milliseconds.  
 * Specify a 0 to never automatically recalculate the table.</p><p>
 *
 * <code>timeout</code>: the timeout of statistic collection packets in
 * milliseconds.  If no acknowledgement is received within this amount
 * of time, the packet is considered lost.  If this tag is not present,
 * it defaults to a value of 120000 milliseconds.</p><p>
 *
 * <code>optimize_on</code>: set this to <code>true</code> to enable
 * this stage to optimize the routing table.  Currently, this only
 * involves checking for 100% packet loss and switching to a backup
 * node as appropriate.</p><p>
 *
 * To change the periods of the stage during runtime, send out a 
 * <code>PatchworkChangePeriodMsg</code> message, specifying
 * what type of period to change and to what to change it.  Change it
 * to 0 to turn that particular period off.</p><p>
 *
 * Currently, the statistics collected by this stage per node are:</p><p>
 *
 * <b>Latency</b>: the total amount of round-trip delay of sending one 
 * packet to the node, in milliseconds.  Measured by sending one packet at t1,
 * receiving an acknowledgement at t2, and letting latency = t2 - t1.</p><p>
 *
 * <b>Bandwidth</b>: the bottleneck bandwidth between this node and 
 * another node, in bytes/milliseconds (or roughly KBytes/sec).  Measured
 * using Sender-Based Packet Pair where two packets of exactly the same
 * size are sent at the same time and acknowledgement is received at times 
 * s1 and s2, then bandwidth = (size of packet)/(s2 - s1).  Current size 
 * of packet is 1461 bytes and clock is of millisecond accuracy, so largest
 * measurable bandwidth is 1461 bytes/ms.</p><p>
 *
 * <b>Loss</b>: the percentage of packets lost (as determined by the 
 * timeout period) since measurements were last taken.</p><p>
 *
 * @author 	Jeremy Stribling
 * @version	$Id: PatchworkStage.java,v 1.3 2002/01/15 23:19:34 strib Exp $
 */

public class PatchworkStage implements EventHandlerIF
{
    ////////////////////       CONSTANTS	////////////////////
    private boolean DEBUG = DebugFlags.OIL_PATCHWORK_PATCHWORKSTAGE;
    private static final String class_tag = "PatchworkStage"; 

    private static final String TIMEOUT_KEY = "timeout";
    private static final int TIMEOUT_DEFAULT = 120000;
    private static final int M_COUNT_SCALAR = 4;

    private static final String LATENCY_PERIOD_KEY = "latency_period";
    private static final int LATENCY_PERIOD_DEFAULT = 5000;
    private static final String BANDWIDTH_PERIOD_KEY = "bandwidth_period";
    private static final int BANDWIDTH_PERIOD_DEFAULT = 10000;
    private static final String LOSS_PERIOD_KEY = "loss_period";
    private static final int LOSS_PERIOD_DEFAULT = 1000;
    private static final String RECALC_PERIOD_KEY = "recalc_period";
    private static final int RECALC_PERIOD_DEFAULT = 20000;
    private static final String OPTIMIZE_ON_KEY = "optimize_on";

    private static final int BANDWIDTH_MTU = 1461;

    private Classifier classifier;

    private int latency_period;
    private int bandwidth_period;
    private int loss_period;
    private int recalc_period;

    private boolean optimize_on;

    // How long to wait for packets before moving on (in ms)
    private int timeout;

    // from routing table
    private int digits_per_guid;
    private int digit_values;

    // max value of m_count = M_COUNT_SCALAR*digits_per_guid
    private int m_count_max;

    // count from 1 to m_count_max-1 circularly
    private int loss_count;
    private int bandwidth_count;
    private int latency_count;
    private int optimize_count;

    private SinkIF my_sink;
    private static ssTimer my_timer;
    private RoutingTable routing_table;

    // Contain the relevant timestamp in Long format
    // indexed by NodeId.
    private HashMap lastLatencyProbes;
    private HashMap lastBandwidthProbes;
    private HashMap lastBandwidthResps;
    private HashMap lastLossProbes;

    // Patchwork observation model
    private PatchworkModel model;

    ////////////////////  STATIC CONSTRUCTORS	////////////////////

    static {
	my_timer = new ssTimer();

	Debug.turnOn( class_tag + ".init" );
	Debug.turnOn( class_tag + ".dispatch" );
	Debug.turnOn( class_tag + ".handleEvent" );
	Debug.turnOn( class_tag + ".handleAlarmMsg" );
	Debug.turnOn( class_tag + ".handleDeltaRequest" );
	Debug.turnOn( class_tag + ".handleRoutingTableAvailable" );
	Debug.turnOn( class_tag + ".handleProbe" );
	Debug.turnOn( class_tag + ".handleOptimize" );

    }

    ////////////////////      CONSTRUCTORS	////////////////////
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Specified by mdw.sandStorm.api.EventHandlerIF 
     */
    public void handleEvents(QueueElementIF elemarr[]) 
	throws EventHandlerException
    {

	// make sure all events are handled at the same time
	long timeReceived = System.currentTimeMillis();
	
	final String tag = class_tag+".handleEvents";
	if( DEBUG ) Debug.printtagln( tag, "called: "
				      + elemarr.length + " elements" );
	for( int i=0; i<elemarr.length; i++ ) {
	    handleEvent( elemarr[i], timeReceived );
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );
    }

    /** 
     * Specified by mdw.sandStorm.api.EventHandlerIF 
     */
    public void handleEvent(QueueElementIF element) 
	throws EventHandlerException
    {

	handleEvent( element, System.currentTimeMillis() );

    }

    /**
     * Handle events given the exact time they were received by this stage
     */
    private void handleEvent( QueueElementIF element, long timeReceived ) {

	final String tag = class_tag+".handleEvent";
	if( DEBUG ) Debug.printtagln( tag, "called" );

        if (DEBUG) Debug.printtagln (tag, "got " + element.getClass () +
                " event at " + timeReceived + "ms: " + element + ".");

	if( element instanceof StagesInitializedSignal ) {

	  // set first alarms only if they're to be probed
	  if( latency_period != 0 ) {
	    setAlarm( PatchworkAlarmMsg.LATENCY, latency_period );
	  }
	  if( bandwidth_period != 0 ) {
	    setAlarm( PatchworkAlarmMsg.BANDWIDTH, bandwidth_period );
	  }
	  if( loss_period != 0 ) {
	    setAlarm( PatchworkAlarmMsg.LOSS, loss_period );
	  }
	  if( recalc_period != 0 ) {
	    setAlarm( PatchworkAlarmMsg.RECALC_MODEL, recalc_period );
	  }
 
	} else if( element instanceof PatchworkAlarmMsg ) {

	  handleAlarmMsg( (PatchworkAlarmMsg) element, timeReceived );

        } else if( element instanceof RoutingTable.Available ) {
	
	  handleRoutingTableAvailable( (RoutingTable.Available) element );
	
	} else if( element instanceof PatchworkProbeMsg ) {
	
	  handleProbe( (PatchworkProbeMsg) element, timeReceived );

	} else if( element instanceof PatchworkDeltaRequestMsg ) {

	  handleDeltaRequest();

	} else if( element instanceof PatchworkChangePeriodMsg ) {

	  handleChangePeriod( (PatchworkChangePeriodMsg) element );

	} else {

	  Carp.error( "Dropping unexpected message: " + element );

	}

    }	

    /**
     * It's time to take some measurements!
     */
    private void handleAlarmMsg( PatchworkAlarmMsg msg, long timeReceived ) {

	final String tag = class_tag+".handleAlarmMsg";
	if (DEBUG) Debug.printtagln( tag, " called with type " + msg.type );

	int m_count;
	int m_period;
	HashMap measureMap;

	// find the right period
	if( msg.type == PatchworkAlarmMsg.LATENCY ) {
	  m_period = latency_period;
	} else if( msg.type == PatchworkAlarmMsg.BANDWIDTH ) {
	  m_period = bandwidth_period;
	} else {
	  m_period = loss_period;
	}

	// no routing table yet
	if( routing_table == null ) {
	  if( DEBUG ) Debug.printtagln( tag, "No routing table yet" );
	  setAlarm( msg.type, m_period );
	  return;
	}

	// if it's a recalc message, then recalculate
	if( msg.type == PatchworkAlarmMsg.RECALC_MODEL ) {
	  handleDeltaRequest();
	
	  // if we want to optimize, now's the time
	  if( optimize_on ) {
	    handleOptimize();
	  }

	  setAlarm( msg.type, recalc_period );
	  return;
	}

	// find the right count and increase (think circularly . . .)
	if( msg.type == PatchworkAlarmMsg.LATENCY ) {
	  m_count = latency_count;
	  latency_count++;
	  if( latency_count == m_count_max ) {
	    latency_count = 1;
 	  }
	  measureMap = lastLatencyProbes;
	} else if( msg.type == PatchworkAlarmMsg.BANDWIDTH ) {
	  m_count = bandwidth_count;
	  bandwidth_count++;
	  if( bandwidth_count == m_count_max ) {
	    bandwidth_count = 1;
 	  }
	  measureMap = lastBandwidthProbes;
	} else {
	  m_count = loss_count;
	  loss_count++;
	  if( loss_count == m_count_max ) {
	    loss_count = 1;
 	  }
	  measureMap = lastLossProbes;
	}

	// measure every peer in routing table level i if:
        //    m_count is divisible by (i+1)
	//    that peer has no outstanding pings	
        for( int i = 0; i < digits_per_guid; i++ ) {

	  if( m_count % (i+1) == 0 ) {

	    if( DEBUG ) Debug.printtagln( tag, "scanning table level " + 
					       i );

	    NodeId currpeer;
	    RouteEntry entry;
	    int startIndex = 0;
	    for( int j = 0; j < digit_values; j++ ) {

	      entry = routing_table.getEntry( i, j );
	      if( entry == null ) {
		continue;
	      }

	      // if we're optimizing, then we should check the
	      // the depths of the routing table
	      if( optimize_on ) {
		currpeer = entry.getAtIndex(startIndex);
		startIndex++;
		if( startIndex == entry.size() ) {
		  // start at beginning of next digit
		  startIndex = 0;
		} else {
		  // stay on the same digit
		  j--;
		}
	      } else {
		currpeer = entry.getFirst();
	      }

	      Long lastMeasureTime = (Long) measureMap.get( currpeer );

	      // check for timeouts if not null
	      if( currpeer != null ) {
		
		  // if the response got in, record one no-loss,
		  // if not and time expired then record a loss
		  if( lastMeasureTime == null ) {

		    model.record( new PatchworkModel.Observation( currpeer, 0,
							 PatchworkModel.LOSS ));
		    if( DEBUG ) Debug.printtagln( tag, "Recorded a no-loss"
						+ " for " + currpeer ); 

		  } else if( (timeReceived - lastMeasureTime.longValue()) 
			     >= timeout ) {

		    model.record( new PatchworkModel.Observation( currpeer, 100,
							 PatchworkModel.LOSS ));
		    measureMap.put( currpeer, null );
		    lastMeasureTime = null;
		    if( DEBUG ) Debug.printtagln( tag, "Recorded a no-loss"
						+ " for " + currpeer );

		  }

		  // more stuff to check if bandwidth
		  if( msg.type == PatchworkAlarmMsg.BANDWIDTH && 
		      lastMeasureTime == null ) {

		    Long firstResp = (Long) lastBandwidthResps.get(currpeer);
	
		    // if second one received, record no-loss
		    if( firstResp == null ) {

		      model.record( new PatchworkModel.Observation( currpeer, 0,
							 PatchworkModel.LOSS ));
		      if( DEBUG ) Debug.printtagln( tag, "Recorded a no-loss"
						+ " for " + currpeer ); 

		    } else if( (timeReceived - firstResp.longValue()) 
			     >= timeout ) {

		      model.record( new PatchworkModel.Observation( currpeer, 
						100, PatchworkModel.LOSS ));
		      lastBandwidthResps.put( currpeer, null );
		      if( DEBUG ) Debug.printtagln( tag, "Recorded a no-loss"
						+ " for " + currpeer );

		    }

		  }

	      }

	      // send out messages if:
	      //   non-null peer
              //   last measurement came in
	      //   if bandwidth, both responses came in
	      if( (currpeer != null) && (lastMeasureTime == null) &&
		  ( (msg.type != PatchworkAlarmMsg.BANDWIDTH) ||
		    (lastBandwidthResps.get(currpeer) == null) ) ) {

		if( DEBUG ) Debug.printtagln( tag, "m_count = " + m_count +
						   ": pinging nodeID " + 
						   currpeer + " at digit " +
						   i + ", value " + j + ".");

		// probe the peer
		PatchworkProbeMsg ppm;
		PatchworkProbeMsg ppm2 = null;
		if( msg.type == PatchworkAlarmMsg.LATENCY ) {
		  ppm = new PatchworkProbeMsg( currpeer, 
					       PatchworkProbeMsg.MINIMUM_SIZE,
					       PatchworkProbeMsg.LATENCY, 
					       false );
		} else if( msg.type == PatchworkAlarmMsg.BANDWIDTH ) {
		  ppm = new PatchworkProbeMsg( currpeer, 
					       BANDWIDTH_MTU,
					       PatchworkProbeMsg.BANDWIDTH_1,
					       false );
		  ppm2 = new PatchworkProbeMsg( currpeer, 
					       BANDWIDTH_MTU,
					       PatchworkProbeMsg.BANDWIDTH_2,
					       false );
		} else {
		  ppm = new PatchworkProbeMsg( currpeer, 
					       PatchworkProbeMsg.MINIMUM_SIZE,
					       PatchworkProbeMsg.LOSS,
					       false );
		}
		
		measureMap.put( currpeer, 
				   new Long( System.currentTimeMillis() ) );
		dispatch( ppm );
		if( ppm2 != null ) {
		  dispatch( ppm2 );
		}

	      }

	    }

	  }

        }

	setAlarm( msg.type, m_period );

    }

    /**
     * Save a reference to the routing table
     */
    private void handleRoutingTableAvailable( RoutingTable.Available msg ) {

	final String tag = class_tag+".handleRoutingTableAvailable";

	if( DEBUG ) { 
	  Debug.printtagln( tag, "called" );
	  Assert.assert (routing_table == null, tag + ": got two versions " +
                "of the routing table.");
	}

	routing_table = msg.table;

	lastLatencyProbes = new HashMap();
	lastBandwidthProbes = new HashMap();
	lastLossProbes = new HashMap();
	lastBandwidthResps = new HashMap();

	digits_per_guid = routing_table.digits_per_guid();
	digit_values = routing_table.digit_values();
	m_count_max = M_COUNT_SCALAR*digits_per_guid;

	loss_count = 1;
	bandwidth_count = 1;
	latency_count = 1;
	optimize_count = 1;

	model = new PatchworkModel();

    }	

    /**
     * Handle a resp
     */
    private void handleProbe( PatchworkProbeMsg msg, long timeReceived ) {

	final String tag = class_tag+".handleProbe";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	HashMap measureMap;

	// no routing table yet
	if( routing_table == null ) {
	  if( DEBUG ) Debug.printtagln( tag, "No routing table yet" );
	  return;
	}

	// look at the right map to figure out measurement
	if( msg.type == PatchworkProbeMsg.LATENCY ) {
	  measureMap = lastLatencyProbes;
	} else if( msg.type == PatchworkProbeMsg.BANDWIDTH_1 ) {
	  measureMap = lastBandwidthProbes;
	} else if( msg.type == PatchworkProbeMsg.BANDWIDTH_2 ) {
	  measureMap = lastBandwidthResps;
	} else {
	  measureMap = lastLossProbes;
	}

	Long measureFrom = (Long) measureMap.get( msg.peer );

	// if resp was unexpected, then drop
	if( measureFrom == null ) {
	  if( DEBUG ) Debug.printtagln( tag, "Dropping unexpected resp from "
					     + msg.peer );
	  // if it's bandwidth #2, stop looking for bandwidth #1
	  if( msg.type == PatchworkProbeMsg.BANDWIDTH_2 ) {
	    lastBandwidthProbes.put( msg.peer, null );
	  }
	  return;
	}

	// make the measurement
	if( msg.type == PatchworkProbeMsg.LATENCY ) {

	  // calculate latency

	  long latency = timeReceived - measureFrom.longValue();
	  if( DEBUG ) Debug.printtagln( tag, "Latency to " + msg.peer + 
					     " is " + latency );
	  model.record( new PatchworkModel.Observation( msg.peer, 
						    (double) latency, 
					            PatchworkModel.LATENCY ) );

	} else if( msg.type == PatchworkProbeMsg.BANDWIDTH_1 ) {

	  // get ready for next message	
	  lastBandwidthResps.put( msg.peer, new Long( timeReceived ) );
	  if( DEBUG ) Debug.printtagln( tag, "Received bandwidth #1 for " +
					     msg.peer ); 

	} else if( msg.type == PatchworkProbeMsg.BANDWIDTH_2 ) {

	  // calculate bandwidth

	  long lpTime = measureFrom.longValue();

	  // if times are equal
	  if( lpTime == timeReceived ) {
	    if( DEBUG ) Debug.printtagln( tag, "Ignoring time compression" );
	    measureMap.put( msg.peer, null );
	    return;
	  } 

	  // otherwise calculate the bandwidth
	  double bandwidth = (BANDWIDTH_MTU/(timeReceived - lpTime ));

	  if( DEBUG ) Debug.printtagln( tag, "Bandwidth to " + msg.peer + 
					     " is " + bandwidth );
	  model.record( new PatchworkModel.Observation( msg.peer, bandwidth, 
					         PatchworkModel.BANDWIDTH ) );

	} else {

	  // nothing to calculate for loss

	}		

	// ack this resp
	measureMap.put( msg.peer, null );

    }

    /** 
     * Send back a delta to the requestor 
     */
    private void handleDeltaRequest() {

	final String tag = class_tag+".handleDeltaRequest";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	PatchworkModel.Delta d = (PatchworkModel.Delta) model.recalculate();

	if( DEBUG ) {
	  Debug.printtagln( tag, "Recalculating model" );
	  Debug.printtagln( tag, "Current model:\n" + model );
	  Debug.printtagln( tag, "Current delta:\n" + d );
	}

	dispatch( new PatchworkDeltaResponseMsg( d ) );

    }

    /** 
     * Change one of the calculation periods
     */
    private void handleChangePeriod( PatchworkChangePeriodMsg msg ) {

	final String tag = class_tag+".handleChangePeriod";
	if( DEBUG ) Debug.printtagln( tag, "called with msg " + msg );

	// set the right period
	if( msg.type == PatchworkChangePeriodMsg.LATENCY ) {
	  latency_period = msg.newPeriod;
	} else if( msg.type == PatchworkChangePeriodMsg.BANDWIDTH ) {
	  bandwidth_period = msg.newPeriod;
	} else if( msg.type == PatchworkChangePeriodMsg.LOSS ) {
	  loss_period = msg.newPeriod;
	} else {
	  recalc_period = msg.newPeriod;
	}

    }

    /**
     * Tries to optimize the routing table based on the model.
     * Right now, just looks at losses.
     */
     private void handleOptimize() {

	final String tag = class_tag+".handleOptimize";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	// no routing table yet
	if( routing_table == null ) {
	  if( DEBUG ) Debug.printtagln( tag, "No routing table yet" );
	  return;
	}

	// optimize every peer in routing table level i if:
        //    m_count is divisible by (i+1)
        for( int i = 0; i < digits_per_guid; i++ ) {

	  if( optimize_count % (i+1) == 0 ) {

	    if( DEBUG ) Debug.printtagln( tag, "optimizing table level " + 
					       i );

	    NodeId currpeer;
	    RouteEntry entry;
	    double loss;
	    for( int j = 0; j < digit_values; j++ ) {

	      entry = routing_table.getEntry( i, j );
	      if( entry == null ) {
		continue;
	      }

	      loss = model.lookupDouble( entry.getFirst(), 
					 PatchworkModel.LOSS );
	      if( Double.isNaN( loss ) ) {
		// no calculations have been made yet, so skip
		continue;
	      }

	      // if loss is 100%, try to switch it with a backup
	      if( loss == 100.0 ) {

		int entrySize = entry.size();
		double otherLoss;
		for( int k = 1; k < entry.size(); k++ ) {
		  otherLoss = model.lookupDouble( entry.getAtIndex( k ),
						  PatchworkModel.LOSS );
		  if( !Double.isNaN( otherLoss ) && (otherLoss < 100.0) ) {
		    // this entry is better!  switch!
		    entry.switchRoutes( 0, k );
		    if( DEBUG ) Debug.printtagln( tag, "in the entry (" + i 
						       + ", " + j + 
						       "), switching in index " 
						       + k );
		    break;
		  }
		}

  	      }

	    }  

	  }

	}

	// increase optimize_count (think circularly . . .)
	optimize_count++;
	if(optimize_count == m_count_max ) {
	 optimize_count = 1;
 	}

     }

    /**
     * Sets this stage's alarm clock for <code>millis</code> milliseconds
     * from now.
     */	
    private void setAlarm( int type, long millis ) {

	final String tag = class_tag+".setAlarm";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	classifier.dispatch_later( new PatchworkAlarmMsg( type ), millis );

    }

    /**
     * An internal method used for dispatching events
     */
    private void dispatch( QueueElementIF event ) {
                                                
	final String tag = class_tag+".dispatch";
	if( DEBUG ) Debug.printtagln( tag, "called with event " + event );

      try {
        classifier.dispatch( event );
      } catch ( SinkFullException sfe ) {
        //TODO: Better way to deal with SinkFull
        try {
          Thread.currentThread().sleep( 5000 );
        } catch( InterruptedException ie ) {
          //ignore
        }
      }                                         
      
    }


    /** 
     * Specified by mdw.sandStorm.api.EventHandlerIF 
     */
    public void init(ConfigDataIF config) throws Exception
    {
	final String tag = class_tag+".init";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	///////////////////
	// get config info

	latency_period = config.getInt( LATENCY_PERIOD_KEY );
	if( latency_period == -1 ) {
	  latency_period = LATENCY_PERIOD_DEFAULT;
	} else if( latency_period < 0 ) {
	  throw new InvalidConfigFileException( "The " + LATENCY_PERIOD_KEY +
						"tag must be >= 0");
 	}

	if( DEBUG ) Debug.printtagln( tag, "latency_period is " + 
					   latency_period );

	bandwidth_period = config.getInt( BANDWIDTH_PERIOD_KEY );
	if( bandwidth_period == -1 ) {
	  bandwidth_period = BANDWIDTH_PERIOD_DEFAULT;
	} else if( bandwidth_period < 0 ) {
	  throw new InvalidConfigFileException( "The " + BANDWIDTH_PERIOD_KEY +
						"tag must be >= 0");
 	}

	if( DEBUG ) Debug.printtagln( tag, "bandwidth_period is " + 
					   bandwidth_period );

	loss_period = config.getInt( LOSS_PERIOD_KEY );
	if( loss_period == -1 ) {
	  loss_period = LOSS_PERIOD_DEFAULT;
	} else if( loss_period < 0 ) {
	  throw new InvalidConfigFileException( "The " + LOSS_PERIOD_KEY +
						"tag must be >= 0");
 	}

	if( DEBUG ) Debug.printtagln( tag, "loss_period is " + loss_period );

	recalc_period = config.getInt( RECALC_PERIOD_KEY );
	if( recalc_period == -1 ) {
	  recalc_period = RECALC_PERIOD_DEFAULT;
	} else if( recalc_period < 0 ) {
	  throw new InvalidConfigFileException( "The " + RECALC_PERIOD_KEY +
						"tag must be >= 0");
 	}

	if( DEBUG ) Debug.printtagln( tag, "recalc_period is " + 
					   recalc_period );

	timeout = config.getInt( TIMEOUT_KEY );
	if( timeout == -1 ) {
	  timeout = TIMEOUT_DEFAULT;
	} else if( timeout <= 0 ) {
	  throw new InvalidConfigFileException( "The " + TIMEOUT_KEY +
						"tag must be greater than 0");
 	}

	if( DEBUG ) Debug.printtagln( tag, "Timeout is " + timeout );

	optimize_on = config.getBoolean( OPTIMIZE_ON_KEY );
	if( DEBUG ) Debug.printtagln( tag, "Optimize_on is " + optimize_on );
	

	//////////////////
	// initialize filters

	classifier = Classifier.getClassifier( 
			new NodeId( config.getString( "node_id" ) ) );
	my_sink = config.getStage().getSink();
	Filter filter;

	// Subscribe to all my favorite events

	String[] eventNames = {
	    "mdw.sandStorm.api.StagesInitializedSignal",	
	    "ostore.tapestry.impl.RoutingTable$Available",
	    "ostore.oil.patchwork.PatchworkDeltaRequestMsg",
	    "ostore.oil.patchwork.PatchworkChangePeriodMsg",
	    "ostore.oil.patchwork.PatchworkAlarmMsg"
	};

	String[] inb_eventNames = {
	};

	// require probe responses
	try {
	  filter = new Filter();
	  Class probeClass = Class.forName( 
				"ostore.oil.patchwork.PatchworkProbeMsg" );
	  if( DEBUG ) Debug.printtagln( tag, "Requiring probe class" );

	    if( !filter.requireType( probeClass ) ||
		!filter.requireValue ( "inbound", new Boolean(true) ) || 
		!filter.requireValue ( "response", new Boolean(true) ) || 
	        !classifier.subscribe( filter, my_sink ) ) {

	      Carp.die( "Could not require probe class " ); 

	    } 
        } catch ( ClassNotFoundException cnfe ) {
            Carp.die( cnfe );
        }

	for( int i = 0; i < eventNames.length; i++ ) {
	  filter = new Filter();
	  Class eventClass;
          if( DEBUG ) Debug.printtagln( tag, "Finding class: " +
                                               eventNames[i] );
          try {
            eventClass = Class.forName( eventNames[i] );
	    if( DEBUG ) Debug.printtagln( tag, "Requiring class: " +
	  				       eventNames[i] );
	    if( !filter.requireType( eventClass ) || 
	        !classifier.subscribe( filter, my_sink ) ) {

	      Carp.die( "Could not require class " + eventNames[i] ); 

	    } 

          } catch ( ClassNotFoundException cnfe ) {
            Carp.die( cnfe );
          }

        }

	// require inbounders
	for( int i = 0; i < inb_eventNames.length; i++ ) {
	  filter = new Filter();
	  Class eventClass;
          if( DEBUG ) Debug.printtagln( tag, "Finding class: " +
                                               inb_eventNames[i] );
          try {
            eventClass = Class.forName( inb_eventNames[i] );
	    if( DEBUG ) Debug.printtagln( tag, "Requiring class: " +
	  				       inb_eventNames[i] );
	    if( !filter.requireType( eventClass ) ||
		!filter.requireValue ( "inbound", new Boolean(true) ) || 
	        !classifier.subscribe( filter, my_sink ) ) {

	      Carp.die( "Could not require class " + inb_eventNames[i] ); 

	    } 

          } catch ( ClassNotFoundException cnfe ) {
            Carp.die( cnfe );
          }

        }

    }

    /** 
     * Specified by mdw.sandStorm.api.EventHandlerIF 
     */
    public void destroy()
    {
        final String tag = class_tag+".destroy";
        if( DEBUG ) Debug.printtagln( tag, "destroyed" );
    }


    ////////////////////     INNER CLASSES	////////////////////

}
