/**
 * GroupKey.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.security.thresh;
import java.math.BigInteger;
import java.security.*;

/**
 * A GroupKey with information common to all the keyshares.
 * Generated by the Dealer class.
 *
 * <p>TODO: Investigate the security of reusing the key parameters (q,p)
 * with a new Poly, since they are computationally expensive
 *
 * <p>Reference: "Practical Threshold Signatures", Victor Shoup
 * (sho@zurich.ibm.com), IBM Research Paper RZ3121, 4/30/99
 *
 * @author <a href=mailto:sweis@theory.lcs.mit.edu>Steve Weis</a> 
 * @version $Id: GroupKey.java,v 1.4 2001/10/17 01:50:40 srhea Exp $
 */
class GroupKey {
   
    /** Keeps track of initialization state */
    private boolean initialized;

    /** An arbitrary security parameter for generating secret shares */
    private final static int L1 = 128;
    private int keysize;
   

    /** Group Verifier */
    private BigInteger vk=null;

    /** (k,l) Parameters. k out of l shares needed for a signature */
    private int k, l;

    /** Randomly generated polynomial used to generate shares */
    private Poly poly;

    /** The group keypair containing special primes */
    private KeyPair groupKeyPair=null;
   
    /** The exponent of the groupKeyPair */
    private BigInteger e;

    /** The RSA modulus of the groupKeyPair */
    private BigInteger n;

    // Constructor does nothing. Each parameter must be individually set.
    public GroupKey() {
	initialized = false;
    }

    public ThresholdPublicKey getPublicKey () {
	return new ThresholdPublicKey (k, l, n, e);
    }

    public void initialize(int k, int l, int keysize, BigInteger v,
			   BigInteger e, BigInteger n,
			   KeyPair groupKeyPair) {
	this.k = k;
	this.l = l;
	this.vk = v;
	this.e = e;
	this.n = n;
	this.groupKeyPair = groupKeyPair;
	this.keysize = keysize;
	initialized = true;
    }


    /**
     * Returns the minimum threshold size
     */
    public int getK()
	throws ThresholdSigException {
	if (initialized = false)
	    throw new ThresholdSigException("Group key not initialized");
	return this.k;
    }

    /**
     * Returns the group size
     */
    public int getL()
	throws ThresholdSigException {
	if (initialized = false)
	    throw new ThresholdSigException("Group key not initialized");
	return this.l;
    }


    /**
     * Returns the group key pair
     */
    public KeyPair getKeyPair()
	throws ThresholdSigException {
	if (initialized = false)
	    throw new ThresholdSigException("Group key not initialized");
	return this.groupKeyPair;
    }

    /**
     * Returns the group key modulus
     */
    public BigInteger getModulus()
	throws ThresholdSigException {
	if (initialized = false)
	    throw new ThresholdSigException("Group key not initialized");
	return this.n;
    }

    /**
     * Returns the group key exponent
     */
    public BigInteger getExponent()
	throws ThresholdSigException {
	if (initialized = false)
	    throw new ThresholdSigException("Group key not initialized");
	return this.e;
    }
   
}
