/**
 * KeyShare.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.security.thresh;
import java.math.BigInteger;

import java.security.*;
import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import cryptix.jce.provider.random.DevRandom;
import ostore.util.Assert;
import ostore.util.QSException;
import ostore.util.Digestible;
import ostore.util.Assert;
import ostore.util.ByteUtils;
import ostore.util.SecureHashBuffer;
import ostore.util.Types;
import ostore.util.SHA1Hash;

/**
 * A Secret Key Share for an RSA (k,l) Threshold Scheme.
 *
 * <p>Reference: "Practical Threshold Signatures",
 * Victor Shoup (sho@zurich.ibm.com), IBM Research Paper RZ3121, 4/30/99
 *
 * @author  <a href=mailto:sweis@theory.lcs.mit.edu>Steve Weis</a> 
 * @version $Id: KeyShare.java,v 1.6 2002/04/06 22:52:43 eaton Exp $
 */
public class KeyShare implements Digestible {

    private static final String class_tag = "KeyShare";

    static {
	final String tag = class_tag + ".static";
	Provider cryptixRandom = new cryptix.jce.provider.CryptixRandom ();
	Security.addProvider (cryptixRandom);
	Assert.assert (DevRandom.isAvailable (), tag + ": need DevRandom.");
    }

    public KeyShare (byte [] data, int [] offset) 
    throws QSException {
	secret = ByteUtils.bytesToBigInteger (data, offset);
	verifier = ByteUtils.bytesToBigInteger (data, offset);
	groupVerifier = ByteUtils.bytesToBigInteger (data, offset);
	n = ByteUtils.bytesToBigInteger (data, offset);
	delta = ByteUtils.bytesToBigInteger (data, offset);
	signVal = ByteUtils.bytesToBigInteger (data, offset);
	/*
	int length = ByteUtils.bytesToInt (data, offset);
	ByteArrayInputStream is = 
	    new ByteArrayInputStream (data, offset [0], length);
	try {
	    secret = BI.fromStream(is);
	    verifier = BI.fromStream(is);
	    groupVerifier = BI.fromStream(is);
	    n = BI.fromStream(is);
	    delta = BI.fromStream(is);
	    signVal = BI.fromStream(is);
	}
	catch (IOException e) {
	    throw new QSException ("Bad encoding.");
	}
	offset [0] += length;
	*/
	id = ByteUtils.bytesToInt (data, offset);
    }

    public void to_bytes (byte [] data, int [] offset) {
	final String tag = class_tag + ".to_bytes";

	ByteUtils.bigIntegerToBytes (secret, data, offset);
	ByteUtils.bigIntegerToBytes (verifier, data, offset);
	ByteUtils.bigIntegerToBytes (groupVerifier, data, offset);
	ByteUtils.bigIntegerToBytes (n, data, offset);
	ByteUtils.bigIntegerToBytes (delta, data, offset);
	ByteUtils.bigIntegerToBytes (signVal, data, offset);
	/*
	ByteArrayOutputStream os = new ByteArrayOutputStream ();
	try {
	    BI.toStream (secret, os);
	    BI.toStream (verifier, os);
	    BI.toStream (groupVerifier, os);
	    BI.toStream (n, os);
	    BI.toStream (delta, os);
	    BI.toStream (signVal, os);
	}
	catch (IOException e) {
	    Assert.assert (false, tag + ": could not encode BigInteger.");
	}
	byte [] tmp = os.toByteArray ();
	ByteUtils.intToBytes (tmp.length, data, offset);
	offset [0] += ByteUtils.memcpy (data, offset [0], tmp, 0, tmp.length);
	*/
	ByteUtils.intToBytes (id, data, offset);
    }

    public int type_code ()
    {
        throw new NoSuchMethodError();
        //return Types.SECURITY_THRESH_KEY_SHARE;
    } 

    public void add_to_digest(SecureHashBuffer buffer)
    {
	buffer.add (secret);
	buffer.add (verifier);
	buffer.add (groupVerifier);
	buffer.add (n);
	buffer.add (delta);
	buffer.add (signVal);
        buffer.add(id);
    }

// Constants and variables
//............................................................................
    private final static BigInteger TWO = BigInteger.valueOf(2l);
    private final static BigInteger FOUR = BigInteger.valueOf(4l);

    /** An arbitrary security parameter for generating secret shares */
    private final static int L1 = 128;

    /** Secret key value */
    private BigInteger secret;

    /** Verifier used to authenticate self to other shares */
    private BigInteger verifier;
    private BigInteger groupVerifier;

    // TODO: This information should be moved to a GroupKey object
    // It's redundant to put into every share
    private BigInteger n;
    private final BigInteger delta;

    /** The secret key value used to sign messages. Should not be exposed */
    private BigInteger signVal;

    // TODO: Maybe get a better id scheme. Use verifier
    private int id;

// Constructors
//............................................................................
   
    /**
     * Create a new share
     *
     * @param id - the identifier of this share
     * @param secret - a secret value generated by a Dealer

     * @param n - the modulo of the group public key
     * @param delta - l! (group size factorial)
     */
    public KeyShare(int id, BigInteger secret,
		    BigInteger n, BigInteger delta)
    {
	this.id = id;
	this.secret = secret;

	// Verifier must be set later using setVerifier()
	// TODO: Merge Dealer.generateShares and Dealer.generateVerifiers
	// and generate them simultaneously
	this.verifier = null;

	this.n = n;
	this.delta = delta;
	this.signVal = FOUR.multiply(delta).multiply(secret);
    }

// Public Methods
//............................................................................

    public int getId() { return id; }

    public BigInteger getSecret() { return secret; }


    public void setVerifiers(BigInteger verifier, BigInteger groupVerifier) {
	this.verifier = verifier;
	this.groupVerifier = groupVerifier;
    }

    public BigInteger getVerifier() { return verifier; }

    public BigInteger getSignVal() { return signVal; }

   
    public String toString() {
	return "KeyShare[" + id + "]";
    }

    // public SigShare sign(BigInteger x) {
    //	return new SigShare(this.id, x.modPow(signVal,n));
    //    }

    /**
     * Create a SigShare and a Verifier for byte[] b<BR>
     *
     * Refer to Shoup pg. 8 <BR>
     * @return a sig share with a verifier
     */
    public SigShare sign(byte [] data) {
	
	byte [] b = (new SHA1Hash (data)).bytes ();
	final String tag = class_tag + ".sign";
	
	SecureRandom random = null;
	try {
	    random = SecureRandom.getInstance ("DevRandom", "CryptixRandom");
	    Assert.assert (random != null, tag + ": need DevRandom.");
	}
	catch (Exception e) {
	    Assert.assert (random != null, tag + ": caught " + e + 
		    " while trying to init DevRandom.");
	}
    
	BigInteger x = (new BigInteger(b)).mod(n);

	int randbits = n.bitLength()+3*L1;

	// r \elt (0, 2^L(n)+3*l1)
	BigInteger r = (new BigInteger(randbits,random));
	BigInteger vprime = groupVerifier.modPow(r,n);
	BigInteger xtilde = x.modPow(FOUR.multiply(delta),n);
	BigInteger xprime = xtilde.modPow(r,n);


	BigInteger c = null;
	BigInteger z = null;
	// Try to generate C and Z
	try {
	    MessageDigest md = MessageDigest.getInstance("SHA");
	    md.reset();

	    //	    debug("v: " + groupVerifier.mod(n));
	    md.update(groupVerifier.mod(n).toByteArray());
	   
	    //	    debug("xtilde: " + xtilde);
	    md.update(xtilde.toByteArray());

	    //	    debug("vi: " + verifier.mod(n));
	    md.update(verifier.mod(n).toByteArray());
	   
	    //	    debug("xi^2: " + x.modPow(signVal,n).modPow(TWO,n));
	    md.update(x.modPow(signVal,n).modPow(TWO,n).toByteArray());

	    //	    debug("v': "+ vprime);
	    md.update(vprime.toByteArray());
	   
	    //	    debug("x': " + xprime);
	    md.update(xprime.toByteArray());
	    c = new BigInteger(md.digest()).mod(n);
	    z = (c.multiply(secret)).add(r);
	} catch (Exception e) { e.printStackTrace(); }

	Verifier ver = new Verifier(this.id,z,c,verifier,groupVerifier);

	return new SigShare(this.id, x.modPow(signVal,n), ver);
    }
   
// Debugging
//............................................................................

    private final static boolean DEBUG = true;

    private static void debug(String s)
    {
	System.err.println("KeyShare: " + s);
    }

}



