/**
 * Poly.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.security.thresh;
import java.math.BigInteger;
import java.util.Random;
import java.security.Security;
import java.security.Provider;
import java.security.SecureRandom;
import cryptix.jce.provider.random.DevRandom;
import ostore.util.Assert;

/**
 * Polynomial handling code.
 *
 * <p>Reference: "Practical Threshold Signatures",
 * Victor Shoup (sho@zurich.ibm.com), IBM Research Paper RZ3121, 4/30/99
 *
 * @author  <a href=mailto:sweis@theory.lcs.mit.edu>Steve Weis</a> 
 * @version $Id: Poly.java,v 1.3 2001/10/17 01:50:40 srhea Exp $
 */
class Poly {
    private static void debug(String s)
    {
	System.err.println("Poly: " + s);
    }
    private BigInteger[] coeff;
    private int size;
    private static SecureRandom random;

    static {
	final String tag = "Poly.static";

	byte[] randSeed = new byte[20];
	(new Random()).nextBytes(randSeed);

	Provider cryptixRandom = new cryptix.jce.provider.CryptixRandom ();
	Security.addProvider (cryptixRandom);
	Assert.assert (DevRandom.isAvailable (), tag + ": need DevRandom.");
	try {
	random = SecureRandom.getInstance ("DevRandom", "CryptixRandom");
	Assert.assert (random != null, tag + ": need DevRandom.");
	}
	catch (Exception e) {
	Assert.assert (random != null, tag + ": caught " + e + 
		" while trying to init DevRandom.");
	}
    }
   
    /**
     * Generates a polynomial, f(x) mod m such that f(0) = d<BR
     * All coefficients are chosen randomly from Z(m) (modulo m)
     *
     * @param d - The first coeffecient of the polynomial
     * @param size - The size of the polynomial
     * @param m - modulo of field coeffiecients are from
     */

    public Poly(BigInteger d, int size, BigInteger m) {
	int bitLength = m.bitLength();

	this.size = size;
	coeff = new BigInteger[size];
	coeff[0] = d;
	for (int i=1; i<size; i++)
	    coeff[i] = (new BigInteger(bitLength,random)).mod(m);

    }
   
    public BigInteger eval(BigInteger x) {
	BigInteger retVal = coeff[size-1];

	for (int i = size-2; i>=0; i--) {
	    //	    debug(retVal.toString(10));
	    retVal = (retVal.multiply(x)).add(coeff[i]);
	}
	return retVal;
    }

    public BigInteger eval(int x) {
	BigInteger bx = BigInteger.valueOf((long)x);
	return this.eval(bx);
    }

    public String toString() {
        String out = "[";
	for (int i=0; i<this.size; i++)
	    out += coeff[i].toString() + " ";
	out += "]";
	return out;
    }
   
}
