/**
 * SigShare.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.security.thresh;
import java.math.BigInteger;
import java.security.MessageDigest;
import ostore.util.Digestible;
import ostore.util.QSException;
import ostore.util.ByteUtils;
import ostore.util.Assert;
import ostore.util.SecureHashBuffer;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

/**
 * Signature Shares Class.
 * Associates a signature share with an id and wraps a static verifier.
 *
 * <p>Reference: "Practical Threshold Signatures",
 * Victor Shoup (sho@zurich.ibm.com), IBM Research Paper RZ3121, 4/30/99
 *
 * @author  <a href=mailto:sweis@theory.lcs.mit.edu>Steve Weis</a> 
 * @version $Id: SigShare.java,v 1.7 2002/04/06 22:52:43 eaton Exp $
 */
public class SigShare implements Digestible {
   
// Constants and variables
//............................................................................
   
    private int id;
    private BigInteger sig;
    private Verifier sigVerifier;

    private static final String class_tag = "SigShare";


// Constructors
//............................................................................

    public SigShare(int id, BigInteger sig, Verifier sigVerifier) {
	this.id = id;
	this.sig = sig;
	this.sigVerifier = sigVerifier;
    }
   
    public SigShare(int id, byte[] sig) {
	this.id = id;
	this.sig = new BigInteger(sig);
	this.sigVerifier = sigVerifier;
    }

    public SigShare (byte [] data, int [] offset) throws QSException {
	id = ByteUtils.bytesToInt (data, offset);
	sig = ByteUtils.bytesToBigInteger (data, offset);
	/*
	int length = ByteUtils.bytesToInt (data, offset);
	ByteArrayInputStream is = 
	    new ByteArrayInputStream (data, offset [0], length);
	try {
	    sig = BI.fromStream(is);
	}
	catch (IOException e) {
	    throw new QSException ("Bad encoding.");
	}
	offset [0] += length;
	*/
	sigVerifier = new Verifier (data, offset);
    }

    public void to_bytes (byte [] data, int [] offset) {
	final String tag = class_tag + ".to_bytes";

	ByteUtils.intToBytes (id, data, offset);
	ByteUtils.bigIntegerToBytes (sig, data, offset);
	/*
	ByteArrayOutputStream os = new ByteArrayOutputStream ();
	try {
	    BI.toStream (sig, os);
	}
	catch (IOException e) {
	    Assert.assert (false, tag + ": could not encode BigInteger.");
	}
	byte [] tmp = os.toByteArray ();
	ByteUtils.intToBytes (tmp.length, data, offset);
	offset [0] += ByteUtils.memcpy (data, offset [0], tmp, 0, tmp.length);
	*/
	sigVerifier.to_bytes (data, offset);
    }

    public int type_code () 
    {
        throw new NoSuchMethodError();
        //return 0; /* don't intend to use this */
    }

    public void add_to_digest(SecureHashBuffer buffer)
    {
	buffer.add (id);
	buffer.add (sig);
	buffer.add(sigVerifier);
    }

    public static byte [] createSignature (SigShare [] sigs) {
	byte [] result = null;
	while (true) {
	    int [] offset = {0};
	    ByteUtils.intToBytes (sigs.length, result, offset);
	    for (int i = 0; i < sigs.length; ++i)
		sigs [i].to_bytes (result, offset);
	    if (result == null)
		result = new byte [offset [0]];
	    else
		break;
	}
	return result;
    }

// Public Methods
//............................................................................
  
    /**
     * Return this share's id. Needed for Lagrange interpolation
     *
     * @return the id of this key share
     */
    public int getId() { return id; }

    /**
     * Return a BigInteger representation of this signature
     *
     * @return a BigInteger representation of this signature
     */
    public BigInteger getSig() { return sig; }

    /**
     * Return this signature's verifier
     *
     * @return A verifier for this signaute
     */
    public Verifier getSigVerifier() { return sigVerifier; }

    /**
     * Return a byte array representation of this signature
     *
     * @return a byte array representation of this signature
     */
    public byte[] getBytes() { return sig.toByteArray(); }

    public String toString() {
	return "Sig["+id+"]: " + sig.toString();
    }

    private static final BigInteger TWO = BigInteger.valueOf(2l);
    private static final BigInteger FOUR = BigInteger.valueOf(4l);

    public boolean verify (byte [] data, ThresholdPublicKey public_key) {
	BigInteger n = public_key.n;
	BigInteger x = (new BigInteger(data)).mod(n);	   
	int l = public_key.l;
	BigInteger delta = SigShare.factorial(l);
	BigInteger xtilde = x.modPow(FOUR.multiply(delta),n);

	try {
	    MessageDigest md = MessageDigest.getInstance("SHA");
	    md.reset();
	    Verifier ver = getSigVerifier();
	    BigInteger v = ver.getGroupVerifier();
	    BigInteger vi = ver.getShareVerifier();
	    md.update(v.toByteArray());
	    md.update(xtilde.toByteArray());
	    md.update(vi.toByteArray());
	    BigInteger xi = getSig();
	    md.update(xi.modPow(TWO,n).toByteArray());
	    BigInteger vz = v.modPow(ver.getZ(),n);
	    BigInteger vinegc = vi.modPow(ver.getC(),n).modInverse(n);
	    md.update(vz.multiply(vinegc).mod(n).toByteArray());

	    BigInteger xtildez = xtilde.modPow(ver.getZ(),n);

	    // TODO: CHECK PAPER
	    BigInteger xineg2c =
		xi.modPow(ver.getC(),n).modInverse(n);
	    // According to Shoup, pg. 8 this should be
	    // xi.modPow(TWO,n).modPow(ver.getC(),n).modInverse(n);

	    // Something to do with working in Q_n since every
	    // element is a square

	    md.update(xineg2c.multiply(xtildez).mod(n).toByteArray());
	    BigInteger result = new BigInteger(md.digest()).mod(n);

	    return result.equals(ver.getC());
	}
	catch (Exception ex) {
	    debug("Encountered error while attempting to check verifiers");
	    ex.printStackTrace();
	}
	    
	return false;
    }

// Static methods
//............................................................................

    public static boolean verify(byte[] data, SigShare[] sigs, 
	ThresholdPublicKey public_key) throws ThresholdSigException
	    /* int k, int l, BigInteger n, BigInteger e*/
    {
	int k = public_key.k;
	int l = public_key.l;
	BigInteger n = public_key.n;
	BigInteger e = public_key.e;

	// Sanity Check - make sure there are at least k unique sigs out of l
	// possible

	boolean[] haveSig = new boolean[l];
	for (int i=0; i<k; i++) {
	    //	    debug("Checking sig " + sigs[i].getId());
	    if (sigs[i] == null)
		throw new ThresholdSigException("Null signature");
	    if (haveSig[sigs[i].getId()-1])
		throw new ThresholdSigException("Duplicate signature: " +
						sigs[i].getId());
	    haveSig[sigs[i].getId()-1] = true;
	}

	BigInteger x = (new BigInteger(data)).mod(n);	   
	BigInteger delta = SigShare.factorial(l);

	// This code is somewhat new.  The idea is that first we just
	// assume the shares are valid and try to verify the signature.  If
	// it fails, we then go look at each of the shares to see it one of
	// them was bogus.  --srhea

	boolean returnme = false;
	boolean CHECKVERIFIER = false;
	while (true) {

	// Test the verifier of each signature to ensure there are
	// no dummy sigs thrown in to corrupt the batch
	if (CHECKVERIFIER) {
	    BigInteger FOUR = BigInteger.valueOf(4l);
	    BigInteger TWO = BigInteger.valueOf(2l);
	    BigInteger xtilde = x.modPow(FOUR.multiply(delta),n);

	    // TODO: currently, if one SigShare is bad, this entire
	    // procedure will fail, even if there exist sufficient good
	    // SigShares to still verify the signature.  We should possibly
	    // fix this in the future, although I imagine that anyone
	    // constructing one of these signatures will make sure all
	    // shares they include in it are good to start with.  --srhea
	   
	    try {
	
		for (int i = 0; i<k; i++) {
		    MessageDigest md = MessageDigest.getInstance("SHA");
		    md.reset();
		    Verifier ver = sigs[i].getSigVerifier();
		    BigInteger v = ver.getGroupVerifier();
		    BigInteger vi = ver.getShareVerifier();

		    //debug("v :" + v);
		    md.update(v.toByteArray());

		    //		    debug("xtilde :" + xtilde);
		    md.update(xtilde.toByteArray());

		    //		    debug("vi :" + vi);
		    md.update(vi.toByteArray());

		    BigInteger xi = sigs[i].getSig();
		    //		    debug("xi^2 :" + xi.modPow(TWO,n));
		    md.update(xi.modPow(TWO,n).toByteArray());

		    BigInteger vz = v.modPow(ver.getZ(),n);

		    BigInteger vinegc = vi.modPow(ver.getC(),n).modInverse(n);
		    //		    debug("v^z*v^-c :" + vz.multiply(vinegc).mod(n));
		    md.update(vz.multiply(vinegc).mod(n).toByteArray());

		    BigInteger xtildez = xtilde.modPow(ver.getZ(),n);

		    // TODO: CHECK PAPER
		    BigInteger xineg2c =
			xi.modPow(ver.getC(),n).modInverse(n);
		    // According to Shoup, pg. 8 this should be
		    // xi.modPow(TWO,n).modPow(ver.getC(),n).modInverse(n);

		    // Something to do with working in Q_n since every
		    // element is a square
		   
		    //		    debug("xi^-2cx: " + xineg2c.multiply(xtildez).mod(n));
		    md.update(xineg2c.multiply(xtildez).mod(n).toByteArray());
		    BigInteger result = new BigInteger(md.digest()).mod(n);
		   
		    if (!result.equals(ver.getC())) {
			//			debug("Share verifier is OK");
			//		    else
			debug("Share verifier is not OK");
			return false;
		    }
		   
		}
	    }
	    catch (Exception ex) {
		debug("Encountered error while attempting to check verifiers");
		ex.printStackTrace();
	    }
	}

	BigInteger w = BigInteger.valueOf(1l);


	for (int i=0; i<k; i++)
	    w = w.multiply(sigs[i].getSig().
			   modPow(SigShare.lambda(sigs[i].getId(),
						 sigs, delta), n));

	// eprime = delta^2*
	BigInteger eprime = delta.multiply(delta).shiftLeft(2);

	w = w.mod(n);
	BigInteger xeprime = x.modPow(eprime,n);
	BigInteger we = w.modPow(e,n);
	returnme = (xeprime.compareTo(we) == 0);

	if (returnme || CHECKVERIFIER)
	    break;
	else
	    CHECKVERIFIER = true;
	}

	return returnme;
    }

    /**
     * Returns the factorial of the given integer as a BigInteger
     *
     * @return l
     */
    private static BigInteger factorial(int l) {
	BigInteger x = BigInteger.valueOf(1l);
	for (int i=1; i<=l; i++)
	    x = x.multiply(BigInteger.valueOf((long)i));

	return x;
    }
   

    /**
     * Compute lagarange interpolation points
     * Reference: Shoup, pg 7.
     *
     * @param p - a polynomial to evaluate these points on
     * @param ik - a point in S
     * @param S - a set of k points in {0...l}
     * @param delta - the factorial of the group size
     *
     * @return the Lagarange interpolation of these points at 
     */
    private static BigInteger lambda(int ik, SigShare[] S,
				    BigInteger delta) {
	// lambda(id,l) = PI {id!=j, 0<j<=l} (i-j')/(id-j')
	BigInteger value = delta;

	for (int i=0; i<S.length; i++)
	    if (S[i].getId() != ik)
		value = value.multiply(BigInteger.valueOf((long)S[i].getId()));

	for (int i=0; i<S.length; i++)
	    if (S[i].getId() != ik)
		value =
		    value.divide(BigInteger.valueOf((long)(S[i].getId()-ik)));
	   
	return value;
    }



// Debugging
//............................................................................
   
    private final static boolean DEBUG = true;
   
    private static void debug(String s)
    {
	System.err.println("SigShare: " + s);
    }

}



