/**
 * ThresholdPublicKey.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.security.thresh;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.security.PublicKey;
import ostore.security.QSPublicKey;
import ostore.util.QSException;
import ostore.util.ByteUtils;
import ostore.util.Assert;
import ostore.util.Types;
import ostore.util.SHA1Hash;
import ostore.util.SecureHashBuffer;

/**
 * A public key class for threshold signatures.
 *
 * @author  Sean C. Rhea
 * @version $Id: ThresholdPublicKey.java,v 1.6 2002/03/29 01:59:58 srhea Exp $
 */
public class ThresholdPublicKey implements PublicKey, QSPublicKey {

    public String toString () {
	return "(ThresholdPublicKey " + new SHA1Hash (this) + ")";
    }

    private final String class_tag = "ThresholdPublicKey";

    ThresholdPublicKey (int k, int l, BigInteger n, BigInteger e) {
	this.k = k;
	this.l = l;
	this.n = n;
	this.e = e;
    }

    ///////////////// QuickSerializable Interface ////////////////////

    public ThresholdPublicKey (byte [] data, int [] offset) 
    throws QSException {
	k = ByteUtils.bytesToInt (data, offset);
	l = ByteUtils.bytesToInt (data, offset);
	n = ByteUtils.bytesToBigInteger (data, offset);
	e = ByteUtils.bytesToBigInteger (data, offset);
	/*
	int length = ByteUtils.bytesToInt (data, offset);
	ByteArrayInputStream is = 
	    new ByteArrayInputStream (data, offset [0], length);
	try {
	    n = BI.fromStream(is);
	    e = BI.fromStream(is);
	}
	catch (IOException e) {
	    throw new QSException ("Bad encoding.");
	}
	offset [0] += length;
	*/
    }

    public void to_bytes (byte [] data, int [] offset) {
	final String tag = class_tag + ".to_bytes";

	ByteUtils.intToBytes (k, data, offset);
	ByteUtils.intToBytes (l, data, offset);
	ByteUtils.bigIntegerToBytes (n, data, offset);
	ByteUtils.bigIntegerToBytes (e, data, offset);
	/*
	ByteArrayOutputStream os = new ByteArrayOutputStream ();
	try {
	    BI.toStream (n, os);
	    BI.toStream (e, os);
	}
	catch (IOException e) {
	    Assert.assert (false, tag + ": could not encode BigInteger.");
	}
	byte [] tmp = os.toByteArray ();
	ByteUtils.intToBytes (tmp.length, data, offset);
	offset [0] += ByteUtils.memcpy (data, offset [0], tmp, 0, tmp.length);
	*/
    }

    public void add_to_digest( SecureHashBuffer buffer ) {
	buffer.add (k);
	buffer.add (l);
	buffer.add (n);
	buffer.add (e);
    }

    public int type_code () { return Types.SECURITY_THRESH_PUBLIC_KEY; } 

    ///////////////////// QSPublicKey Interface //////////////////////

    public PublicKey key () { return this; }

    /////////////////////////// Key Interface ////////////////////////

    public String getAlgorithm () { return "Threshold"; }
    public String getFormat () { return "Default"; }

    public byte [] getEncoded () {
	int [] offset = {0};
	to_bytes (null, offset);
	byte [] result = new byte [offset [0]];
	offset [0] = 0;
	to_bytes (result, offset);
	return result;
    }

    /////////////////////////// Private Parts ////////////////////////

    int k, l;
    BigInteger n, e; 

}

