/**
 * Verifier.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.security.thresh;
import java.math.BigInteger;
import ostore.util.Digestible;
import ostore.util.QSException;
import ostore.util.ByteUtils;
import ostore.util.Assert;
import ostore.util.SecureHashBuffer;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

/**
 * Verifier for signature shares.
 *
 * <p>Reference: "Practical Threshold Signatures",
 * Victor Shoup (sho@zurich.ibm.com), IBM Research Paper RZ3121, 4/30/99
 *
 * @author  <a href=mailto:sweis@theory.lcs.mit.edu>Steve Weis</a> 
 * @version $Id: Verifier.java,v 1.6 2002/04/06 22:52:43 eaton Exp $
 */
class Verifier implements Digestible {

    private static final String class_tag = "Verifier";

    private BigInteger z;
    private BigInteger c;
    private BigInteger groupVerifier;
    private BigInteger shareVerifier;

    private int id;

    public Verifier(int id, BigInteger z, BigInteger c,
		    BigInteger shareVerifier,
		    BigInteger groupVerifier) {
	this.id = id;
	this.z = z;
	this.c = c;
	this.shareVerifier = shareVerifier;
	this.groupVerifier = groupVerifier;
    }

    public BigInteger getZ() {
	return this.z;
    }

    public BigInteger getShareVerifier() {
	return this.shareVerifier;
    }
    public BigInteger getGroupVerifier() {
	return this.groupVerifier;
    }

    public BigInteger getC() {
	return this.c;
    }

    public Verifier (byte [] data, int [] offset) throws QSException {
	id = ByteUtils.bytesToInt (data, offset);
	z = ByteUtils.bytesToBigInteger (data, offset);
	c = ByteUtils.bytesToBigInteger (data, offset);
	groupVerifier = ByteUtils.bytesToBigInteger (data, offset);
	shareVerifier = ByteUtils.bytesToBigInteger (data, offset);
	/*
	int length = ByteUtils.bytesToInt (data, offset);
	ByteArrayInputStream is = 
	    new ByteArrayInputStream (data, offset [0], length);
	try {
	    z = BI.fromStream(is);
	    c = BI.fromStream(is);
	    groupVerifier = BI.fromStream(is);
	    shareVerifier = BI.fromStream(is);
	}
	catch (IOException e) {
	    throw new QSException ("Bad encoding.");
	}
	offset [0] += length;
	*/
    }

    public void to_bytes (byte [] data, int [] offset) {
	final String tag = class_tag + ".to_bytes";

	ByteUtils.intToBytes (id, data, offset);
	ByteUtils.bigIntegerToBytes (z, data, offset);
	ByteUtils.bigIntegerToBytes (c, data, offset);
	ByteUtils.bigIntegerToBytes (groupVerifier, data, offset);
	ByteUtils.bigIntegerToBytes (shareVerifier, data, offset);
	/*
	ByteArrayOutputStream os = new ByteArrayOutputStream ();
	try {
	    BI.toStream (z, os);
	    BI.toStream (c, os);
	    BI.toStream (groupVerifier, os);
	    BI.toStream (shareVerifier, os);
	}
	catch (IOException e) {
	    Assert.assert (false, tag + ": could not encode BigInteger.");
	}
	byte [] tmp = os.toByteArray ();
	ByteUtils.intToBytes (tmp.length, data, offset);
	offset [0] += ByteUtils.memcpy (data, offset [0], tmp, 0, tmp.length);
	*/
    }

    public void add_to_digest(SecureHashBuffer buffer)
    {
	buffer.add (id);
	buffer.add (z);
	buffer.add (c);
	buffer.add (groupVerifier);
	buffer.add (shareVerifier);
    }

    public int type_code () 
    {
        throw new NoSuchMethodError();
        //return 0; /* don't intend to use this */
    } 
}

