/*
 * RoutingTable.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.tapestry.impl;
import java.util.HashMap;
import java.util.Map;
import ostore.util.Assert;
import ostore.util.DebugFlags;
import ostore.util.Debug;
import ostore.util.NodeId;
import ostore.util.SecureHash;
import mdw.sandStorm.api.QueueElementIF;

/**
 * Implements the routing table interface.
 *
 * @author     Ben Y. Zhao
 * @version    $Id: RoutingTable.java,v 1.7 2002/03/27 00:08:07 ravenben Exp $
 */
public class RoutingTable {

    public static class Available implements QueueElementIF {
    	public Available (RoutingTable table) {
	    this.table = table;
	}

	public RoutingTable table;
    }

    private static final String class_tag = "RoutingTable";
    private static final boolean DEBUG = DebugFlags.TAPESTRY_ROUTER;

    public static final int BITS_PER_DIGIT = 4;
    public static final int DIGITS_PER_GUID = 40;
    public static final int DIGIT_VALUES = 16;

    public RoutingTable (NodeId self_node_id) {
	Debug.turnOn (class_tag);

	_node_map = new HashMap ();
	_routing_table = new RouteEntry [DIGITS_PER_GUID][];
	for (int digit = 0; digit < DIGITS_PER_GUID; ++digit) {
	    _routing_table [digit] = new RouteEntry [DIGIT_VALUES];
	}
	_self_node_id = self_node_id;
	_nodeid_distance = new HashMap();
    }

    public int digit_values    () { return DIGIT_VALUES; } 
    public int digits_per_guid () { return DIGITS_PER_GUID; } 
    public int bits_per_digit  () { return BITS_PER_DIGIT; } 

    synchronized public SecureHash node_id_to_guid (NodeId node_id) {
	return (SecureHash) _node_map.get (node_id);
    }

    synchronized public NodeId read (int which_digit, int digit_value) {
       RouteEntry re = _routing_table [which_digit][digit_value];
       if (re !=null) return re.getFirst();
       return null;
    }

    synchronized public void write (int which_digit, int digit_value, 
	    NodeId node_id, SecureHash node_guid, Long latency) {
      if (DEBUG) System.err.println("Writing nodeid="+node_id+
				    " node_guid="+node_guid);

      // System.out.println("Writing nodeid="+node_id+" node_guid="+node_guid
      //	 +" digit="+which_digit+" value="+digit_value+" "
      //			 +this.toString());
      RouteEntry re = _routing_table [which_digit][digit_value];
      if (re == null) {
	_routing_table[which_digit][digit_value] = 
	  new RouteEntry(node_id, latency);
	//	System.out.println(" new value"+_routing_table[which_digit][digit_value]+"\n");
      } else {
	re.add(node_id, latency);
      }

      _node_map.put (node_id, node_guid);
      //MARK
      //      System.out.println("Done writing "+this.toString());
    }

  /**
   * @param newguid The guid of some node 
   * @param nid The nodeid of that node
   * @param delta is the estimated RTT distance from us to that node
   */
  synchronized public void optimize(long delta, NodeId nid,
				    SecureHash newguid) {
    int [] selfguidint = digits((SecureHash) _node_map.get(_self_node_id));
    int [] newguidint = digits(newguid);

    int i;
    for (i = 0; selfguidint[i] == newguidint[i] && i<DIGITS_PER_GUID; i++) {
    }

    //now at the level where newguid can be looked at.
    RouteEntry re = _routing_table[i][newguidint[i]];
    if (re != null) {
      NodeId oldnode = re.getFirst();
      Long oldlatency = (Long) _nodeid_distance.get(oldnode);
      if (delta <= oldlatency.longValue()) {
	if (DEBUG) System.err.println("Modifying entry at digit="+
				      i+", value="+newguidint[i]);
	write(i, newguidint[i], nid, newguid, new Long(delta));
	_nodeid_distance.put(nid, new Long(delta));
      } else if (DEBUG)
	System.err.println("Compared value, but no change made.");
    } else {
      write(i, newguidint[i], nid, newguid, new Long(delta));
      _nodeid_distance.put(nid, new Long(delta));
      if (DEBUG) 
	System.err.println("Optimize replacing null entry with nid="+
			   nid+", time="+delta+", digit="+i+", v="+
			   newguidint[i]);
    }
  }

  /**
   * Add alternate/backup routes to the routing table
   * @param digit digit of entry being added
   * @param value value of entry's Guid at the relevant digit
   * @param nid NodeId of entry being added
   * @param latency Last tested latency of new entry
   * @param guid Guid of new entry
   */
  synchronized public void addBackup(int digit, int value, NodeId nid,
				     Long latency, SecureHash guid) {
    _routing_table[digit][value].add(nid, latency);
    _nodeid_distance.put(nid, latency);
    _node_map.put(nid, guid);
  }

  synchronized public RouteEntry getEntry(int digit, int value) {
    return _routing_table[digit][value];
  }

  /**
   * Method to add the latency measurement to the data structures
   * @param rtt Estimated roundtrip distance
   * @param nid NodeId of new node
   * @param newguid Guid of new node
   */
  synchronized public void addDistance(long rtt, NodeId nid,
				       SecureHash newguid) {
    if (DEBUG) 
      System.err.println("adding distance for nodeId="+nid);
    _nodeid_distance.put(nid, new Long(rtt));
    if (_node_map.get(nid) == null) _node_map.put(nid, newguid);
  }

    /**
     * Find the surrogate for an empty routing table entry.
     *
     * @param dest_digits The array of ints representing the destination guid
     * @param dest_digit  The digit value that points to an empty entry
     * @param hop         Which hop are we looking for the surrogate route on
     *
     * @return            <code>null</code> if we are <i>definitely</i> the
     *                    surrogate, and the <code>NodeId</code> of the
     *                    next person to ask if we cannot tell from this
     *                    hop yet.
     **/
    synchronized public NodeId surrogate (int [] dest_digits, 
	    int dest_digit, int hop) {
	final String tag = class_tag + ".surrogate";

	boolean empty = true;
	if (DEBUG) System.err.println("Surrogate called on table: "+
				      toString());
	for (int next_hop = hop; next_hop < DIGITS_PER_GUID; ++next_hop) {
	  if (DEBUG) Debug.printtagln (tag, "checking level " +
				       next_hop + ".");
	  for (int i = 0; i < DIGIT_VALUES; ++i) {
	    if ((_routing_table [next_hop][i] != null) && 
		((RouteEntry) (_routing_table [next_hop][i])).getFirst() 
		!= _self_node_id) {
	      if (DEBUG) Debug.printtagln (tag, "neighbor " +
			 _routing_table [next_hop][i] + " present.");
	      empty = false;
	      break;
	    }
	  }
	  if (! empty) break;
	}

	if (empty) {
	    if (DEBUG) Debug.printtagln (tag, "we are definitely the root.");
	    return null;
	}

	// Surrogate routing to next higher non-null value (circularly).
	int i = (dest_digit == DIGIT_VALUES - 1) ? 0 : dest_digit;

	while (_routing_table [hop][i] == null)
	    i = (i == DIGIT_VALUES - 1) ? 0 : i + 1;

	if (DEBUG) Debug.printtagln (tag, "surrogate is " + 
		_routing_table [hop][i]);

	return ((RouteEntry) _routing_table [hop][i]).getFirst();
    }

  public void addSelf (SecureHash self) {
    if (DEBUG) System.err.println("Writing entries for self into map.\n");

    int [] digits = RoutingTable.digits(self);
    for (int i=0; i<digits_per_guid(); i++) {
      write(i, digits[i], _self_node_id, self, new Long(0));
    }
  }

    public static int [] digits (SecureHash h) {
	final String tag = class_tag + ".digits";
	int [] offset = {0};
	h.to_bytes (null, offset);
	byte [] buf = new byte [offset [0]];
	offset [0] = 0;
	h.to_bytes (buf, offset);

	// Completely assume that we use four bit digits.

	Assert.assert (BITS_PER_DIGIT == 4, tag + ": Ack!");
	Assert.assert (offset [0] * 2 == DIGITS_PER_GUID, tag + ": Ack!");

	int [] result = new int [offset [0] * 2];
	int i = offset [0] * 2 - 1;
	while (i >= 0) {
	    result [i]   =  (int) (buf [i/2] & 0xf);
	    result [i-1] = ((int) (buf [i/2] >>> 4)) & 0xf;
	    i -= 2;
	}

	return result;
    }

  public String toString() {
    String val ="RoutingTable for "+_self_node_id+"\n";
    for (int i=0; i<3; i++) {
      for (int j=0; j<DIGIT_VALUES; j++) {
	if (_routing_table[i][j] == null)
	  val+=" -";
	else 
	  val+=" -["+_routing_table[i][j].size()+"]"+
	    printGuid((SecureHash) 
		      _node_map.get(_routing_table[i][j].getFirst()));
      }
      val+="\n";
    }
    return val;
  }

  public String printGuid(SecureHash x) {
    String out = "";
    int [] digits = digits(x);
    for (int i=0; i<digits.length; i++) {
      out += Integer.toHexString(digits[i]);
    }
    return out+"/"+x;
  }

    private Map _node_map;        //NodeId to Guid mapping
    private RouteEntry [] [] _routing_table;
    private NodeId _self_node_id;
    private Map _nodeid_distance; //mapping nodeids to last latency measured
                                  //but only for entries in the routingtable
}

