/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;

/**
 * QuickSerializable is an interface that should be implemented by any
 * OceanStore object intended to be sent "over the wire", stored to disk,
 * or otherwise leave the happy confines of the JVM.  It interacts with the
 * {@link TypeTable} and {@link Types} classes to encode and decode Java
 * objects to/from byte arrays.
 * 
 * <p>Note that in addition to the interface defined here, all subtypes of
 * <code>QuickSerializable</code> should define a constructor of one of the
 * following two forms:
 * 
 * <pre>Foo (byte [] data, int [] offset)</pre>
 * or
 * <pre>Foo (byte [] data, int [] offset) throws QSException</pre>
 *
 * for use in decoding.  The meanings of <code>data</code> and
 * <code>offset</code> for this constructor are the same as in
 * {@link #to_bytes}, below.  The second form should be used if there
 * is the possiblity that the decoding process can detect a bad encoding,
 * in which case a {@link QSException} should be thrown.
 * 
 * @author   Sean C. Rhea
 * @version  "$Id: QuickSerializable.java,v 1.6 2001/10/18 00:20:03 eaton Exp $"
 * 
 * @see ostore.util.TypeTable
 * @see ostore.util.Types
 * @see ostore.util.QSException
 */
public interface QuickSerializable {
    /**
     * Store the object in its "on the wire" form in the byte array
     * <code>data</code>, starting at index <code>offset [0]</code>, and
     * increment <code>offset [0]</code> by the number of bytes written;
     * if <code>data == null</code>, increment <code>offset [0]</code> by
     * the number of bytes that would have been written otherwise.
     *
     * @param	data	The byte array to store into, or <code>null</code>.
     * @param	offset	A single element array whose first
     *                  element is the index in data to begin writing at on
     *                  function entry, and which on function exit has been
     *                  incremented by the number of bytes written.
     */
    public void to_bytes (byte [] data, int [] offset);

    /**
     * Returns the unique integer associated with this class in the
     * <code>TypeTable</code>.
     * 
     * @return		an integer unique to this class
     */
    public int type_code ();
}

