/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;
import java.security.*;

/**
 * SHA1Hash is the hash class used by all entities which are SHA1 hashes.
 * (In the prototype, this means Guids and VHashes).
 *
 * @author     Chris WElls
 * @version    $Id: SHA1Hash.java,v 1.23 2002/04/06 22:52:44 eaton Exp $
 */
public class SHA1Hash implements SecureHash {

	public static final int TYPE_CODE = Types.SHA1_HASH;
        private final static boolean VERIFY_DIGESTIBLE_AGAINST_QS = false;

	/**
	 * The number of bytes all <code>SHA1Hash</code>s are composed of
	 */
	public static final int NUM_BYTES = 20;

	/**
	 * This is the value of the null hash for SHA1
	 */
	public static final byte[] NULL_HASH_VALUE =
		{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

	/**
	 * This is the null hash for SHA1
	 */
	public static final SHA1Hash NULL_HASH = new SHA1Hash();

	/**
	 * These are the bytes in this Hash.
	 */
	protected byte _bytes[];

	/**
	 * default constructor (creates null hash)
	 */
	public SHA1Hash() {
		_bytes = new byte[NUM_BYTES];
	}

	/**
	 * copy constructor (creates new hash with value of old hash)
	 * @param shash == old hash whose value we are to copy
	 */
	public SHA1Hash(SHA1Hash shash) {
		_bytes = new byte[NUM_BYTES];
		for (int i = 0; i < NUM_BYTES; ++i) {
			_bytes[i] = shash._bytes[i];
		}
	}

	/**
	 * hashes the given buffer to create a new <code>SHA1Hash</code>
	 * @param the array of bytes over which to hash
	 */
	public SHA1Hash(byte[] bytes) {
		hash(bytes);
	}

	/**
	 * hashes the given buffer to create a new <code>SHA1Hash</code>
	 * @param the array of bytes over which to hash
	 * @param offset in bytes at which to start hashing
	 * @param number of bytes over which to hash
	 */
	public SHA1Hash(byte[] bytes, int offset, int len) {
		hash(bytes, offset, len);
	}

	/**
	 * Reads the bytes of the value of this hash in from the input byte
	 * array.
	 * @param array out of which to read the bytes of this hash
	 * @param offset in data at which to begin
	 * reading. <code>offset[0]</code> is incremented by
	 * <code>NUM_BYTES</code> on completion of this constructor.  */
	public SHA1Hash (byte [] data, int [] offset) {
		_bytes = new byte[NUM_BYTES];
		for (int i = 0; i < NUM_BYTES; ++i) {
			_bytes[i] = data [i+offset [0]];
		}
		offset [0] += NUM_BYTES;
	}

	/**
	 * hashes over the bytes of the input String
	 * @param string over which to hash
	 */
	public SHA1Hash(String s) {
		hash(s);
	}

	/**
	 * hashes over the input object
	 * @param object over which to hash
	 */
	public SHA1Hash (Digestible d) {
            //  Use the first line if you want to hash objects using
            //  the old QS procedure with excessive memory allocation
            //  and all of its other warts.  Use the second line to
            //  hash objects with the new, improved, super-speedy
            //  Digestible method.
            //  hash(qs_to_bytes((QuickSerializable) d));
            hash(d);
	}

	/**
	 * Computes and returns the digest of the specified buffer.
	 * @param buffer	A <code>SecureHashBuffer</code>.
	 */
	public SHA1Hash (SHA1HashBuffer buffer) {
	    _bytes = buffer.compute_bytes();	    
	}

	/**
	 * Create an exact copy of this hash
	 * @return new SHA1Hash identical in value to this one
	 * @exception CloneNotSupportedException if clone() is not supported
	 */
        public Object clone() throws CloneNotSupportedException {
		return super.clone();
	}

	/**
	 * This routine compares this <code>SHA1Hash</code> to another
	 * <code>Object</code>
	 * @param o 	SHA1Hash
	 * @return -1 if <code>this < o</code>, 0 if <code>this == o</code>,
	 * 	or 1 if <code>this > o</code>. 
	 * @exception ClassCastException if <code>o</code> is not a SHA1Hash
	 */
	public int compareTo( Object o ) throws ClassCastException {
	    if( o == null ) {
		throw new NullPointerException();
	    } else if( !(o instanceof SHA1Hash) ) {
		throw new ClassCastException( "cannot compareTo non-SHA1Hash" );
	    } else if( o == this ) {
		return 0;
	    } else {
		SHA1Hash other = (SHA1Hash)o;
		// Compare each successive byte of the two hashes
		for (int i = 0; i < NUM_BYTES; ++i) {
			if( _bytes[i] < other._bytes[i] )
				return -1;
			else if( _bytes[i] > other._bytes[i] )
				return 1;
		}
		return 0;
	    }
	}

	/**
	 * Write the bytes of this hash to the input byte array
	 * @param array into which to write the bytes of this hash
	 * @param offset in data at which to begin writing
	 * @return number of bytes written (always NUM_BYTES)
	 */
	public void to_bytes (byte [] data, int [] offset) {
		if (data != null) {
		    for (int i = 0; i < NUM_BYTES; ++i) {
			data[i+offset [0]] = _bytes[i];
		    }
		}
		offset [0] += NUM_BYTES;
	}

	/**
	 * Return the type code of a SHA1Hash
	 * @return the type code defined at the top of this class
	 */
	public int type_code () {
		return TYPE_CODE;
	}

	/**
	 * Return the number of bytes used in the representation of this hash
	 * (for example, SHA1 uses 20 bytes)
	 * @return number of bytes used in the implementation
	 */
        public int size() {
		return NUM_BYTES;
	}

	/**
	 * Determines if this SecureHash is a NULL hash or not.
	 * @return true iff this Hash has a NULL value
	 */
	public boolean isNull() {
		for (int i = 0; i < NUM_BYTES; ++i) {
			if (_bytes[i] != NULL_HASH_VALUE[i])
				return false;
		}
		return true;
	}

	/**
	 * Returns the implementation's NULL hash
	 * (typically all zeroes).
	 * @return the implementation's null hash
	 */
	public SecureHash nullHash() {
		return NULL_HASH;
	}

	/**
	 * Returns the bytes contained in this hash
	 * @return the bytes of this hash
	 */
	public byte[] bytes() {
		return _bytes;
	}

	/**
	 * Compares this SecureHash with another hash, and returns
	 * true iff they are the same type of hash and have the
	 * same byte representation
	 * @param the object with which to compare this SecureHash
	 * @return true iff this and o contain the same byte values
	 */
	public boolean equals(Object other) {
	    boolean equal;
	    if (other == null) {
		equal = false;
	    } else if (other == this) {
		equal = true;
	    } else if (!(other instanceof SHA1Hash)) {
		equal = false;
	    } else {
		SHA1Hash rhs = (SHA1Hash)other;
		equal = true;
		
		for (int i = 0; i < NUM_BYTES; ++i) {
		    if (_bytes[i] != rhs._bytes[i]) {
			equal = false;
			break;
		    }
		}
	    }
	    return equal;
	}

	/**
	 * Calculates a Java hash code for this SecureHash object
	 * (by taking its lower order bytes)
	 * @return the Java hash code for this SecureHash
	 */
	public int hashCode(){
		int result = 0; 
		for (int i = 0; i < 4; ++i) {
			result <<= 8;
			result |= 0xff & ((int) _bytes[i]);
		}
		return result;
	}

	/**
	 * Calculates a Java hash code for this SecureHash object
	 * (by taking its lower order bytes)
	 * @return the Java hash code for this SecureHash
	 */
	public long lower64bits(){
		long result = 0; 
		for (int i = 0; i < 8; ++i) {
			result <<= 8;
			result |= 0xff & ((int) _bytes[i]);
		}
		return result;
	}

	/**
	 * Hashes over the input bytes, starting at offset, going for len
	 * @param byte array over which to hash
	 * @param offset in bytes[] at which to begin hashing
	 * @param number of bytes over which to hash
	 */
	public void hash(byte bytes[], int offset, int len) {
		try {
			MessageDigest md = MessageDigest.getInstance("SHA");
			md.update(bytes, offset, len);
			_bytes = md.digest();
		} catch (java.security.NoSuchAlgorithmException e) {
			System.err.println(
				"SHA1Hash: cannot locate SHA-1 algorithm");
		}
	}

	/**
	 * Hashes over the input bytes
	 * @param byte array over which to hash
	 */
	public void hash(byte bytes[]) {
		hash(bytes, 0, bytes.length);
	}

	/**
	 * Hashes over a Digestible object
	 * @param the object over which to hash
	 */
	public void hash(Digestible o) {
	    SHA1HashBuffer digest = new SHA1HashBuffer();
	    digest.add( o );
	    _bytes = digest.compute_bytes();

            /*
            if (VERIFY_DIGESTIBLE_AGAINST_QS)
            {
                SHA1Hash qs_hash = hash_of_qs(o);
                Assert.assert(this.equals(qs_hash),
                              "SHA1Hash check failed.  " +
                              "Digestible hash=" + this +
                              "QS hash=" + qs_hash + ".");
            }
            */
	}

	/**
	 * Hashes over the bytes of the input String
	 * @param the String over which to hash
	 */
	public void hash(String str) {
		hash(str.getBytes());
	}

	/**
	 * This routine concatenates two hashes and stores the result
	 * in the input byte array.
	 * @param hash1 the first hash to concatenate
	 * @param hash2 the second hash to concatenate
	 * @param buf the byte array in which to storet the concatenation
	 */
	public static void concatenateHashes(SHA1Hash hash1, SHA1Hash hash2,
						byte buf[])
    	{
		int [] offset = {0};
		hash1.to_bytes(buf, offset);
		hash2.to_bytes(buf, offset);
	}

	/**
	 * Creates a hierarchical hash over an array of byte arrays,
	 * and returns an array of arrays, the i^th element
	 * of which contains the hashes along the tree of hashes
	 * to the root hash for the i^th byte array in the input array;
	 * This hash is set to the value of that root.
	 * @param array of byte arrays.
	 * @return array of verification information
	 */
	public SecureVerify [] hash(byte array[][]) {
		SHA1Hash tree[] = new SHA1Hash[2*array.length - 1];
		SHA1Verify ret[] = new SHA1Verify[array.length];
		int parents[] = new int[tree.length];
		int sibling[] = new int[tree.length];

		// This array is where the concatenation of two
		// hashes is stored
		byte buf[] = new byte[2*NUM_BYTES];

		int left, right, parent, cousinCount, node;

		// Initialize the leaves of the verification tree
		// to be the hashes over the input byte arrays
		for (int i = 0; i < array.length; ++i) {
			tree[i] = new SHA1Hash(array[i]);
		}

		// Build the tree, where a parent is the hash over the
		// concatenation of its left and right children
		left = 0;
		right = 1;
		parent = array.length;
		while (parent > right) {
			concatenateHashes(tree[left], tree[right], buf);
			
			tree[parent] = new SHA1Hash(buf);
			parents[left] = parent;
			parents[right] = parent;
			sibling[left] = right;
			sibling[right] = left;

			left = right + 1;
			right = left + 1;
			parent++;
		}
		parent--;
		// NOTE:  At this point, parent is the index of the root hash

		// For each input byte array, trace up the path from it to
		// the root hash and store each sibling node along that path
		// in the byte array's verification infromation
		for (int i = 0; i < array.length; ++i) {

			cousinCount = 0;
			node = i;
			// Count the number of nodes just off the path from
			// the current node to the root of the tree
			while (node < parent) {
				cousinCount++;
				node = parents[node];
			}
			ret[i] = new SHA1Verify(cousinCount);

			cousinCount = 0;
			node = i;
			// Store each of the node's cousins in its
			// verification info
			while (node < parent) {
				ret[i].setHash(cousinCount,
					tree[sibling[node]]);
				cousinCount++;
				node = parents[node];
			} // End while(node < parent)
		}  // End for (i < array.length)

		_bytes = tree[parent]._bytes;
	        return ret;
	}

	/**
	 * Creates a hierarchical hash over an array of QuickSerializable
	 * objects, and returns an array of arrays, the i^th element
	 * of which contains the hashes along the tree of hashes
	 * to the root hash for the i^th object in the input array;
	 * This hash is set to the value of that root.
	 * @param array of QuickSerializable objects.
	 * @return array of arrays of verification information
	 */
	public SecureVerify [] hash(QuickSerializable array[]) {
		byte buf[][] = new byte[array.length][];

		for (int i = 0; i < array.length; ++i) {
			buf[i] = qs_to_bytes (array [i]);
		}
		return hash(buf);
	}

	/**
	 * Verifies that the input bytes hash to this hash value
	 * @param byte array containing the bytes to verify
	 * @param offset in bytes[] at which to begin hashing
	 * @param number of bytes over which to hash
	 * @return true iff the input bytes hash to
	 *                   a hash value equal to this hash value
	 */
	public boolean verify(byte bytes[], int offset, int len) {
		SHA1Hash temp = new SHA1Hash(bytes, offset, len);
		return equals(temp);
	}

        /**
         * Verifies that the input bytes hash to this hash value
         * @param byte array to verify
         * @return true iff the input bytes hash to
         *                   a hash value equal to this hash value
         */
        public boolean verify(byte bytes[]) {
		return verify(bytes, 0, bytes.length);
	}

	/**
	 * Verifies a QuickSerializable object
	 * @param the object over which to hash
	 * @return true iff the input object hashes to
	 *                   a hash value equal to this hash value
	 */
	public boolean verify(Digestible o) {
            SecureHashBuffer buffer = new SHA1HashBuffer();
            buffer.add(o);
            SecureHash hash = buffer.toHash();
            return equals(hash);
	}

	/**
	 * Verifies the input String
	 * @param the String to verify
	 * @return true iff the input String hashes to
	 * a hash value equal to this hash value
	 */
	public boolean verify(String str) {
		return verify(str.getBytes());
	}

	/**
	 * Verifies that the input byte array, coupled with the input
	 * verification information, hashes to this hash's value
	 * @param bytes to verify
	 * @param verification verification info with which to verify the bytes
	 * against this SHA1Hash
	 * @return true iff the bytes hash to the value contained in this hash
	 */
	public boolean verify(byte bytes[], SecureVerify verification) {
		SHA1Hash temp = new SHA1Hash();
		temp.calculate(bytes, verification);
		return equals(temp);
	}

	/**
	 * Verifies that the input object, coupled with the input
	 * verification information, hashes to this hash's value
	 * @param bytes to verify
	 * @param verification verification info with which to verify the bytes
	 * against this SHA1Hash
	 * @return true iff the bytes hash to the value contained in this hash
	 */
        public boolean verify(QuickSerializable o, SecureVerify verification) {
		return verify(qs_to_bytes (o), verification);
	}

	/**
	 * This routine calculates the root hash for a given object
	 * and its slice of its verification tree (basically, this routine
	 * does what verify does, but instead of comparing the last hash
	 * to this hash and returning the result, it sets this hash's value
	 * to the last hash's value)
	 * @param bytes bytes to calc a root hash for
	 * @param verification verification info to use in the calculation
	 */
	public void calculate(byte bytes[], SecureVerify verification) {
		SHA1Hash verify[] = ((SHA1Verify)verification).getArray();
		byte buf[] = new byte[2*NUM_BYTES];
		hash(bytes);

		// Successively concatenate and hash over the
		// verification hashes
		for (int i = 0; i < verify.length; ++i) {
			concatenateHashes(this, verify[i], buf);
			this.hash(buf);
		}
	}

        /**
         * This routine calculates the root hash for a given object
         * and its slice of its verification tree (basically, this routine
         * does what verify does, but instead of comparing the last hash  
         * to this hash and returning the result, it sets this hash's value
         * to the last hash's value)
         * @param qs object to calc a root hash for
         * @param verify verification information to use in the calculation
         */
        public void calculate(QuickSerializable qs, SecureVerify verify) {
		calculate(qs_to_bytes (qs), verify);
	}

	/**
	 * Converts a few bytes of this SecureHash to a human-readable String
	 * @return human-readable String representation of this Hash
	 */
	public java.lang.String toString() {
	    int num_to_print = 6;
	    String s = "(SHA1Hash:" +
		ByteUtils.print_bytes( _bytes, 0, num_to_print ) + ')';
	    return s;
	}

	/**
	 * Converts all bytes of this SecureHash to a human-readable String
	 * @return human-readable String representation of this Hash
	 * containing every nibble of this Hash's value
	 */
	public java.lang.String fullString() {
		String s = "(SHA1Hash:" +
		    ByteUtils.print_bytes( _bytes, 0, _bytes.length ) + ')';
		return s;
	}    

	public void add_to_digest( SecureHashBuffer buffer ) {
	    buffer.add (_bytes);
	}

	/**
	 * Used to test hashing by creating hierarchical hashes
	 * @param arguments; should contain one entry
	 *                which is the number of byte arrays over which
	 *                to create a hierarchical hash
	 */
	public static void main(String args[]) {
		byte array[][];
		int LEN = 100;
		int MAX = 55;

		for (int e = 1; e < MAX; ++e) {
			array = new byte[e][];
			for (int i = 0; i < array.length; i++) {
				array[i] = new byte[LEN];
				for (int j = 0; j < LEN; j++) {
					array[i][j] =
						(byte)(Math.random() *
							Byte.MAX_VALUE);
				}
			}

			SHA1Hash sh = new SHA1Hash();
			SecureVerify verify[]  = sh.hash(array);

			if (array.length > 15) {
				if (array[15][0] == Byte.MAX_VALUE)
					array[15][0] = 0;
				else
					array[15][0] += 1;
			}

			for (int i = 0; i < array.length; i++) {
				if (i == 15 && sh.verify(array[i], verify[i])) {
					System.out.println("SHA1Hash ERROR");
				}
				else if (i != 15 && !sh.verify(array[i],
					verify[i])) {
					System.out.println("SHA1Hash ERROR");
				}
			}
		}
	}

	private static byte [] qs_to_bytes (QuickSerializable o) {
	    int [] offset = {0};
	    o.to_bytes(null, offset);
	    byte buf[] = new byte[offset [0]];
	    offset [0] = 0;
	    o.to_bytes(buf, offset);
	    return buf;
	}

    private static SHA1Hash hash_of_qs(QuickSerializable o)
    {
        int[] offset = {0};
        o.to_bytes(null, offset);
        byte[] bytes = new byte[offset[0]];
        offset[0] = 0;
        o.to_bytes(bytes, offset);
        return new SHA1Hash(bytes);
    }
}
