/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;

/**
 * SecureHash is the interface used to represent classes which
 * are secure hashes
 * An example is the SHA1 hash
 *
 * @author	Chris WElls
 * @version	$Id: SecureHash.java,v 1.10 2002/04/06 22:52:44 eaton Exp $
 */
public interface SecureHash extends Cloneable, Comparable, Digestible {

	/**
	 * Return the number of bytes used in the representation of this hash
	 * (for example, SHA1 uses 20 bytes)
	 * @return number of bytes used in the implementation
	 */
	public int size();

	/**
	 * Determines if this SecureHash is a NULL hash or not.
	 * @return true iff this Hash has a NULL value
	 */
	public boolean isNull();

	/**
	 * Returns the implementation's NULL hash
         * (typically all zeroes).
	 * @return the implementation's null hash
	 */
	public SecureHash nullHash();

	/**
	 * Returns the bytes contained in this hash
	 * @return the bytes of this hash
	 */
	public byte[] bytes();

	/**
	 * Compares this SecureHash with another hash, and returns
	 * true iff they are the same type of hash and have the
	 * same byte representation
	 * @param o the object with which to compare this SecureHash
	 * @return true iff this and o contain the same byte values
	 */
	public boolean equals(Object o);

	/**
	 * Calculates a Java hash code for this SecureHash object
	 * (typically by taking its lower order bytes)
	 * @return the Java hash code for this SecureHash
	 */
	public int hashCode();

	/**
	 * Hashes over the input bytes, starting at offset, going for len
	 * @param bytes byte array over which to hash
	 * @param offset offset in bytes[] at which to begin hashing
	 * @param len number of bytes over which to hash
	 */
	public void hash(byte bytes[], int offset, int len);

	/**
	 * Hashes over the input bytes
	 * @param bytes byte array over which to hash
	 */
	public void hash(byte bytes[]);

	/**
	 * Hashes over a Digestible object
	 * @param o the object over which to hash
	 */
	public void hash(Digestible o);

	/**
	 * Hashes over the bytes of the input String
	 * @param str the String over which to hash
	 */
	public void hash(String str);

	/**
	 * Creates a hierarchical hash over an array of byte arrays,
	 * and returns an array of SecureVerifiers, the i^th element
	 * of which contains the hashes along the tree of hashes
	 * to the root hash for the i^th byte array in the input array;
	 * This hash is set to the value of that root.
	 * @param array array of byte arrays
	 * @return array of arrays of verification information
	 */
	public SecureVerify[] hash(byte array[][]);

	/**
	 * Creates a hierarchical hash over an array of objects,
	 * and returns an array of SecureVerifiers, the i^th element
	 * of which contains the hashes along the tree of hashes
	 * to the root hash for the i^th byte array in the input array;
	 * This hash is set to the value of that root.
	 * @param array array of byte arrays
	 * @return array of arrays of verification information
	 */
	public SecureVerify[] hash(QuickSerializable array[]);

	/**
	 * Creates a hierarchical hash over an array of objects,
	 * and returns an array of SecureVerifiers, the i^th element
	 * of which contains the hashes along the tree of hashes
	 * to the root hash for the i^th byte array in the input array;
	 * This hash is set to the value of that root.
	 * @param array array of Digestible objects
	 * @return array of arrays of verification information
	public SecureVerify[] hash(Digestible array[]);
	 */

	/**
	 * Verifies that the input bytes hash to this hash value
	 * @param bytes byte array containing the bytes to verify
	 * @param offset offset in bytes[] at which to begin hashing
	 * @param len number of bytes over which to hash
	 * @return true iff the input bytes hash to
	 * a hash value equal to this hash value
	 */
	public boolean verify(byte bytes[], int offset, int len);

	/**
	 * Verifies that the input bytes hash to this hash value
	 * @param bytes byte array to verify
	 * @return true iff the input bytes hash to
	 * a hash value equal to this hash value
	 */
	public boolean verify(byte bytes[]);

	/**
	 * Verifies a Digestible object
	 * @param o the object over which to hash
	 * @return true iff the input object hashes to
	 * a hash value equal to this hash value
	 */
	public boolean verify(Digestible o);

	/**
	 * Verifies the input String
	 * @param str the String to verify
	 * @return true iff the input String hashes to
	 * a hash value equal to this hash value
	 */
	public boolean verify(String str);

	/**
	 * Verifies that the input byte array, coupled with the input
	 * verification information, hashes to this hash's value
	 * @param bytes bytes to verify
	 * @param verify verification information to use to verify bytes
	 * @return true iff the bytes hash to the value contained in this hash
	 */
	public boolean verify(byte bytes[], SecureVerify verify);

	/**
	 * Verifies that the input object, coupled with the input
	 * verification information, hashes to this hash's value
	 * @param o object to verify
	 * @param verify verification information to use to verify bytes
	 * @return true iff object hashes to the value contained in this hash
	 */
	public boolean verify(QuickSerializable o, SecureVerify verify);

	/**
	 * Verifies that the input object, coupled with the input
	 * verification information, hashes to this hash's value
	 * @param o object to verify
	 * @param verify verification information to use to verify bytes
	 * @return true iff object hashes to the value contained in this hash
	public boolean verify(Digestible o, SecureVerify verify);
	 */

	/**
	 * This routine calculates the root hash for a given object
	 * and its slice of its verification tree (basically, this routine
	 * does what verify does, but instead of comparing the last hash
	 * to this hash and returning the result, it sets this hash's value
	 * to the last hash's value)
	 * @param bytes bytes to calc a root hash for
	 * @param verify verification information to use in the calculation
	 */
	public void calculate(byte bytes[], SecureVerify verify);

	/**
	 * This routine calculates the root hash for a given object
	 * and its slice of its verification tree (basically, this routine
	 * does what verify does, but instead of comparing the last hash
	 * to this hash and returning the result, it sets this hash's value
	 * to the last hash's value)
	 * @param qs object to calc a root hash for
	 * @param verify verification information to use in the calculation
	 */
	public void calculate(QuickSerializable qs, SecureVerify verify);

	/**
	 * This routine calculates the root hash for a given object
	 * and its slice of its verification tree (basically, this routine
	 * does what verify does, but instead of comparing the last hash
	 * to this hash and returning the result, it sets this hash's value
	 * to the last hash's value)
	 * @param o object to calc a root hash for
	 * @param verify verification information to use in the calculation
	public void calculate(Digestible o, SecureVerify verify);
	 */

	/**
	 * Converts a few bytes of this SecureHash to a human-readable String
	 * @return human-readable String representation of this Hash
	 */
	public java.lang.String toString();

	/**
	 * Converts all bytes of this SecureHash to a human-readable String
	 * @return human-readable String representation of this Hash
	 * containing every nibble of this Hash's value
	 */
	public java.lang.String fullString();

    /**
     * This is the clone routine for Fragment.  It deeply clones a fragment
     * (meaning that it copies all of the data, not just references). */
    public Object clone() throws CloneNotSupportedException;
}
