/**
 * SecureHashBuffer.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.util;
import java.math.BigInteger;

/**
 * A <code>MessageDigest</code>-like interface for building
 * <code>SecureHash</code>es.
 * <p>
 * Users should call <code>add</code> for each object to include in the
 * <code>SecureHash</code> and then call <code>toHash</code> once.
 * Invocations of <code>toHash</code> reset the
 * <code>SecureHashBuffer</code>, so do <b>not</b> attempt to produce
 * intermediate hashes with a single buffer.
 *
 * @see Digestible
 * @see java.security.MessageDigest
 *
 * @author	Dennis Geels
 * @version	$Id: SecureHashBuffer.java,v 1.4 2002/03/29 21:14:53 eaton Exp $
 */

public abstract class SecureHashBuffer
{
    ////////////////////       CONSTANTS	////////////////////
    ////////////////////    STATIC VARIABLES 	////////////////////
    ////////////////////       VARIABLES	////////////////////
    ////////////////////  STATIC CONSTRUCTORS	////////////////////
    ////////////////////      CONSTRUCTORS	////////////////////
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Adds the specified byte to the digest.
     */
    public abstract void add( byte input );

    /**
     * Adds the specified byte array to the digest.
     */
    public abstract void add( byte[] input );

    /**
     * Adds <code>length</code> bytes of the specified array to the digest,
     * starting at <code>offset</code>.
     */
    public abstract void add( byte[] input, int offset, int length );

    /**
     * Adds the specified object to the digest.
     * This method invokes the {@link Digestible#add_to_digest
     * input.add_to_digest} method; be careful to avoid infinite loops.
     */
    public abstract void add( Digestible input );

    public final void add(boolean b)
    {
        add((byte) (b ? 1 : 0));
        return;
    }

    public final void add (int i) {
	// must be MSBF
	for (int j = 3; j >= 0; --j) {
	    add ((byte) (0xff & (i>>(8*j))));
	}
    }

    public final void add (long i) {
	// must be MSBF
	for (int j = 7; j >= 0; --j) {
	    add ((byte) (0xff & (i>>(8*j))));
	}
    }

    public final void add(String s)
    {
        byte s_bytes[] = s.getBytes();

        add(s_bytes.length);
        add(s_bytes);

        return;
    }

    public final void add (BigInteger value) {
	// TODO: I wish I could do this without allocating memory, but the
	// BigInteger interface doesn't allow for it.

	byte [] bytes = value.toByteArray ();
	add (bytes.length);
	add (bytes);
    }

    /**
     * Computes and returns the digest of all inputs.  Also resets internal
     * state, so do <b>not</b> attempt to produce intermediate hashes by
     * calling this method prematurely.
     */
    public abstract SecureHash toHash();

    ////////////////////     INNER CLASSES	////////////////////
}
