/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.InstantiationException;
import java.lang.IllegalAccessException;
import java.util.Hashtable;
import java.util.HashMap;
import java.util.Map;

/**
 * Allows for the creation of new QuickSerializable objects from their type
 * codes and a byte array.  This class facilitates the de-serialization
 * process by providing what are, in effect, virtual constructors.
 *
 * @author   Sean C. Rhea
 * @version  "$Id: TypeTable.java,v 1.22 2002/03/28 02:29:11 ravenben Exp $"
 * 
 * @see ostore.util.QuickSerializable
 * @see ostore.util.QSException
 * @see ostore.util.Types
 */
public class TypeTable {

    private static final boolean DEBUG = ostore.util.DebugFlags.TYPE_TABLE;

    /**
     * An exception class thrown when a given type code is passed to
     * {@link #register_type} for the second time with a different Java
     * type than the original call.
     */
    static public class DuplicateTypeCode extends Exception {
	/**
	 * The type code that was registered twice.
	 */
	public long type_code;

	/**
	 * Construct with the given type code.
	 *
	 * @param t the duplicate type code
	 */
	public DuplicateTypeCode (long t) {
	    super( Long.toHexString(t) );
	    type_code = t;
	}
    }

    /**
     * An exception class thrown when a request to {@link #new_object}
     * passes in an un-registered type code.
     */
    static public class NoSuchTypeCode extends QSException {
	public long type_code;
	public NoSuchTypeCode (long t) {
	    super( Long.toHexString(t) );
	    type_code = t;
	}
	public String toString () {
	    return "(TypeTable.NoSuchTypeCode 0x" + getMessage () + ")";
	}
        public long getTypeCode() {
            return type_code;
        }
    }

    /** 
     * An exception class thrown when the object named by the parameter
     * <code>name</code> to {@link #register_type} does not refer to a
     * class which implements <code>QuickSerializeable</code>.
     */
    static public class NotQuickSerializable extends Exception {}

    /**
     * Register a new type, so that {@link #new_object} can be called
     * later.
     *
     * @param name       the fully qualified name of the Java class being
     *                   registered
     * @param type_code  the type code to associate with this class
     * 
     * @exception DuplicateTypeCode       if the given type code has 
     *                                    already been associated with
     *                                    another class
     * @exception ClassNotFoundException  if the given class name cannot be
     *                                    resolved to any loadable class by
     *                                    the JVM
     * @exception NoSuchMethodException   if the given class does not have
     *                                    a contructor of the form 
     *   
     *     <code>Foo (byte [] data, int [] offset)</code>
     *   or
     *     <code>Foo (byte [] data, int [] offset) throws QSException</code>
     *
     * @exception NotQuickSerializable    if the given class is not a
     *                                    subtype of QuickSerializable
     */
    static public void register_type (
	    String name, 
	    int type_code) 
	throws DuplicateTypeCode, 
	       ClassNotFoundException,
               NoSuchMethodException,
	       NotQuickSerializable {

	_register_type (name, (long) type_code);
	register_type (name);
    }

    static public void register_type (String name) 
	throws DuplicateTypeCode, 
	       ClassNotFoundException,
               NoSuchMethodException,
	       NotQuickSerializable {

	long new_type_code = type_code (name);
	_register_type (name, new_type_code);
    }

    static private void _register_type (
	    String name, 
	    long type_code) 
	throws DuplicateTypeCode, 
	       ClassNotFoundException,
               NoSuchMethodException,
	       NotQuickSerializable {

        Long Type_code = new Long (type_code);

	synchronized (_constructors) {

	    Class this_class = Class.forName (name);
	    Class existing_class = (Class) _classes.get (Type_code);

	    if (existing_class != null)
	    {
		// This "type_code" is already registered to some type.  If
		// that type is the same as the one specified by "name",
		// then everything is okay, otherwise we need to throw a
		// DuplicateTypeCode exception.

		if (! existing_class.isAssignableFrom (this_class) ||
			! this_class.isAssignableFrom (existing_class))
		    throw new DuplicateTypeCode (type_code);
	    }
	    else {
		// This "type_code" is not yet registered.

		// First, make sure this type is actually a subtype of
		// QuickSerializable.

		Class qs_class = 
		    Class.forName ("ostore.util.QuickSerializable");

		if (! qs_class.isAssignableFrom (this_class))
		    throw new NotQuickSerializable ();

		// Now get the constructor...

		Class [] args = {(new byte [1]).getClass (),
		                 (new int  [1]).getClass ()};
		Constructor constructor = this_class.getConstructor (args);

		// ...and check its exception types.

		Class qsec = Class.forName ("ostore.util.QSException");
		Class [] etypes = constructor.getExceptionTypes ();
		if ((etypes.length > 0) && 
		    ((etypes.length != 1) ||
		     (! qsec.isAssignableFrom (etypes [0]))))
		    throw new NoSuchMethodException ();

		// Make sure it's not an interface.

		if (this_class.isInterface ())
		    throw new NotQuickSerializable ();

		// Everything checks out.  Register the constructor.

		_classes.put (Type_code, this_class);
		_constructors.put (Type_code, constructor);

		if (DEBUG) System.out.println ("TypeTable: type_code " + 
			   Long.toHexString (Type_code.longValue ()) + 
					       ", class " + name + ".");
	    }
	}
    }
    
    /**
     * Constructs a new object of the type associated with the given type
     * code through a call to {@link #register_type} using the data in
     * the given byte array.
     *
     * @param type_code  the type code of the class to construct
     * @param data       an array of bytes containing an encoding of an
     *                   object of the given type (!= null)
     * @param offset     a single element array which on function entry
     *                   should contain the starting index into the 
     *                   <code>data</code> array to begin reading from, and
     *                   which on function exit will be incremented by the
     *                   number of bytes read.
     *
     * @exception QSException  if the constructor for the given type throws 
     *                         one
     */
    static public QuickSerializable new_object (
	    int type_code,
	    byte [] data,
	    int [] offset) 
	throws QSException {

        return new_object ((long) type_code, data, offset);
    }

    static public QuickSerializable new_object (
	    long type_code,
	    byte [] data,
	    int [] offset) 
	throws QSException {

        Long Type_code = new Long (type_code);

	Assert.assert (data != null, "TypeTable: constructor from byte "
		+ "stream assumes data is non-null.");

	Constructor c = null;
	synchronized (_constructors) { 
	    c = (Constructor) _constructors.get (Type_code); 
	}
	if (c == null)
	    throw new NoSuchTypeCode (type_code);

	QuickSerializable result = null;
	try {
	    Object [] args = {data, offset};
	    result = (QuickSerializable) c.newInstance (args);
	}
	catch (InvocationTargetException e) {
	    System.err.print ("ostore.util.TypeTable: caught target " +
		    "exception: " + e.getTargetException () + " while " +
		    "trying to create a new object with type code 0x");
	    System.err.print (Long.toHexString (type_code));
	    System.err.println (".\n  Stack trace:");
	    e.printStackTrace ();
            if (e.getTargetException() instanceof RuntimeException)
            {
		System.err.println (".  Propogating it as a RTE.");
                throw (RuntimeException) e.getTargetException();
            }
            else if (e.getTargetException() instanceof QSException)
            {
		System.err.println (".  Propogating it as a QSException.");
                throw (QSException) e.getTargetException ();
            }
            else
            {
                Carp.die("TypeTable: This is a bug.  Report to Sean");
            }
	}
	catch (InstantiationException e) {
	    Carp.die ("TypeTable.newobject: class of type code " +
		      type_code + " is an interface or abstract class." + 
		      "  This is a bug in TypeTable.");
	}
	catch (IllegalAccessException e) {
	    Carp.die ("TypeTable.newobject: class of type code 0x" +
		      Long.toHexString (type_code) + 
		      " is an non-accessable constructor." + 
		      "  This is a bug in TypeTable.");
	}
	return result;
    }

    static public long type_code (QuickSerializable obj) {
	String class_name = obj.getClass ().getName ();
	return type_code (class_name);
    }

    static public long type_code (String class_name) {
	Long result = null;
	synchronized (_type_codes) {
	    result = (Long) _type_codes.get (class_name);
	    if (result == null) {
		SHA1Hash hash = new SHA1Hash (class_name);
		result = new Long (hash.lower64bits ());
		_type_codes.put (class_name, result);
	    }
	}
	return result.longValue ();
    }

    static private Hashtable _classes;
    static private Hashtable _constructors;
    static private Map _type_codes = new HashMap ();

    static {
	_classes = new Hashtable ();
	_constructors = new Hashtable ();
    }
}

