/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;

import ostore.cache.Cacheable;

/**
***  The <code>VerifiableBlock</code> interface should be implemented
***  by all objects that need to be <em>self-verifiable</em>.  In
***  order to be self-verifiable, an object must maintain a field to
***  record the <code>fhash</code>, the secure hash that lies at the
***  root of the verification tree constructed from the
***  erasure-encoded fragments produced by the archival layer.  A
***  self-verifiable object must also provide methods for computing
***  and verifying guids and verification hashes.  In computing a
***  guid, an object should reference all information (including its
***  <code>fhash</code>).  In computing a vhash, however, an object
***  should not reference any guids that are created by the archive.
***  If the archive has not yet computed the fragments for an object,
***  the object will lack an <code>fhash</code> and thus will be
***  unable to compute or verify its own guid.  If such a request is
***  made, a {@link BlockNotGuidVerifiableException} is thrown.
***
***  @author  Patrick R. Eaton
***  @version $Id: VerifiableBlock.java,v 1.4 2002/04/02 23:52:05 eaton Exp $
**/
public interface VerifiableBlock extends Cacheable, ostore.util.Digestible
{
    /**
    ***  Return the secure hash at the root of the verification tree
    ***  constructed from the erasure-encoded fragments of a block.
    ***  The root hash verifies the the fragments of the block and is
    ***  a component in the computation of the block GUID.
    ***
    ***  @return
    ***      the secure hash at the root of the verification tree
    **/
    SecureHash getFhash();

    /**
    ***  Record the secure hash at the root of the verification tree
    ***  constructed from the erasure-encoded fragments of a block.
    ***  The root hash verifies the the fragments of the block and is
    ***  a component in the computation of the block GUID.
    ***
    ***  @param fhash
    ***      the secure hash at the root of the verification tree
    **/
    void setFhash(SecureHash fhash);

    /**
    ***  Compute and return the verification hash of the object.
    ***
    ***  @return
    ***      the vhash of the object
    **/
    SecureHash computeVhash();

    /**
    ***  Serialize the object in a form suitable for computing the
    ***  verification hash of the object.  The parameters and their
    ***  meaning mimic those of the QuickSerializable.to_bytes()
    ***  method.  @see ostore.util.QuickSerializable#to_bytes
    ***
    ***  @param data
    ***      See {@link ostore.util.QuickSerializable#to_bytes}
    ***  @param offset
    ***      See {@link ostore.util.QuickSerializable#to_bytes}
    **/
    void toVhashBytes(byte[] data, int[] offset);

    /**
    ***  Verify the contents of the object against its given
    ***  verification hash.  This method will compute the verification
    ***  hash of the object and compare that hash against the alleged
    ***  hash of the object to determine its validity.
    ***
    ***  @param vhash
    ***      the vhash against which to verify the object
    ***  @return
    ***      <code>true</code> iff the hash of the object matches
    ***      <code>vhash</code>; <code>false</code> otherwise
    **/
    boolean verifyVhash(SecureHash vhash);

    /**
    ***  Compute and return the guid of the object.  As a
    ***  precondition, the secure hash of the root of the fragment
    ***  verification tree must be set.
    ***
    ***  @see #setFhash
    ***
    ***  @return
    ***      the guid of the object
    ***  @exception
    ***     ostore.util.BlockNotGuidVerifiable if the precondition is
    ***     not satisfied.
    **/
    SecureHash computeGuid()
        throws ostore.util.BlockNotGuidVerifiableException;

    /**
    ***  Serialize the object in a form suitable for computing the
    ***  block guid of the object.  The parameters and their
    ***  meaning mimic those of the QuickSerializable.to_bytes()
    ***  method.  @see ostore.util.QuickSerializable#to_bytes
    ***  As a precondition, the secure hash of the root of the
    ***  fragment verification tree must be set.
    ***  @param data
    ***      See {@link ostore.util.QuickSerializable#to_bytes}
    ***  @param offset
    ***      See {@link ostore.util.QuickSerializable#to_bytes}
    ***  @exception
    ***     ostore.util.BlockNotGuidVerifiable if the precondition is
    ***     not satisfied.
    **/
    void toGuidBytes(byte[] data, int[] offset)
        throws ostore.util.BlockNotGuidVerifiableException;

    /**
    ***  Verify the contents of the object against its given guid.
    ***  This method will compute the guid of the object and compare
    ***  that hash against the alleged name of the object to determine
    ***  its validity.  As a precondition, the secure hash of the root
    ***  of the fragment verification tree must be set.
    ***
    ***  @param bguid
    ***      the block guid against which to verify the object
    ***  @return
    ***      <code>true</code> iff the guid of the object matches
    ***      <code>bguid</code>; <code>false</code> otherwise
    ***  @exception
    ***     ostore.util.BlockNotGuidVerifiable if the precondition is
    ***     not satisfied.
    **/
    boolean verifyGuid(SecureHash bguid)
        throws ostore.util.BlockNotGuidVerifiableException;
}
