# Toolbox.pm
#
# $Id: Toolbox.pm,v 1.31 2002/04/07 02:23:50 geels Exp $
#
# Useful common subroutines and data structures for regression test scripts.

package Toolbox;
use Exporter();
use File::Path "rmtree";

@ISA = (Exporter);
@EXPORT = qw( @child_pids @logs @configs @dirs @profiles
	      $hostname $ostore_home $key_dir $tmp_dir $script_name
	      $FED_INDEX $RP_INDEX $CLIENT_INDEX 
	      &fork_process &fork_sandstorm &fork_java &fork_simulator
	      &remote_tmp_dir &wait_for_ready &wait_for_tapestry_ready 
	      &wait_for_network_ready &wait_for 
	      &returned_zero &is_clean &rewrite &override &insert &interpolate
	      &done &clean_up &croak $seed);
$| = 1;

$seed = 1000;

# Script name - reasonable default
$script_name = $0;
$script_name =~ s|.*/||;

# PIDs of all children forked (will be killed)
@child_pids = ();

# Names of all log files (will be deleted)
@logs = ();

# Names of all config files (will be deleted)
@configs = ();

# Names of all temp dirs (will be deleted)
@dirs = ();

# Names of all profiles files (will be deleted)
@profiles = ();
# TODO: don't use default name ever - add specialized name to @profiles
push @profiles, "/tmp/sandstorm-graph.txt";	# a common one

# Get the hostname.
$hostname = `hostname`;
chomp $hostname;

# And the directory with the OceanStore sources.
if (! $ENV{"OSTOREHOME"}) {
    $ostore_home = $ENV{"HOME"} . "/oceanstore";
} else {
    $ostore_home = $ENV{"OSTOREHOME"};
}
$key_dir = $ostore_home . "/pond/regression/keys";
$ostore_home .= "/pond/ostore";

# Choose how/where to run sandstorm
# options are "single", "multiple", and "remote"
if( $ENV{"OSTORE_TEST_MODE"}) {
    $mode = $ENV{"OSTORE_TEST_MODE"};
} else {
    $mode = "single";	# default
}

if( $mode eq "remote" ) {
    # not supported yet.
    &croak;
}

# parameterize the local /tmp or alternative.
$tmp_dir = ($ENV{"OSTORE_TMP"} or "/tmp");

# parameterize the remote /tmp or alternative.
#
# Reads a value from an environment variable if possible.
# chooses a preset /tmp alternative for certain machines.
# Otherwise, uses /tmp.
sub remote_tmp_dir {
    my $remote_host = shift;
    my $remote_tmp;
    
    if( $ENV{"OSTORE_REMOTE_TMP"} ) {
	$remote_tmp = $ENV{"OSTORE_REMOTE_TMP"};
    } elsif( $remote_host =~ /^ibm\d/ ) {
	# ibm cluster
	$remote_tmp = "/scratch2";
    } elsif( $remote_host =~ /^x\d/ ) {
	# ROC cluster
	$remote_tmp = "/stripe";
    } else {
	# who knows.
	$remote_tmp = "/tmp";
    }
    return $remote_tmp;
}


# fork a process; return its PID.
# redirect STDOUT and STDERR to the named log.
# automatically adds the PID to @child_pids and the log filename to @logs.
#
# Usage: &fork_process( <log_filename>, <executable>, [<arg1>, ..., <argN>] );

sub fork_process {
    my $log = shift( @_ );
    push @logs, $log;
    
    my $pid = fork();
    if( $pid > 0 ) {
	#parent;  return PID for child.
	push @child_pids, $pid;
	return $pid;
	
    } elsif( defined $pid ) {
	# child; run the named process..
	
	open STDOUT, ">$log" or warn "can't redirect STDOUT" and &croak;
	open STDERR, ">&STDOUT" or warn "can't redirect STDERR" and &croak;

	# now exec the process as specified.
	exec (@_);
    }
    
    # should never reach this point.
    warn( "fork failed" ) and &croak;
}

# fork a process to run sandstorm; return its PID.
# redirect STDOUT and STDERR to the named log.
# automatically adds the PID to @child_pids and the log filename to @logs.
#
# Usage: &fork_sandstorm( <config_filename>, <log_filename> );

sub fork_sandstorm {
    my ($cfg, $log) = @_;

    my $run_sandstorm = "$ostore_home/run-sandstorm";
    
    return &fork_process( $log, $run_sandstorm, $cfg );
}

# fork a process to run java; return its PID.
# redirect STDOUT and STDERR to the named log.
# automatically adds the PID to @child_pids and the log filename to @logs.
#
# Usage: &fork_java( <log_filename>, <mainclass>, [<arg1>, ..., <argN>] );

sub fork_java {
    my $log = shift( @_ );

    my $run_java = "$ostore_home/run-java";
    
    return &fork_process( $log, $run_java, @_ );
}

# fork a process to run soss; return its PID.
# redirect STDOUT and STDERR to the named log.
# automatically adds the PID to @child_pids and the log filename to @logs.
#
# Usage: &fork_simulator( <log_filename>, <config_file> );

sub fork_simulator {
    my ($log, $cfg) = @_;

    &fork_process( $log, "$ostore_home/../soss/run-java", 
		   "soss.core.Main",
		   "$cfg");
}

# Greps the specified log to find the sandstorm ready signal
#
# Usage: &wait_for_ready( <log_filename> )
sub wait_for_ready {
    my $log = shift @_;

    &wait_for( $log, "^Sandstorm: Ready" );
}

# Greps the specified log to find the Tapestry ready signal
#
# Usage: &wait_for_ready( <log_filename> )
sub wait_for_tapestry_ready {
    my $log = shift @_;

    &wait_for( $log, "^Tapestry: ready" );
}

# Greps the specified log to find the Network ready signal
#
# Usage: &wait_for_network_ready( <log_filename> )
sub wait_for_network_ready {
    my $log = shift @_;

    &wait_for( $log, "^Network.*now listening on port.*" );
}

# Greps a log for a string
#
# Usage: &wait_for( <log_filename>, <string> )
# TODO: grep through the log as it is written, instead of repeatedly
#  scanning the whole file.
sub wait_for {
    my ($log, $string) = @_;
    my $timeout = time + 120;

    while( ! -r $log ) {
	if( time > $timeout ) {
	    warn "Stopping - Timed out\n";
	    kill "TERM", $$;
	}
	sleep 1;
    }
    while( `grep -c \'$string\' $log` == 0 ) {
	if( time > $timeout ) {
	    warn "Stopping - Timed out\n";
	    kill "TERM", $$;
	} elsif( ! &is_clean( $log ) ) {
	    # catch exceptions and kill self
	    # Send a signal (rather than just calling &done) to
	    #  allow main scripts to override &done in interesting ways.
	    kill "TERM", $$;
	} else {
	    # don't hog the processor
	    sleep 2;
	}
    }
}

# Waits for child processes to exit and checks their return values.
# Returns 1 (true) if all returned 0, or 0 (false) if any child returned a
# nonzero return value.
#
# Usage: &returned_zero( (<pid1>,..., <pidN>) );
sub returned_zero {
    foreach $child (@_) {
	waitpid $child, 0; #no flags
	if( $? > 0 ) {
	    return 0;
	}
    }
    return 1;
}

# greps files for any word in @bad_words
# Returns 1 (true) if clean, 0 (false) if dirty.
#
# Usage: &is_clean( (<filename1>,..., <filenameN>) );
sub is_clean {
    my @bad_words = ("ERROR", "WARNING", "JAVA.*EXCEPTION",
		     "ASSERTIONVIOLATED", "ILLEGAL", "\"DROPPING IT\"");

    foreach $file (@_) {
	if( -r $file ) {
	    foreach $word (@bad_words) {
		if( `grep -ic $word $file` > 0 ) {
		    return 0;
		}
	    }
	}
    }
    return 1;
}

# Substitute runtime values into configuration files.
# Any stand-alone occurrence of a key is replaced inline with its value.
#
# Pass in a placeholder->value hashtable,
#	(e.g. $map{"\$\{FederationID\}"} = "$hostname:3630").
# Usage: &interpolate( <template>, <output_filename>, <map> )

sub interpolate {
    my ($template, $output, %map) = @_;

    open (IN, "<$template");
    open (OUT, ">$output");

    # Write each line of the template into the output file, replacing
    # special keys with runtime values.
    my $line_no = 0;
    while( <IN> ) {
	$line_no++;
	while( /\$\{(.*)\}/ ) {
	    # found a variable - replace from map.
	    my $variable = $1;
	    my $value = $map{$variable};
	    if( defined $value ) {
		s/\$\{$variable\}/$value/;
	    } else {
		print "ERROR: undefined variable \"$variable\"" .
		    " on line $line_no of file $template\n";
		kill "TERM", $$;
	    }
	}
	print OUT $_;
    }

    close( IN );
    close( OUT );
}


########################################
# Common value map, with index-independent values
# We need to triple-backslash the keys so that perl doesn't dereference them.

my %value_map;
my $fed_port = 3630;
my $rp_port = $fed_port+1;
my $client_port = $fed_port+2;
my @pr_pkeys = glob( "$key_dir/pr-*pub" );
my @pr_skeys = glob( "$key_dir/pr-*priv" );
my @ir_keyshares = glob( "$key_dir/ir-*priv" );
my @storage_pkeys = glob( "$key_dir/storage-*pub" );
my @storage_skeys = glob( "$key_dir/storage-*priv" );

$value_map{"Host"} = "$hostname";

$value_map{"FederationID"} = "$hostname:$fed_port";
$value_map{"FederationPort"} = "$fed_port";

$value_map{"MemberPublicKeys"} = "@pr_pkeys";
$value_map{"BasePort"} = $fed_port + 3;
$value_map{"GatewayID"} = "$hostname:$value_map{BasePort}";

my $i;
for( $i=0; $i<scalar(@pr_pkeys); $i++ ){
    $value_map{"MemberPublicKey\\\[$i\\\]"} = "$pr_pkeys[$i]";
    $value_map{"MemberPrivateKey\\\[$i\\\]"} = "$pr_skeys[$i]";
    my $port = $value_map{"BasePort"} + $i;    
    $value_map{"MemberID\\\[$i\\\]"} = "$hostname:$port";
}

$value_map{"IRPublicKey"} = "$key_dir/ir.pub";
$value_map{"IRPrivateKeyShares"} = "@ir_keyshares";
for( $i=0; $i<scalar(@ir_keyshares); $i++ ){
    $value_map{"IRPrivateKeyShare\\\[$i\\\]"} = "$ir_keyshares[$i]";
}

for( $i=0; $i<scalar(@storage_pkeys); $i++ ){
    $value_map{"StoragePublicKey\\\[$i\\\]"} = "$storage_pkeys[$i]";
    $value_map{"StoragePrivateKey\\\[$i\\\]"} = "$storage_skeys[$i]";
    my $port = $value_map{"BasePort"} + $i;    
    $value_map{"MemberID\\\[$i\\\]"} = "$hostname:$port";
}

$value_map{"RPID"} = "$hostname:$rp_port";
$value_map{"RPPort"} = "$rp_port";
$value_map{"RPPublicKey"} = "$key_dir/rp.pub";
$value_map{"RPPrivateKey"} = "$key_dir/rp.priv";

$value_map{"PrincipalPublicKey"} = "$key_dir/principal.pub";
$value_map{"PrincipalPrivateKey"} = "$key_dir/principal.priv";

# TODO: client could be generic
$value_map{"ClientID"} = "$hostname:$client_port";
$value_map{"ClientPort"} = "$client_port";
$value_map{"ClientPublicKey"} = "$key_dir/client.pub";
$value_map{"ClientPrivateKey"} = "$key_dir/client.priv";
    
$FED_INDEX = -1;
$RP_INDEX = -2;
$CLIENT_INDEX = -3;

my %tags;
$tags{$FED_INDEX} = "federation";
$tags{$RP_INDEX} = "rp";
$tags{$CLIENT_INDEX} = "client";

# Produces sandstorm configuration files with runtime values.
#
# The "index" parameter should be the index of the configured machine, if
#   it is a member of the inner ring.  Otherwise use one of the "*_INDEX"
#   constants above.
# Usage: &rewrite( <template>, <output_filename>, <index> )

sub rewrite {
    my ($template, $output, $index) = @_;

    # fill in index-specific portions of %value_map.
    my $port = $value_map{"BasePort"} + $index;
    my $tag = $tags{$index};
    if( ! defined $tag ) {
	$tag = "node-$index";
    }
    $value_map{"OSTOREHOME"} = "$ostore_home";
    $value_map{"MemberIndex"} = "$index";
    $value_map{"MemberID"} = "$hostname:$port";
    $value_map{"MemberPort"} = "$port";
    $value_map{"MemberPublicKey"} = "$key_dir/pr-$index.pub";
    $value_map{"MemberPrivateKey"} = "$key_dir/pr-$index.priv";
    $value_map{"IRPrivateKeyShare"} = "$key_dir/ir-$index.priv";
    $value_map{"SSProfileFile"} = "$tmp_dir/$script_name-profile-$tag-$$.txt";
    $value_map{"SSGraphFile"} = "$tmp_dir/$script_name-graph-$tag-$$.txt";
    $value_map{"CacheDir"} = "$tmp_dir/$script_name-blocks-$tag-$$/";

    # Add the newly created filenames to appropriate lists.
    push @configs, $output;
    push @dirs, $value_map{"CacheDir"};
    push @profiles, $value_map{"SSProfileFile"};
    push @profiles, $value_map{"SSGraphFile"};
    
    # Now write the new, interpolated file.
    &interpolate( $template, $output, %value_map );
}

# Swap in new values for named variables
# Any line defining a specified key will be rewritten.  The new line will
#	set the key to its provided value.
#
# Pass in a variable->value hashtable, (e.g. $map{"Profile"} = "false").
# Usage: &override( <cfg_filename> <%hashtable> )

sub override {
    my ($cfg, %map) = @_;
    my $tmp_name = "$cfg.$$.tmp";
    
    open( IN, "<$cfg" );
    open( OUT, ">$tmp_name" );

    # Fix here, too.
  LINE: while( <IN> ) {
      foreach $key (keys %map) {
	  if( /^(\s*)$key\b.*/ ) {
	      print OUT "$1$key\t$map{$key}\n";
	      next LINE;
	  }
      }
      print OUT $_;	# no match
  }
    
    close( IN );
    close( OUT );

    # now replace the old copy with the new one.
    rename( $tmp_name, $cfg );
}

# Adds new variable-value mappings to a config file
#
# All fields will be added to the "<initargs>" section of the named stage,
#  or to the global initargs if "global" is specified.
#
# Pass in a variable->value hashtable, (e.g. $map{"Profile"} = "false").
# Usage: &insert( <cfg_filename>, (<stagename>|"global"), <%hashtable> )

sub insert {
    my ($cfg, $stage, %map) = @_;
    my $tmp_name = "$cfg.$$.tmp";
    
    open( IN, "<$cfg" );
    open( OUT, ">$tmp_name" );

    # find stage/global
    while( <IN> ) {
	print OUT $_;
	last if( /^\s*<$stage>/ );
    }
    # find initargs
    while( <IN> ) {
	print OUT $_;
	last if( /^\s*<initargs>/ );
    }
    # replace old mappings
    LINE: while( <IN> ) {
	foreach $key (keys %map) {
	    if( /^(\s*)$key\b.*/ ) {
		# found one
		print OUT "$1$key\t$map{$key}\n";
		undef $map{$key};	# avoid repeats
		next LINE;
	    }
	}
	
	if( /^(\s*)<\/initargs>/ ) {
	    # insert remaining mappings
	    foreach $key (keys %map) {
		print OUT "$1\t$key\t$map{$key}\n";
	    }
	    print OUT $_;	# "</initargs>"
	    last;
	}
	print OUT $_;	# non-matching mapping
    }
    # copy rest of file
    while( <IN> ) {
	print OUT $_;
    }
    
    close( IN );
    close( OUT );

    # now replace the old copy with the new one.
    rename( $tmp_name, $cfg );
}

# Clean up.
#
# Usage: &done( <exit_code> )
sub done {

    if ($_[0] == 0) {
	print "*** $script_name: Success! ***\n";
    } else {
	print "*** $script_name: Failure! Logs follow ***\n";
	foreach $log ( @logs ) {
	    if( -f $log ) {
		system( "cat", $log );
	    }
	}
    }

    &clean_up();
    
    exit $_[0];
}

# Removes any remaining files or child processes
#
sub clean_up {
    foreach $gonner (@child_pids) {
	kill 9, $gonner;
    }

    foreach $file (@logs, @configs, @profiles) {
    #foreach $file (@configs, @profiles) {
        if( -f $file ) {
          unlink $file;
	}
    }

    foreach $dir (@dirs) {
	if( -d $dir ) {
	    rmtree( $dir );
	}
    }
}

# Handle signals
sub croak {
    &done( 1 );
}

# Set up a signal handler to clean up if the script is killed.
$SIG{INT} = 'croak';
$SIG{QUIT} = 'croak';
$SIG{HUP} = 'croak';
$SIG{PIPE} = 'croak';
$SIG{TERM} = 'croak';
