/*
 * Main.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package soss.core;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.InstantiationException;
import java.lang.IllegalAccessException;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Vector;
import java.util.Map;
import java.util.Date;
import java.util.Random;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.StreamTokenizer;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.Reader;
import ostore.util.Assert;
import ostore.util.Clock;
import ostore.util.NodeId;
import ostore.util.SHA1Hash;
import ostore.dispatch.Classifier;
import ostore.tapestry.api.TapestryPublishMsg;
import soss.network.Network;
import mdw.sandStorm.main.SandstormConfig;
import mdw.sandStorm.main.stageDescr;
import mdw.sandStorm.internal.ConfigData;
import mdw.sandStorm.api.StageIF;
import mdw.sandStorm.api.EventHandlerIF;
import mdw.sandStorm.api.StagesInitializedSignal;
import mdw.sandStorm.api.SinkException;
import mdw.sandStorm.api.SinkIF;
import mdw.sandStorm.api.QueueElementIF;

/**
 * Sets up a simulation and runs it.
 *
 * @author Sean C. Rhea
 * @version $Id: Main.java,v 1.10 2002/03/15 06:21:07 hweather Exp $
 */
public class Main {
    private static class SimulatorClockCB implements Clock.ClockCB { 
	public SimulatorClockCB (EventQueue event_queue) {
	    _event_queue = event_queue;
	}
	public Date date (NodeId node_id) {
	    return new Date ((long) (_event_queue.current_time () * 1000.0));
	}
	private EventQueue _event_queue;
    }

    private static class SimulatorTimerCB implements Classifier.TimerCB {

	public SimulatorTimerCB (Classifier classifier, 
		EventQueue event_queue) {
	    _classifier = classifier;
	    _event_queue = event_queue;
	}

	public Object schedule (long millis, QueueElementIF event) {
	    long now_ms = (long) (_event_queue.current_time () * 1000.0);
	    long then_ms = now_ms + millis;
	    double then_s = ((double) then_ms) / 1000.0;
	    _event_queue.schedule (event, then_s, _classifier);
	    return new Integer (0);
	}

	public void cancel (Object token) {
	    System.err.println ("soss.core.Main.SimulatorTimerCB.cancel " +
		    "not yet implemented.");
	    System.exit (1);
	}

	private Classifier _classifier;
	private EventQueue _event_queue;
    }

    public static void main (String [] args) throws Exception {

	if (args.length < 1) {
	    System.err.println ("usage: <simulation description file>");
	    System.exit (1);
	}

	int argnum = 0;
	String description_filename = args [argnum++];

	// Guard values
	int desired_successes = -1;
	String graph_filename = null;
	long seed = -1;

	FileInputStream is = new FileInputStream (description_filename);
	Reader reader = new BufferedReader (new InputStreamReader (is));
	StreamTokenizer tok = new StreamTokenizer (reader);

	tok.resetSyntax ();
	tok.commentChar ((int) '#');
	tok.whitespaceChars ((int) ' ', (int) ' ');
	tok.whitespaceChars ((int) '\t', (int) '\t');
	tok.wordChars ((int) '-', (int) '-');
	tok.wordChars ((int) '_', (int) '_');
	tok.wordChars ((int) '.', (int) '.');
	tok.wordChars ((int) ':', (int) ':');
	tok.wordChars ((int) '/', (int) '/');
	tok.wordChars ((int) '0', (int) '9');
	tok.wordChars ((int) 'a', (int) 'z');
	tok.wordChars ((int) 'A', (int) 'Z');

	int lineno = 1;
	tok.nextToken ();
	while (tok.ttype == StreamTokenizer.TT_EOL) {
	    tok.nextToken ();
	    lineno++;
	}

	while (tok.ttype != StreamTokenizer.TT_EOF) {
	    if (tok.ttype != StreamTokenizer.TT_WORD) {
		System.err.println ("Expected parameter name on line " + 
			lineno + ": " + tok);
		System.exit (1); 
	    }

	    if (tok.sval.equals ("Seed")) {
		tok.nextToken ();
		if (tok.ttype != StreamTokenizer.TT_WORD) {
		    System.err.println ("Expected seed on line " + 
			    lineno + ": " + tok);
		    System.exit (1); 
		}
		seed = Long.parseLong (tok.sval);
	    }
	    else if (tok.sval.equals ("Graph")) {
		tok.nextToken ();
		if (tok.ttype != StreamTokenizer.TT_WORD) {
		    System.err.println ("Expected graph file name on line " + 
			    lineno + ": " + tok);
		    System.exit (1); 
		}
		graph_filename = tok.sval;
	    }
	    else if (tok.sval.equals ("DesiredSuccesses")) {
		tok.nextToken ();
		if (tok.ttype != StreamTokenizer.TT_WORD) {
		    System.err.println ("Expected graph file name on line " + 
			    lineno + ": " + tok);
		    System.exit (1); 
		}
		desired_successes = Integer.parseInt (tok.sval);
		if (desired_successes < 1) {
		    System.err.println (
			    "DesiredSuccesses must be > 0 on line " + 
			    lineno + ": " + tok);
		    System.exit (1); 
		}
	    }
	    else if (tok.sval.equals ("Algorithm")) {
		System.err.println ("Must specify Graph and " +
			"DesiredSuccesses first on line " + 
			lineno + ": " + tok);
		System.exit (1); 
	    }
	    else {
		System.err.println ("Expected parameter name on line " + 
			lineno + ": " + tok);
		System.exit (1); 
	    }

	    tok.nextToken ();
	    while (tok.ttype == StreamTokenizer.TT_EOL) {
		tok.nextToken ();
		lineno++;
	    }

	    if ((graph_filename != null) && (desired_successes != -1)
		    && (seed != -1)) 
		break;
	}

	// Create the simulation monitor stage.

	SimulationMonitorStage simulation_monitor = 
	    new SimulationMonitorStage (desired_successes);
	{
	    ConfigData config_data = new ConfigData (null);
	    SinkIF sink = new SimulatorSink (simulation_monitor);
	    StageIF stage = new SimulatorStage ("Monitor", null, sink);
	    config_data.setStage (stage);
	    simulation_monitor.init (config_data);
	}

	// Read in the nodes from the config file.

	System.out.println ("*** Reading config files. ***");
	System.out.print ("Constructing nodes");

	Class [] arg_types = {(new String [1]).getClass ()};

	Vector node_ids = new Vector ();
	Map node_id_to_config = new HashMap ();
	Map node_id_to_classifier = new HashMap ();
	while (tok.ttype != StreamTokenizer.TT_EOF) {

	    if ((tok.ttype != StreamTokenizer.TT_WORD) ||
		    (! tok.sval.equals ("Algorithm"))) {
		System.err.println ("Expected Algorithm on line " + 
			lineno + ": " + tok);
		System.exit (1); 
	    }

	    tok.nextToken ();
	    if (tok.ttype != StreamTokenizer.TT_WORD) {
		System.err.println ("Expected algorithm name on line " + 
			    lineno + ": " + tok);
		    System.exit (1); 
	    }

	    // Create the builder.
	    String class_name = tok.sval;
	    Class this_class = Class.forName (class_name);
	    Constructor constructor = this_class.getConstructor (arg_types);

	    tok.nextToken ();
	    while (tok.ttype == StreamTokenizer.TT_EOL) {
		tok.nextToken ();
		lineno++;
	    }

	    if ((tok.ttype != StreamTokenizer.TT_WORD) ||
		    (! tok.sval.equals ("Arguments"))) {
		System.err.println ("Expected Arguments on line " + 
			lineno + ": " + tok);
		System.exit (1); 
	    }

	    tok.nextToken ();
	    if (tok.ttype != StreamTokenizer.TT_WORD) {
		System.err.println ("Expected argument count on line " + 
			    lineno + ": " + tok);
		    System.exit (1); 
	    }
	    int arg_cnt = Integer.parseInt (tok.sval);
	    String [] builder_args = 
		(arg_cnt > 0 ? new String [arg_cnt] : null);

	    for (int i = 0; i < arg_cnt; ++i) {
		tok.nextToken ();
		if (tok.ttype != StreamTokenizer.TT_WORD) {
		    System.err.println ("Expected argument " + i + 
			    "on line " + lineno + ": " + tok);
		    System.exit (1); 
		}
		builder_args [i] = tok.sval;
	    }

	    BuilderIF builder = null;
	    try {
		Object [] wrapper = {builder_args};
		builder = (BuilderIF) constructor.newInstance (wrapper);
	    }
	    catch (Exception e) {
		System.err.println ("Could not construct object of " +
			"type " + class_name + " using " + constructor + 
			": " + e);
		System.exit (1);
	    }

	    tok.nextToken ();
	    while (tok.ttype == StreamTokenizer.TT_EOL) {
		tok.nextToken ();
		lineno++;
	    }

	    if ((tok.ttype != StreamTokenizer.TT_WORD) ||
		    (! tok.sval.equals ("ParticipatingNodes"))) {
		System.err.println ("Expected ParticipatingNodes on line " + 
			lineno + ": " + tok);
		System.exit (1); 
	    }

	    tok.nextToken ();
	    if (tok.ttype != StreamTokenizer.TT_WORD) {
		System.err.println ("Expected node count on line " + 
			    lineno + ": " + tok);
		    System.exit (1); 
	    }
	    int node_cnt = Integer.parseInt (tok.sval);

	    tok.nextToken ();
	    while (tok.ttype == StreamTokenizer.TT_EOL) {
		tok.nextToken ();
		lineno++;
	    }

	    for (int j = 0; j < node_cnt; ++j) {
		if ((tok.ttype != StreamTokenizer.TT_WORD) ||
			(! tok.sval.equals ("Node"))) {
		    System.err.println ("Expected Node on line " + 
			    lineno + ": " + tok);
		    System.exit (1); 
		}

		tok.nextToken ();
		if (tok.ttype != StreamTokenizer.TT_WORD) {
		    System.err.println ("Expected argument count on line " + 
			    lineno + ": " + tok);
		    System.exit (1); 
		}

		int node_arg_cnt = Integer.parseInt (tok.sval);
		String [] node_args = new String [node_arg_cnt];

		for (int i = 0; i < node_arg_cnt; ++i) {
		    tok.nextToken ();
		    if (tok.ttype != StreamTokenizer.TT_WORD) {
			System.err.println ("Expected argument " + i + 
				"on line " + lineno + ": " + tok);
			System.exit (1); 
		    }
		    node_args [i] = tok.sval;
		}

		SandstormConfig config = null;
		try {
		    config = builder.build_node (node_args);
		}
		catch (IOException e) {
		    System.err.println ("Error creating node on line " +
			    lineno + ": " + e + ".  Stack trace follows.");
		    e.printStackTrace ();
		    System.exit (1);
		}

		System.out.print (".");

		Enumeration stages = config.getStages ();
		stageDescr first = (stageDescr) stages.nextElement ();
		String node_id_str = (String) first.initargs.get ("node_id");
		NodeId node_id = null;
		try {
		    node_id = new NodeId (node_id_str);
		}
		catch (Exception e) {
		    System.err.println ("Error: could not resolve node id " +
			    node_id_str + ".  Stack trace follows:");
		    e.printStackTrace ();
		    System.exit (1);
		}
		node_ids.addElement (node_id);

		if (node_id_to_config.containsKey (node_id)) {
		    System.err.println ("Error: duplicate node id " + 
			    node_id + ".");
		    System.exit (1);
		}
		node_id_to_config.put (node_id, config);
		node_id_to_classifier.put (node_id, 
			Classifier.getClassifier (node_id));

		tok.nextToken ();
		while (tok.ttype == StreamTokenizer.TT_EOL) {
		    tok.nextToken ();
		    lineno++;
		}
	    }
	}

	System.out.println ("");

	// Create the event queue.

	EventQueue event_queue = new EventQueue ();
	Clock.set_cb (new SimulatorClockCB (event_queue));

	// Create the network.

	System.out.println ("*** Loading network information. ***");

	Network network = new Network (event_queue, node_id_to_classifier,
		graph_filename);

	// Create and initialize all of the other stages.

	System.out.println ("*** Initializing stages. ***");

	for (int node_num = 0; node_num < node_ids.size (); ++node_num) {
	    NodeId node_id = (NodeId) node_ids.elementAt (node_num);

	    Classifier classifier = 
		(Classifier) node_id_to_classifier.get (node_id);
	    classifier.set_timer_cb (new SimulatorTimerCB (
			classifier, event_queue));

	    simulation_monitor.add_classifier (classifier);

	    // Create the network ``stage''.

	    {
		ConfigData config_data = new ConfigData (null);
		Network.Stage handler = network.new_stage (node_id);
		SinkIF sink = new SimulatorSink (handler);
		StageIF stage = 
		    new SimulatorStage ("Network", null, sink);
		config_data.setStage (stage);

		try {
		    handler.init (config_data);
		}
		catch (Exception exception) {
		    System.err.println ("Error: caught an exception " +
			    exception +
			    " while trying to initialize the Network " +
			    "on node " + node_id + "." +
			    "  Stack trace follows:");
		    exception.printStackTrace ();
		    System.exit (1);
		}
	    }	

	    SandstormConfig config = (SandstormConfig)
		node_id_to_config.get (node_id);
	    Manager mgr = new Manager (config);
	    Enumeration e = config.getStages ();
	    while (e.hasMoreElements ()) {
		stageDescr descr = (stageDescr) e.nextElement ();

		descr.initargs.put ("simulator_running", "true");
		ConfigData config_data = 
		    new ConfigData (mgr, descr.initargs);

		String nid_param = config_data.getString ("node_id");
		if ((nid_param == null) || 
		    (! nid_param.equals (node_id.toString ()))) {
		    System.err.println ("Error: the node_id parameter " +
			    "passed into the " + descr.stageName + 
			    " stage of node " + node_id + " was " + 
			    nid_param + ".");
		    System.exit (1);
		}

		Class c = null;
		try {
		    c = Class.forName (descr.className);
		}
		catch (ClassNotFoundException exception) {
		    System.err.println ("Error: could not load class " +
			    descr.className + ".");
		    System.exit (1);
		}

		EventHandlerIF handler = null;
		try {
		    handler = (EventHandlerIF) c.newInstance ();
		}
		catch (Exception exception) {
		    System.err.println ("Error: caught exception " +
			    exception + " while trying to instanciate " +
			    "an object of type " + descr.className + 
			    ".  Stack trace follows:");
		    exception.printStackTrace ();
		    System.exit (1);
		}

		SinkIF sink = new SimulatorSink (handler);
		StageIF stage = 
		    new SimulatorStage (descr.stageName, null, sink);
		config_data.setStage (stage);
		mgr.addStage (descr.stageName, stage);

		/* System.out.println ("Initializing stage " + descr.stageName
			+ " on node " + node_id.port () + ".");*/
		try {
		    handler.init (config_data);
		}
		catch (Exception exception) {
		    System.err.println ("Error: caught an exception " +
			    exception +
			    " while trying to initialize the stage " +
			    descr.stageName + " on node " + node_id + "." +
			    "  Stack trace follows:");
		    exception.printStackTrace ();
		    System.exit (1);
		}
	    }

	    // Enqueue a StagesInitializedSignal event.

	    /*System.out.println ("Sending StagesInitializedSignal to node " 
		    + node_id.port () + ".");*/
	    try {
		classifier.enqueue (new StagesInitializedSignal ());
	    }
	    catch (SinkException exception) {
		System.err.println ("Error: caught exception " + 
			exception + " while trying to send a " +
			"StagesInitializedSignal to the classifier " +
			"on node " + node_id + ".  Stack trace " +
			"follows:");
		exception.printStackTrace ();
		System.exit (1);
	    }
	}

	// Run the simulator.

	System.out.println ("*** Running simulation. ***");

	event_queue.run ();
	network.print_byte_counts ();
	network.clear_byte_counts ();

	System.out.println ("*** Performing update bandwidth test ***");

	double time = event_queue.current_time () + 1.0;

	Random rand = new Random (seed);

	for (int node_num = 0; node_num < node_ids.size (); ++node_num) {
	    NodeId node_id = (NodeId) node_ids.elementAt (node_num);
	    Classifier classifier = 
		(Classifier) node_id_to_classifier.get (node_id);

	    for (int i = 0; i < 1; ++i) {
		TapestryPublishMsg publish = new TapestryPublishMsg (
			new SHA1Hash ("bw-" + rand.nextInt ()), null);
		event_queue.schedule (publish, time, classifier);
	    }
	}

	event_queue.run ();
	network.print_byte_counts ();

	System.out.println ("*** Shutting down. ***");

	for (int node_num = 0; node_num < node_ids.size (); ++node_num) {
	    NodeId node_id = (NodeId) node_ids.elementAt (node_num);

	    Classifier classifier = 
		(Classifier) node_id_to_classifier.get (node_id);
	    classifier.dispatch (new SimulationShutdownEvent ());
	}

	System.out.println ("*** Simulation finished normally. ***");
	System.out.flush ();
	System.err.flush ();
	System.exit (0);
    }
}
	

