/*
 * Network.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package soss.network;
import java.util.HashMap;
import java.util.Map;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.StreamTokenizer;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.Reader;
import ostore.util.Assert;
import ostore.util.ByteUtils;
import ostore.util.NodeId;
import ostore.util.Pair;
import ostore.network.NetworkMessage;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import soss.core.EventQueue;
import mdw.sandStorm.api.ConfigDataIF;
import mdw.sandStorm.api.EventHandlerException;
import mdw.sandStorm.api.EventHandlerIF;
import mdw.sandStorm.api.QueueElementIF;
import mdw.sandStorm.api.SinkIF;
import mdw.sandStorm.api.SinkException;

/**
 * A network stage for simulated machines.
 *
 * @author Sean C. Rhea
 * @version $Id: Network.java,v 1.6 2002/03/22 00:15:54 srhea Exp $
 */
public class Network {

    private static final String class_tag = "Network";
    private static final boolean DEBUG = true;

    public class Stage implements EventHandlerIF {

	public Stage (NodeId node_id) {
	    _self_node_id = node_id;
	}

	public void init (ConfigDataIF config) throws Exception {

	    final String tag = "soss.network.Network";
	    SinkIF sink = config.getStage().getSink();

	    String [] msg_types = {
		"ostore.network.NetworkMessage"
	    };

	    for (int i = 0; i < msg_types.length; ++i) {
		Filter filter = new Filter ();
		if (! filter.requireType (Class.forName (msg_types [i]))) {
		    System.err.println (tag + ": could not require type " + 
			    msg_types [i]);
		    System.exit (1);
		}
		if (! filter.requireValue ("inbound", new Boolean (false))) {
		    System.err.println (tag + 
			    ": could not require inbound = false for "
			    + msg_types [i]);
		    System.exit (1);
		}
		Classifier.getClassifier(_self_node_id).subscribe (
			filter, sink);
	    }
	}

	public void destroy() throws Exception {
	    // NOT USED
	}

	public void handleEvent(QueueElementIF item) 
	    throws EventHandlerException {
		final String tag = class_tag + ".handleEvent";

		if (item instanceof NetworkMessage) {
		    handle_network_message ((NetworkMessage) item);
		}
		else {
		    Assert.assert (false, tag + ": Oops!");
		}
	    }

	public void handleEvents(QueueElementIF items[]) 
	    throws EventHandlerException {
		for(int i=0; i<items.length; i++)
		    handleEvent(items[i]);
	    }

	void handle_network_message (NetworkMessage msg) {
	    final String tag = class_tag + ".handle_network_message";

	    Assert.assert (! msg.inbound, tag + ": Oops!");

	    // Find the destination's queue.
	    Classifier dest = 
		(Classifier) _node_id_to_classifier.get (msg.peer);

	    // Find the message size.
	    int [] offset = {0};
	    long type_code = ostore.util.TypeTable.type_code (msg);
	    ByteUtils.longToBytes (type_code, null, offset);
	    msg.to_bytes (null, offset);
	    int msg_size = offset [0];

	    // Find the shortest path.

	    // Compute the total latency and minimum bandwidth.

	    int from = node_id_to_int (_self_node_id);
	    int to = node_id_to_int (msg.peer);

	    TransitStub.PathInfo p = _ts.shortest_path (to, from, 
		    msg_size, true, type_code);

	    double routing_time = p.route_time;
	    double recv_time = _event_queue.current_time () + routing_time;

	    // Schedule the event for the receiver's queue

	    msg.inbound = true;  // TODO: violates const semantics
	    msg.peer = _self_node_id;  // TODO: violates const semantics
	    _event_queue.schedule (msg, recv_time, dest);

	    // TODO: test serialization?
	}

	private NodeId _self_node_id;
    }

    public Network (
	    EventQueue event_queue, 
	    Map node_id_to_classifier,
	    String ts_file_name) 
	throws FileNotFoundException, IOException {

	if (DEBUG) 
	    Assert.assert (_instance == null, "Network already created.");

	_instance = this;

	_event_queue = event_queue;
	_node_id_to_classifier = node_id_to_classifier;
	_ts = new TransitStub (ts_file_name);
    }

    public static Network instance () {
	if (DEBUG) 
	    Assert.assert (_instance != null, "No network created yet.");

	return _instance;
    }

    public void print_byte_counts () {
	_ts.print_byte_counts ();
    }

    public void clear_byte_counts () {
	_ts.clear_byte_counts ();
    }

    public long route_time_ms (NodeId a, NodeId b, int msg_size) {
	int from = node_id_to_int (a);
	int to = node_id_to_int (b);
	soss.network.TransitStub.PathInfo p = _ts.shortest_path (to, from, msg_size, false, 0);
	double routing_time = p.route_time;
	return Math.round (routing_time * 1000.0);
    }

    public Stage new_stage (NodeId node_id) {
	return new Stage (node_id);
    }

    private int node_id_to_int (NodeId node_id) {
	return node_id.port ();
    }

    private static Network _instance;

    private TransitStub _ts;
    private EventQueue _event_queue;
    private Map _node_id_to_classifier;
}

